__all__ = ()

from scarletio import export

from .....discord.application_command import ApplicationCommandOptionType


APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND = ApplicationCommandOptionType.sub_command
APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY = ApplicationCommandOptionType.sub_command_group

APPLICATION_COMMAND_FUNCTION_DEEPNESS = -1
APPLICATION_COMMAND_HANDLER_DEEPNESS = 0
APPLICATION_COMMAND_DEEPNESS = 1
APPLICATION_COMMAND_CATEGORY_DEEPNESS_START = 2
APPLICATION_COMMAND_CATEGORY_DEEPNESS_MAX = 3


export(APPLICATION_COMMAND_FUNCTION_DEEPNESS, 'APPLICATION_COMMAND_FUNCTION_DEEPNESS')
