__all__ = ()

from .preinstanced import ButtonStyle, TextInputStyle


BUTTON_STYLE_DEFAULT = ButtonStyle.blue

TEXT_INPUT_STYLE_DEFAULT = TextInputStyle.short

LABEL_LENGTH_MAX = 80

PLACEHOLDER_LENGTH_MAX = 150

OPTIONS_LENGTH_MIN = 1
OPTIONS_LENGTH_MAX = 25

MAX_LENGTH_MIN = 0
MAX_LENGTH_MAX = 4000
MAX_LENGTH_DEFAULT = 0

MIN_LENGTH_MIN = 0
MIN_LENGTH_MAX = 4000
MIN_LENGTH_DEFAULT = 0

MAX_VALUES_MIN = 1
MAX_VALUES_MAX = 25
MAX_VALUES_DEFAULT = 1

MIN_VALUES_MIN = 0
MIN_VALUES_MAX = 15
MIN_VALUES_DEFAULT = 1

VALUE_LENGTH_MAX = 4000
