__all__ = ()

from .emoji import Emoji
from .unicode_type import Unicode


UNICODES = (
    Unicode(
        'skin_tone_1',
        b'\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'skin_tone_2',
        b'\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'skin_tone_3',
        b'\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'skin_tone_4',
        b'\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'skin_tone_5',
        b'\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'sweat',
        b'\xf0\x9f\x98\x93',
        False,
        None,
        (',:(', ',:-(', ',=(', ',=-('),
    ),
    Unicode(
        'sweat_smile',
        b'\xf0\x9f\x98\x85',
        False,
        None,
        (',:)', ',:-)', ',=)', ',=-)'),
    ),
    Unicode(
        'innocent',
        b'\xf0\x9f\x98\x87',
        False,
        None,
        ('0:)', '0:-)', '0=)', '0=-)', 'o:)', 'O:)', 'o:-)', 'O:-)', 'o=)', 'O=)', 'o=-)', 'O=-)'),
    ),
    Unicode(
        'sunglasses',
        b'\xf0\x9f\x98\x8e',
        False,
        None,
        ('8-)', 'B-)'),
    ),
    Unicode(
        'unamused',
        b'\xf0\x9f\x98\x92',
        False,
        None,
        (':$', ':-$', ':-S', ':-Z', ':s', ':z', '=$', '=-$', '=-S', '=-Z', '=s', '=z'),
    ),
    Unicode(
        'cry',
        b'\xf0\x9f\x98\xa2',
        False,
        None,
        (':\'(', ':\'-(', ':,(', ':,-(', '=\'(', '=\'-(', '=,(', '=,-('),
    ),
    Unicode(
        'joy',
        b'\xf0\x9f\x98\x82',
        False,
        None,
        (
            ':\')', ':\'-)', ':\'-D', ':\'D', ':,)', ':,-)', ':,-D', ':,D', '=\')', '=\'-)', '=\'-D', '=\'D',
            '=,)', '=,-)', '=,-D', '=,D'
        ),
    ),
    Unicode(
        'frowning',
        b'\xf0\x9f\x98\xa6',
        False,
        None,
        (':(', ':-(', '=(', '=-('),
    ),
    Unicode(
        'smiley',
        b'\xf0\x9f\x98\x83',
        False,
        None,
        (':)', ':-)', '=)', '=-)'),
    ),
    Unicode(
        'kissing',
        b'\xf0\x9f\x98\x97',
        False,
        None,
        (':*', ':-*', '=*', '=-*'),
    ),
    Unicode(
        'thumbsup',
        b'\xf0\x9f\x91\x8d',
        False,
        (
            'thumbup',
            '+1'
        ),
        None,
    ),
    Unicode(
        'thumbsup_tone1',
        b'\xf0\x9f\x91\x8d\xf0\x9f\x8f\xbb',
        False,
        (
            'thumbup_tone1',
            '+1_tone1'
        ),
        None,
    ),
    Unicode(
        'thumbsup_tone2',
        b'\xf0\x9f\x91\x8d\xf0\x9f\x8f\xbc',
        False,
        (
            'thumbup_tone2',
            '+1_tone2'
        ),
        None,
    ),
    Unicode(
        'thumbsup_tone3',
        b'\xf0\x9f\x91\x8d\xf0\x9f\x8f\xbd',
        False,
        (
            'thumbup_tone3',
            '+1_tone3'
        ),
        None,
    ),
    Unicode(
        'thumbsup_tone4',
        b'\xf0\x9f\x91\x8d\xf0\x9f\x8f\xbe',
        False,
        (
            'thumbup_tone4',
            '+1_tone4'
        ),
        None,
    ),
    Unicode(
        'thumbsup_tone5',
        b'\xf0\x9f\x91\x8d\xf0\x9f\x8f\xbf',
        False,
        (
            'thumbup_tone5',
            '+1_tone5'
        ),
        None,
    ),
    Unicode(
        'sob',
        b'\xf0\x9f\x98\xad',
        False,
        None,
        (':,\'(', ':,\'-(', ';(', ';-(', '=,\'(', '=,\'-('),
    ),
    Unicode(
        'confused',
        b'\xf0\x9f\x98\x95',
        False,
        None,
        (':-/', ':-\\', '=-/', '=-\\'),
    ),
    Unicode(
        'thumbdown',
        b'\xf0\x9f\x91\x8e',
        False,
        (
            'thumbsdown',
        ),
        ('-1',),
    ),
    Unicode(
        'thumbdown_tone1',
        b'\xf0\x9f\x91\x8e\xf0\x9f\x8f\xbb',
        False,
        (
            'thumbsdown_tone1',
            '_1_tone1'
        ),
        ('-1_tone1',),
    ),
    Unicode(
        'thumbdown_tone2',
        b'\xf0\x9f\x91\x8e\xf0\x9f\x8f\xbc',
        False,
        (
            'thumbsdown_tone2',
            '_1_tone2'
        ),
        ('-1_tone2',),
    ),
    Unicode(
        'thumbdown_tone3',
        b'\xf0\x9f\x91\x8e\xf0\x9f\x8f\xbd',
        False,
        (
            'thumbsdown_tone3',
            '_1_tone3'
        ),
        ('-1_tone3',),
    ),
    Unicode(
        'thumbdown_tone4',
        b'\xf0\x9f\x91\x8e\xf0\x9f\x8f\xbe',
        False,
        (
            'thumbsdown_tone4',
            '_1_tone4'
        ),
        ('-1_tone4',),
    ),
    Unicode(
        'thumbdown_tone5',
        b'\xf0\x9f\x91\x8e\xf0\x9f\x8f\xbf',
        False,
        (
            'thumbsdown_tone5',
            '_1_tone5'
        ),
        ('-1_tone5',),
    ),
    Unicode(
        'rage',
        b'\xf0\x9f\x98\xa1',
        False,
        None,
        (':-@', ':@', '=-@', '=@'),
    ),
    Unicode(
        'blush',
        b'\xf0\x9f\x98\x8a',
        False,
        None,
        (':-")', ':")', '=-")', '=")'),
    ),
    Unicode(
        'smile',
        b'\xf0\x9f\x98\x84',
        False,
        None,
        (':-D', ':D', '=-D', '=D'),
    ),
    Unicode(
        'open_mouth',
        b'\xf0\x9f\x98\xae',
        False,
        None,
        (':-o', ':-O', ':o', ':O', '=-o', '=-O', '=o', '=O'),
    ),
    Unicode(
        'stuck_out_tongue',
        b'\xf0\x9f\x98\x9b',
        False,
        None,
        (':-P', ':P', '=-P', '=P'),
    ),
    Unicode(
        'neutral_face',
        b'\xf0\x9f\x98\x90',
        False,
        None,
        (':-|', ':|', '=-|', '=|'),
    ),
    Unicode(
        '100',
        b'\xf0\x9f\x92\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        '1234',
        b'\xf0\x9f\x94\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        '8ball',
        b'\xf0\x9f\x8e\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'a',
        b'\xf0\x9f\x85\xb0',
        True,
        None,
        None,
    ),
    Unicode(
        'ab',
        b'\xf0\x9f\x86\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'abc',
        b'\xf0\x9f\x94\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'abcd',
        b'\xf0\x9f\x94\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'accept',
        b'\xf0\x9f\x89\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'admission_tickets',
        b'\xf0\x9f\x8e\x9f',
        True,
        None,
        None,
    ),
    Unicode(
        'aerial_tramway',
        b'\xf0\x9f\x9a\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'airplane',
        b'\xe2\x9c\x88',
        True,
        None,
        None,
    ),
    Unicode(
        'airplane_arriving',
        b'\xf0\x9f\x9b\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'airplane_departure',
        b'\xf0\x9f\x9b\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'airplane_small',
        b'\xf0\x9f\x9b\xa9',
        True,
        None,
        None,
    ),
    Unicode(
        'alarm_clock',
        b'\xe2\x8f\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'alembic',
        b'\xe2\x9a\x97',
        True,
        None,
        None,
    ),
    Unicode(
        'alien',
        b'\xf0\x9f\x91\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'ambulance',
        b'\xf0\x9f\x9a\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'amphora',
        b'\xf0\x9f\x8f\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'anchor',
        b'\xe2\x9a\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'angel',
        b'\xf0\x9f\x91\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'angel_tone1',
        b'\xf0\x9f\x91\xbc\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'angel_tone2',
        b'\xf0\x9f\x91\xbc\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'angel_tone3',
        b'\xf0\x9f\x91\xbc\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'angel_tone4',
        b'\xf0\x9f\x91\xbc\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'angel_tone5',
        b'\xf0\x9f\x91\xbc\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'anger',
        b'\xf0\x9f\x92\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'anger_right',
        b'\xf0\x9f\x97\xaf',
        True,
        None,
        None,
    ),
    Unicode(
        'angry',
        b'\xf0\x9f\x98\xa0',
        False,
        None,
        ('>:(', '>:-(', '>=(', '>=-('),
    ),
    Unicode(
        'anguished',
        b'\xf0\x9f\x98\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'ant',
        b'\xf0\x9f\x90\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'apple',
        b'\xf0\x9f\x8d\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'aquarius',
        b'\xe2\x99\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'archery',
        b'\xf0\x9f\x8f\xb9',
        False,
        (
            'bow_and_arrow',
        ),
        None,
    ),
    Unicode(
        'aries',
        b'\xe2\x99\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_backward',
        b'\xe2\x97\x80',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_double_down',
        b'\xe2\x8f\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_double_up',
        b'\xe2\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_down',
        b'\xe2\xac\x87',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_down_small',
        b'\xf0\x9f\x94\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_forward',
        b'\xe2\x96\xb6',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_heading_down',
        b'\xe2\xa4\xb5',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_heading_up',
        b'\xe2\xa4\xb4',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_left',
        b'\xe2\xac\x85',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_lower_left',
        b'\xe2\x86\x99',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_lower_right',
        b'\xe2\x86\x98',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_right',
        b'\xe2\x9e\xa1',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_right_hook',
        b'\xe2\x86\xaa',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_up',
        b'\xe2\xac\x86',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_up_down',
        b'\xe2\x86\x95',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_up_small',
        b'\xf0\x9f\x94\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_upper_left',
        b'\xe2\x86\x96',
        True,
        None,
        None,
    ),
    Unicode(
        'arrow_upper_right',
        b'\xe2\x86\x97',
        True,
        None,
        None,
    ),
    Unicode(
        'arrows_clockwise',
        b'\xf0\x9f\x94\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'arrows_counterclockwise',
        b'\xf0\x9f\x94\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'art',
        b'\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'articulated_lorry',
        b'\xf0\x9f\x9a\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'asterisk',
        b'*\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'astonished',
        b'\xf0\x9f\x98\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'athletic_shoe',
        b'\xf0\x9f\x91\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'atm',
        b'\xf0\x9f\x8f\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'atom',
        b'\xe2\x9a\x9b',
        True,
        None,
        None,
    ),
    Unicode(
        'avocado',
        b'\xf0\x9f\xa5\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'b',
        b'\xf0\x9f\x85\xb1',
        True,
        None,
        None,
    ),
    Unicode(
        'baby',
        b'\xf0\x9f\x91\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'baby_tone1',
        b'\xf0\x9f\x91\xb6\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'baby_tone2',
        b'\xf0\x9f\x91\xb6\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'baby_tone3',
        b'\xf0\x9f\x91\xb6\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'baby_tone4',
        b'\xf0\x9f\x91\xb6\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'baby_tone5',
        b'\xf0\x9f\x91\xb6\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'baby_bottle',
        b'\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'baby_chick',
        b'\xf0\x9f\x90\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'baby_symbol',
        b'\xf0\x9f\x9a\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'back',
        b'\xf0\x9f\x94\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'back_of_hand',
        b'\xf0\x9f\xa4\x9a',
        False,
        (
            'raised_back_of_hand',
        ),
        None,
    ),
    Unicode(
        'back_of_hand_tone1',
        b'\xf0\x9f\xa4\x9a\xf0\x9f\x8f\xbb',
        False,
        (
            'raised_back_of_hand_tone1',
        ),
        None,
    ),
    Unicode(
        'back_of_hand_tone2',
        b'\xf0\x9f\xa4\x9a\xf0\x9f\x8f\xbc',
        False,
        (
            'raised_back_of_hand_tone2',
        ),
        None,
    ),
    Unicode(
        'back_of_hand_tone3',
        b'\xf0\x9f\xa4\x9a\xf0\x9f\x8f\xbd',
        False,
        (
            'raised_back_of_hand_tone3',
        ),
        None,
    ),
    Unicode(
        'back_of_hand_tone4',
        b'\xf0\x9f\xa4\x9a\xf0\x9f\x8f\xbe',
        False,
        (
            'raised_back_of_hand_tone4',
        ),
        None,
    ),
    Unicode(
        'back_of_hand_tone5',
        b'\xf0\x9f\xa4\x9a\xf0\x9f\x8f\xbf',
        False,
        (
            'raised_back_of_hand_tone5',
        ),
        None,
    ),
    Unicode(
        'bacon',
        b'\xf0\x9f\xa5\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'badminton',
        b'\xf0\x9f\x8f\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'baggage_claim',
        b'\xf0\x9f\x9b\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'baguette_bread',
        b'\xf0\x9f\xa5\x96',
        False,
        (
            'french_bread',
        ),
        None,
    ),
    Unicode(
        'balloon',
        b'\xf0\x9f\x8e\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'ballot_box',
        b'\xf0\x9f\x97\xb3',
        True,
        None,
        None,
    ),
    Unicode(
        'ballot_box_with_check',
        b'\xe2\x98\x91',
        True,
        None,
        None,
    ),
    Unicode(
        'bamboo',
        b'\xf0\x9f\x8e\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'banana',
        b'\xf0\x9f\x8d\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'bangbang',
        b'\xe2\x80\xbc',
        True,
        None,
        None,
    ),
    Unicode(
        'bank',
        b'\xf0\x9f\x8f\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'bar_chart',
        b'\xf0\x9f\x93\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'barber',
        b'\xf0\x9f\x92\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'baseball',
        b'\xe2\x9a\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'basketball',
        b'\xf0\x9f\x8f\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'basketball_player',
        b'\xe2\x9b\xb9',
        True,
        None,
        None,
    ),
    Unicode(
        'basketball_player_tone1',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbb',
        False,
        (
            'person_with_ball_tone1',
            'person_bouncing_ball_tone1'
        ),
        None,
    ),
    Unicode(
        'basketball_player_tone2',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbc',
        False,
        (
            'person_with_ball_tone2',
            'person_bouncing_ball_tone2'
        ),
        None,
    ),
    Unicode(
        'basketball_player_tone3',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbd',
        False,
        (
            'person_with_ball_tone3',
            'person_bouncing_ball_tone3'
        ),
        None,
    ),
    Unicode(
        'basketball_player_tone4',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbe',
        False,
        (
            'person_with_ball_tone4',
            'person_bouncing_ball_tone4'
        ),
        None,
    ),
    Unicode(
        'basketball_player_tone5',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbf',
        False,
        (
            'person_with_ball_tone5',
            'person_bouncing_ball_tone5'
        ),
        None,
    ),
    Unicode(
        'bat',
        b'\xf0\x9f\xa6\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'bath',
        b'\xf0\x9f\x9b\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'bath_tone1',
        b'\xf0\x9f\x9b\x80\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'bath_tone2',
        b'\xf0\x9f\x9b\x80\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'bath_tone3',
        b'\xf0\x9f\x9b\x80\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'bath_tone4',
        b'\xf0\x9f\x9b\x80\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'bath_tone5',
        b'\xf0\x9f\x9b\x80\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'bathtub',
        b'\xf0\x9f\x9b\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'battery',
        b'\xf0\x9f\x94\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'beach',
        b'\xf0\x9f\x8f\x96',
        True,
        None,
        None,
    ),
    Unicode(
        'beach_umbrella',
        b'\xe2\x9b\xb1',
        True,
        None,
        None,
    ),
    Unicode(
        'bear',
        b'\xf0\x9f\x90\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'bed',
        b'\xf0\x9f\x9b\x8f',
        True,
        None,
        None,
    ),
    Unicode(
        'bee',
        b'\xf0\x9f\x90\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'beer',
        b'\xf0\x9f\x8d\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'beers',
        b'\xf0\x9f\x8d\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'beetle',
        b'\xf0\x9f\x90\x9e',
        True,
        None,
        None,
    ),
    Unicode(
        'beginner',
        b'\xf0\x9f\x94\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'bell',
        b'\xf0\x9f\x94\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'bellhop',
        b'\xf0\x9f\x9b\x8e',
        True,
        None,
        None,
    ),
    Unicode(
        'bento',
        b'\xf0\x9f\x8d\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'bicyclist',
        b'\xf0\x9f\x9a\xb4',
        False,
        (
            'person_biking',
        ),
        None,
    ),
    Unicode(
        'bicyclist_tone1',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbb',
        False,
        (
            'person_biking_tone1',
        ),
        None,
    ),
    Unicode(
        'bicyclist_tone2',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbc',
        False,
        (
            'person_biking_tone2',
        ),
        None,
    ),
    Unicode(
        'bicyclist_tone3',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbd',
        False,
        (
            'person_biking_tone3',
        ),
        None,
    ),
    Unicode(
        'bicyclist_tone4',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbe',
        False,
        (
            'person_biking_tone4',
        ),
        None,
    ),
    Unicode(
        'bicyclist_tone5',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbf',
        False,
        (
            'person_biking_tone5',
        ),
        None,
    ),
    Unicode(
        'bike',
        b'\xf0\x9f\x9a\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'bikini',
        b'\xf0\x9f\x91\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'biohazard',
        b'\xe2\x98\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'bird',
        b'\xf0\x9f\x90\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'birthday',
        b'\xf0\x9f\x8e\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'black_circle',
        b'\xe2\x9a\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'black_heart',
        b'\xf0\x9f\x96\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'black_joker',
        b'\xf0\x9f\x83\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'black_large_square',
        b'\xe2\xac\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'black_medium_small_square',
        b'\xe2\x97\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'black_medium_square',
        b'\xe2\x97\xbc',
        True,
        None,
        None,
    ),
    Unicode(
        'black_nib',
        b'\xe2\x9c\x92',
        True,
        None,
        None,
    ),
    Unicode(
        'black_small_square',
        b'\xe2\x96\xaa',
        True,
        None,
        None,
    ),
    Unicode(
        'black_square_button',
        b'\xf0\x9f\x94\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'blossom',
        b'\xf0\x9f\x8c\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'blowfish',
        b'\xf0\x9f\x90\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'blue_book',
        b'\xf0\x9f\x93\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'blue_car',
        b'\xf0\x9f\x9a\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'blue_heart',
        b'\xf0\x9f\x92\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'boar',
        b'\xf0\x9f\x90\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'bomb',
        b'\xf0\x9f\x92\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'book',
        b'\xf0\x9f\x93\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'bookmark',
        b'\xf0\x9f\x94\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'bookmark_tabs',
        b'\xf0\x9f\x93\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'books',
        b'\xf0\x9f\x93\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'boom',
        b'\xf0\x9f\x92\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'boot',
        b'\xf0\x9f\x91\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'bottle_with_popping_cork',
        b'\xf0\x9f\x8d\xbe',
        False,
        (
            'champagne',
        ),
        None,
    ),
    Unicode(
        'bouquet',
        b'\xf0\x9f\x92\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'bow',
        b'\xf0\x9f\x99\x87',
        False,
        (
            'person_bowing',
        ),
        None,
    ),
    Unicode(
        'bow_tone1',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbb',
        False,
        (
            'person_bowing_tone1',
        ),
        None,
    ),
    Unicode(
        'bow_tone2',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbc',
        False,
        (
            'person_bowing_tone2',
        ),
        None,
    ),
    Unicode(
        'bow_tone3',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbd',
        False,
        (
            'person_bowing_tone3',
        ),
        None,
    ),
    Unicode(
        'bow_tone4',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbe',
        False,
        (
            'person_bowing_tone4',
        ),
        None,
    ),
    Unicode(
        'bow_tone5',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbf',
        False,
        (
            'person_bowing_tone5',
        ),
        None,
    ),
    Unicode(
        'bowling',
        b'\xf0\x9f\x8e\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'boxing_glove',
        b'\xf0\x9f\xa5\x8a',
        False,
        (
            'boxing_gloves',
        ),
        None,
    ),
    Unicode(
        'boy',
        b'\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'boy_tone1',
        b'\xf0\x9f\x91\xa6\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'boy_tone2',
        b'\xf0\x9f\x91\xa6\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'boy_tone3',
        b'\xf0\x9f\x91\xa6\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'boy_tone4',
        b'\xf0\x9f\x91\xa6\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'boy_tone5',
        b'\xf0\x9f\x91\xa6\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'bread',
        b'\xf0\x9f\x8d\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'bride_with_veil',
        b'\xf0\x9f\x91\xb0',
        False,
        (
            'person_with_veil',
        ),
        None,
    ),
    Unicode(
        'bride_with_veil_tone1',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbb',
        False,
        (
            'person_with_veil_tone1',
        ),
        None,
    ),
    Unicode(
        'bride_with_veil_tone2',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbc',
        False,
        (
            'person_with_veil_tone2',
        ),
        None,
    ),
    Unicode(
        'bride_with_veil_tone3',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbd',
        False,
        (
            'person_with_veil_tone3',
        ),
        None,
    ),
    Unicode(
        'bride_with_veil_tone4',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbe',
        False,
        (
            'person_with_veil_tone4',
        ),
        None,
    ),
    Unicode(
        'bride_with_veil_tone5',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbf',
        False,
        (
            'person_with_veil_tone5',
        ),
        None,
    ),
    Unicode(
        'bridge_at_night',
        b'\xf0\x9f\x8c\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'briefcase',
        b'\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'broken_heart',
        b'\xf0\x9f\x92\x94',
        False,
        None,
        ('</3', '<\\3'),
    ),
    Unicode(
        'bug',
        b'\xf0\x9f\x90\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'building_construction',
        b'\xf0\x9f\x8f\x97',
        True,
        None,
        None,
    ),
    Unicode(
        'bulb',
        b'\xf0\x9f\x92\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'bullettrain_front',
        b'\xf0\x9f\x9a\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'bullettrain_side',
        b'\xf0\x9f\x9a\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'burrito',
        b'\xf0\x9f\x8c\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'bus',
        b'\xf0\x9f\x9a\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'busstop',
        b'\xf0\x9f\x9a\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'bust_in_silhouette',
        b'\xf0\x9f\x91\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'busts_in_silhouette',
        b'\xf0\x9f\x91\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'butterfly',
        b'\xf0\x9f\xa6\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'cactus',
        b'\xf0\x9f\x8c\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'cake',
        b'\xf0\x9f\x8d\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'calendar',
        b'\xf0\x9f\x93\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'calendar_spiral',
        b'\xf0\x9f\x97\x93',
        True,
        None,
        None,
    ),
    Unicode(
        'call_me',
        b'\xf0\x9f\xa4\x99',
        False,
        (
            'call_me_hand',
        ),
        None,
    ),
    Unicode(
        'call_me_tone1',
        b'\xf0\x9f\xa4\x99\xf0\x9f\x8f\xbb',
        False,
        (
            'call_me_hand_tone1',
        ),
        None,
    ),
    Unicode(
        'call_me_tone2',
        b'\xf0\x9f\xa4\x99\xf0\x9f\x8f\xbc',
        False,
        (
            'call_me_hand_tone2',
        ),
        None,
    ),
    Unicode(
        'call_me_tone3',
        b'\xf0\x9f\xa4\x99\xf0\x9f\x8f\xbd',
        False,
        (
            'call_me_hand_tone3',
        ),
        None,
    ),
    Unicode(
        'call_me_tone4',
        b'\xf0\x9f\xa4\x99\xf0\x9f\x8f\xbe',
        False,
        (
            'call_me_hand_tone4',
        ),
        None,
    ),
    Unicode(
        'call_me_tone5',
        b'\xf0\x9f\xa4\x99\xf0\x9f\x8f\xbf',
        False,
        (
            'call_me_hand_tone5',
        ),
        None,
    ),
    Unicode(
        'calling',
        b'\xf0\x9f\x93\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'camel',
        b'\xf0\x9f\x90\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'camera',
        b'\xf0\x9f\x93\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'camera_with_flash',
        b'\xf0\x9f\x93\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'camping',
        b'\xf0\x9f\x8f\x95',
        True,
        None,
        None,
    ),
    Unicode(
        'cancer',
        b'\xe2\x99\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'candle',
        b'\xf0\x9f\x95\xaf',
        True,
        None,
        None,
    ),
    Unicode(
        'candy',
        b'\xf0\x9f\x8d\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'canoe',
        b'\xf0\x9f\x9b\xb6',
        False,
        (
            'kayak',
        ),
        None,
    ),
    Unicode(
        'capital_abcd',
        b'\xf0\x9f\x94\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'capricorn',
        b'\xe2\x99\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'card_box',
        b'\xf0\x9f\x97\x83',
        True,
        None,
        None,
    ),
    Unicode(
        'card_index',
        b'\xf0\x9f\x93\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'card_index_dividers',
        b'\xf0\x9f\x97\x82',
        True,
        None,
        None,
    ),
    Unicode(
        'carousel_horse',
        b'\xf0\x9f\x8e\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'carrot',
        b'\xf0\x9f\xa5\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'cartwheel',
        b'\xf0\x9f\xa4\xb8',
        False,
        (
            'person_doing_cartwheel',
        ),
        None,
    ),
    Unicode(
        'cartwheel_tone1',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbb',
        False,
        (
            'person_doing_cartwheel_tone1',
        ),
        None,
    ),
    Unicode(
        'cartwheel_tone2',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbc',
        False,
        (
            'person_doing_cartwheel_tone2',
        ),
        None,
    ),
    Unicode(
        'cartwheel_tone3',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbd',
        False,
        (
            'person_doing_cartwheel_tone3',
        ),
        None,
    ),
    Unicode(
        'cartwheel_tone4',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbe',
        False,
        (
            'person_doing_cartwheel_tone4',
        ),
        None,
    ),
    Unicode(
        'cartwheel_tone5',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbf',
        False,
        (
            'person_doing_cartwheel_tone5',
        ),
        None,
    ),
    Unicode(
        'cat2',
        b'\xf0\x9f\x90\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'cat',
        b'\xf0\x9f\x90\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'cd',
        b'\xf0\x9f\x92\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'chains',
        b'\xe2\x9b\x93',
        True,
        None,
        None,
    ),
    Unicode(
        'champagne_glass',
        b'\xf0\x9f\xa5\x82',
        False,
        (
            'clinking_glass',
        ),
        None,
    ),
    Unicode(
        'chart',
        b'\xf0\x9f\x92\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'chart_with_downwards_trend',
        b'\xf0\x9f\x93\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'chart_with_upwards_trend',
        b'\xf0\x9f\x93\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'checkered_flag',
        b'\xf0\x9f\x8f\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'cheese',
        b'\xf0\x9f\xa7\x80',
        False,
        (
            'cheese_wedge',
        ),
        None,
    ),
    Unicode(
        'cherries',
        b'\xf0\x9f\x8d\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'cherry_blossom',
        b'\xf0\x9f\x8c\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'chestnut',
        b'\xf0\x9f\x8c\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'chicken',
        b'\xf0\x9f\x90\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'children_crossing',
        b'\xf0\x9f\x9a\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'chipmunk',
        b'\xf0\x9f\x90\xbf',
        True,
        None,
        None,
    ),
    Unicode(
        'chocolate_bar',
        b'\xf0\x9f\x8d\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'christmas_tree',
        b'\xf0\x9f\x8e\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'church',
        b'\xe2\x9b\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'cinema',
        b'\xf0\x9f\x8e\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'circus_tent',
        b'\xf0\x9f\x8e\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'city_dusk',
        b'\xf0\x9f\x8c\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'city_sunrise',
        b'\xf0\x9f\x8c\x87',
        False,
        (
            'city_sunset',
        ),
        None,
    ),
    Unicode(
        'cityscape',
        b'\xf0\x9f\x8f\x99',
        True,
        None,
        None,
    ),
    Unicode(
        'cl',
        b'\xf0\x9f\x86\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'clap',
        b'\xf0\x9f\x91\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'clap_tone1',
        b'\xf0\x9f\x91\x8f\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'clap_tone2',
        b'\xf0\x9f\x91\x8f\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'clap_tone3',
        b'\xf0\x9f\x91\x8f\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'clap_tone4',
        b'\xf0\x9f\x91\x8f\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'clap_tone5',
        b'\xf0\x9f\x91\x8f\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'clapper',
        b'\xf0\x9f\x8e\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'classical_building',
        b'\xf0\x9f\x8f\x9b',
        True,
        None,
        None,
    ),
    Unicode(
        'clipboard',
        b'\xf0\x9f\x93\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'clock1030',
        b'\xf0\x9f\x95\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'clock10',
        b'\xf0\x9f\x95\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'clock1130',
        b'\xf0\x9f\x95\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'clock11',
        b'\xf0\x9f\x95\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'clock1230',
        b'\xf0\x9f\x95\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'clock12',
        b'\xf0\x9f\x95\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'clock130',
        b'\xf0\x9f\x95\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'clock1',
        b'\xf0\x9f\x95\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'clock230',
        b'\xf0\x9f\x95\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'clock2',
        b'\xf0\x9f\x95\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'clock330',
        b'\xf0\x9f\x95\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'clock3',
        b'\xf0\x9f\x95\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'clock430',
        b'\xf0\x9f\x95\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'clock4',
        b'\xf0\x9f\x95\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'clock530',
        b'\xf0\x9f\x95\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'clock5',
        b'\xf0\x9f\x95\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'clock630',
        b'\xf0\x9f\x95\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'clock6',
        b'\xf0\x9f\x95\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'clock730',
        b'\xf0\x9f\x95\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'clock7',
        b'\xf0\x9f\x95\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'clock830',
        b'\xf0\x9f\x95\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'clock8',
        b'\xf0\x9f\x95\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'clock930',
        b'\xf0\x9f\x95\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'clock9',
        b'\xf0\x9f\x95\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'clock',
        b'\xf0\x9f\x95\xb0',
        True,
        None,
        None,
    ),
    Unicode(
        'closed_book',
        b'\xf0\x9f\x93\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'closed_lock_with_key',
        b'\xf0\x9f\x94\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'closed_umbrella',
        b'\xf0\x9f\x8c\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'cloud',
        b'\xe2\x98\x81',
        True,
        None,
        None,
    ),
    Unicode(
        'cloud_lightning',
        b'\xf0\x9f\x8c\xa9',
        True,
        None,
        None,
    ),
    Unicode(
        'cloud_rain',
        b'\xf0\x9f\x8c\xa7',
        True,
        None,
        None,
    ),
    Unicode(
        'cloud_snow',
        b'\xf0\x9f\x8c\xa8',
        True,
        None,
        None,
    ),
    Unicode(
        'cloud_tornado',
        b'\xf0\x9f\x8c\xaa',
        True,
        None,
        None,
    ),
    Unicode(
        'clown',
        b'\xf0\x9f\xa4\xa1',
        False,
        (
            'clown_face',
        ),
        None,
    ),
    Unicode(
        'clubs',
        b'\xe2\x99\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'cocktail',
        b'\xf0\x9f\x8d\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'coffee',
        b'\xe2\x98\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'coffin',
        b'\xe2\x9a\xb0',
        True,
        None,
        None,
    ),
    Unicode(
        'cold_sweat',
        b'\xf0\x9f\x98\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'comet',
        b'\xe2\x98\x84',
        True,
        None,
        None,
    ),
    Unicode(
        'compression',
        b'\xf0\x9f\x97\x9c',
        True,
        None,
        None,
    ),
    Unicode(
        'computer',
        b'\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'confetti_ball',
        b'\xf0\x9f\x8e\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'confounded',
        b'\xf0\x9f\x98\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'congratulations',
        b'\xe3\x8a\x97',
        True,
        None,
        None,
    ),
    Unicode(
        'construction',
        b'\xf0\x9f\x9a\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'construction_worker',
        b'\xf0\x9f\x91\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'construction_worker_tone1',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'construction_worker_tone2',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'construction_worker_tone3',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'construction_worker_tone4',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'construction_worker_tone5',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'control_knobs',
        b'\xf0\x9f\x8e\x9b',
        True,
        None,
        None,
    ),
    Unicode(
        'convenience_store',
        b'\xf0\x9f\x8f\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'cookie',
        b'\xf0\x9f\x8d\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'cooking',
        b'\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'cool',
        b'\xf0\x9f\x86\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'cop',
        b'\xf0\x9f\x91\xae',
        False,
        (
            'police_officer',
        ),
        None,
    ),
    Unicode(
        'cop_tone1',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbb',
        False,
        (
            'police_officer_tone1',
        ),
        None,
    ),
    Unicode(
        'cop_tone2',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbc',
        False,
        (
            'police_officer_tone2',
        ),
        None,
    ),
    Unicode(
        'cop_tone3',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbd',
        False,
        (
            'police_officer_tone3',
        ),
        None,
    ),
    Unicode(
        'cop_tone4',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbe',
        False,
        (
            'police_officer_tone4',
        ),
        None,
    ),
    Unicode(
        'cop_tone5',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbf',
        False,
        (
            'police_officer_tone5',
        ),
        None,
    ),
    Unicode(
        'copyright',
        b'\xc2\xa9',
        True,
        None,
        None,
    ),
    Unicode(
        'corn',
        b'\xf0\x9f\x8c\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'couch',
        b'\xf0\x9f\x9b\x8b',
        True,
        None,
        None,
    ),
    Unicode(
        'couple',
        b'\xf0\x9f\x91\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_mm',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8',
        False,
        (
            'couple_with_heart_mm',
        ),
        None,
    ),
    Unicode(
        'couple_with_heart',
        b'\xf0\x9f\x92\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_ww',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9',
        False,
        (
            'couple_ww',
        ),
        None,
    ),
    Unicode(
        'couplekiss',
        b'\xf0\x9f\x92\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'couplekiss_mm',
        (
            b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8'
        ),
        False,
        (
            'kiss_mm',
        ),
        None,
    ),
    Unicode(
        'couplekiss_ww',
        (
            b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9'
        ),
        False,
        (
            'kiss_ww',
        ),
        None,
    ),
    Unicode(
        'cow2',
        b'\xf0\x9f\x90\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'cow',
        b'\xf0\x9f\x90\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'cowboy',
        b'\xf0\x9f\xa4\xa0',
        False,
        (
            'face_with_cowboy_hat',
        ),
        None,
    ),
    Unicode(
        'crab',
        b'\xf0\x9f\xa6\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'crayon',
        b'\xf0\x9f\x96\x8d',
        True,
        None,
        None,
    ),
    Unicode(
        'credit_card',
        b'\xf0\x9f\x92\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'crescent_moon',
        b'\xf0\x9f\x8c\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'cricket',
        b'\xf0\x9f\x8f\x8f',
        True,
        None,
        None,
    ),
    Unicode(
        'crocodile',
        b'\xf0\x9f\x90\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'croissant',
        b'\xf0\x9f\xa5\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'cross',
        b'\xe2\x9c\x9d',
        True,
        None,
        None,
    ),
    Unicode(
        'crossed_flags',
        b'\xf0\x9f\x8e\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'crossed_swords',
        b'\xe2\x9a\x94',
        True,
        None,
        None,
    ),
    Unicode(
        'crown',
        b'\xf0\x9f\x91\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'cruise_ship',
        b'\xf0\x9f\x9b\xb3',
        True,
        None,
        None,
    ),
    Unicode(
        'crying_cat_face',
        b'\xf0\x9f\x98\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'crystal_ball',
        b'\xf0\x9f\x94\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'cucumber',
        b'\xf0\x9f\xa5\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'cupid',
        b'\xf0\x9f\x92\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'curly_loop',
        b'\xe2\x9e\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'currency_exchange',
        b'\xf0\x9f\x92\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'curry',
        b'\xf0\x9f\x8d\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'custard',
        b'\xf0\x9f\x8d\xae',
        False,
        (
            'flan',
            'pudding'
        ),
        None,
    ),
    Unicode(
        'customs',
        b'\xf0\x9f\x9b\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'cyclone',
        b'\xf0\x9f\x8c\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'dagger',
        b'\xf0\x9f\x97\xa1',
        True,
        None,
        None,
    ),
    Unicode(
        'dancer',
        b'\xf0\x9f\x92\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'dancer_tone1',
        b'\xf0\x9f\x92\x83\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'dancer_tone2',
        b'\xf0\x9f\x92\x83\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'dancer_tone3',
        b'\xf0\x9f\x92\x83\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'dancer_tone4',
        b'\xf0\x9f\x92\x83\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'dancer_tone5',
        b'\xf0\x9f\x92\x83\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'dancers',
        b'\xf0\x9f\x91\xaf',
        False,
        (
            'people_with_bunny_ears_partying',
        ),
        None,
    ),
    Unicode(
        'dango',
        b'\xf0\x9f\x8d\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'dark_sunglasses',
        b'\xf0\x9f\x95\xb6',
        True,
        None,
        None,
    ),
    Unicode(
        'dart',
        b'\xf0\x9f\x8e\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'dash',
        b'\xf0\x9f\x92\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'date',
        b'\xf0\x9f\x93\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'deciduous_tree',
        b'\xf0\x9f\x8c\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'deer',
        b'\xf0\x9f\xa6\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'department_store',
        b'\xf0\x9f\x8f\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'derelict_house_building',
        b'\xf0\x9f\x8f\x9a',
        True,
        None,
        None,
    ),
    Unicode(
        'desert',
        b'\xf0\x9f\x8f\x9c',
        True,
        None,
        None,
    ),
    Unicode(
        'desert_island',
        b'\xf0\x9f\x8f\x9d',
        True,
        None,
        None,
    ),
    Unicode(
        'desktop',
        b'\xf0\x9f\x96\xa5',
        True,
        None,
        None,
    ),
    Unicode(
        'diamond_shape_with_a_dot_inside',
        b'\xf0\x9f\x92\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'diamonds',
        b'\xe2\x99\xa6',
        True,
        None,
        None,
    ),
    Unicode(
        'disappointed',
        b'\xf0\x9f\x98\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'disappointed_relieved',
        b'\xf0\x9f\x98\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'dizzy',
        b'\xf0\x9f\x92\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'dizzy_face',
        b'\xf0\x9f\x98\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'do_not_litter',
        b'\xf0\x9f\x9a\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'dog2',
        b'\xf0\x9f\x90\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'dog',
        b'\xf0\x9f\x90\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'dollar',
        b'\xf0\x9f\x92\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'dolls',
        b'\xf0\x9f\x8e\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'dolphin',
        b'\xf0\x9f\x90\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'door',
        b'\xf0\x9f\x9a\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'double_vertical_bar',
        b'\xe2\x8f\xb8',
        True,
        None,
        None,
    ),
    Unicode(
        'doughnut',
        b'\xf0\x9f\x8d\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'dove',
        b'\xf0\x9f\x95\x8a',
        True,
        None,
        None,
    ),
    Unicode(
        'dragon',
        b'\xf0\x9f\x90\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'dragon_face',
        b'\xf0\x9f\x90\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'dress',
        b'\xf0\x9f\x91\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'dromedary_camel',
        b'\xf0\x9f\x90\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'drool',
        b'\xf0\x9f\xa4\xa4',
        False,
        (
            'drooling_face',
        ),
        None,
    ),
    Unicode(
        'droplet',
        b'\xf0\x9f\x92\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'drum',
        b'\xf0\x9f\xa5\x81',
        False,
        (
            'drum_with_drumsticks',
        ),
        None,
    ),
    Unicode(
        'duck',
        b'\xf0\x9f\xa6\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'dvd',
        b'\xf0\x9f\x93\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'e_mail',
        b'\xf0\x9f\x93\xa7',
        False,
        (
            'email',
        ),
        None,
    ),
    Unicode(
        'eagle',
        b'\xf0\x9f\xa6\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'ear',
        b'\xf0\x9f\x91\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_tone1',
        b'\xf0\x9f\x91\x82\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_tone2',
        b'\xf0\x9f\x91\x82\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_tone3',
        b'\xf0\x9f\x91\x82\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_tone4',
        b'\xf0\x9f\x91\x82\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_tone5',
        b'\xf0\x9f\x91\x82\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_of_rice',
        b'\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'earth_africa',
        b'\xf0\x9f\x8c\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'earth_americas',
        b'\xf0\x9f\x8c\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'earth_asia',
        b'\xf0\x9f\x8c\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'egg',
        b'\xf0\x9f\xa5\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'eggplant',
        b'\xf0\x9f\x8d\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'eight',
        b'8\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'eight_pointed_black_star',
        b'\xe2\x9c\xb4',
        True,
        None,
        None,
    ),
    Unicode(
        'eight_spoked_asterisk',
        b'\xe2\x9c\xb3',
        True,
        None,
        None,
    ),
    Unicode(
        'eject',
        b'\xe2\x8f\x8f',
        True,
        None,
        None,
    ),
    Unicode(
        'electric_plug',
        b'\xf0\x9f\x94\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'elephant',
        b'\xf0\x9f\x90\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'end',
        b'\xf0\x9f\x94\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'envelope',
        b'\xe2\x9c\x89',
        True,
        None,
        None,
    ),
    Unicode(
        'envelope_with_arrow',
        b'\xf0\x9f\x93\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'euro',
        b'\xf0\x9f\x92\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'european_castle',
        b'\xf0\x9f\x8f\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'european_post_office',
        b'\xf0\x9f\x8f\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'evergreen_tree',
        b'\xf0\x9f\x8c\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'exclamation',
        b'\xe2\x9d\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'expecting_woman',
        b'\xf0\x9f\xa4\xb0',
        False,
        (
            'pregnant_woman',
        ),
        None,
    ),
    Unicode(
        'expecting_woman_tone1',
        b'\xf0\x9f\xa4\xb0\xf0\x9f\x8f\xbb',
        False,
        (
            'pregnant_woman_tone1',
        ),
        None,
    ),
    Unicode(
        'expecting_woman_tone2',
        b'\xf0\x9f\xa4\xb0\xf0\x9f\x8f\xbc',
        False,
        (
            'pregnant_woman_tone2',
        ),
        None,
    ),
    Unicode(
        'expecting_woman_tone3',
        b'\xf0\x9f\xa4\xb0\xf0\x9f\x8f\xbd',
        False,
        (
            'pregnant_woman_tone3',
        ),
        None,
    ),
    Unicode(
        'expecting_woman_tone4',
        b'\xf0\x9f\xa4\xb0\xf0\x9f\x8f\xbe',
        False,
        (
            'pregnant_woman_tone4',
        ),
        None,
    ),
    Unicode(
        'expecting_woman_tone5',
        b'\xf0\x9f\xa4\xb0\xf0\x9f\x8f\xbf',
        False,
        (
            'pregnant_woman_tone5',
        ),
        None,
    ),
    Unicode(
        'expressionless',
        b'\xf0\x9f\x98\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'eye',
        b'\xf0\x9f\x91\x81',
        True,
        None,
        None,
    ),
    Unicode(
        'eye_in_speech_bubble',
        b'\xf0\x9f\x91\x81\xe2\x80\x8d\xf0\x9f\x97\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'eyeglasses',
        b'\xf0\x9f\x91\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'eyes',
        b'\xf0\x9f\x91\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'face_palm',
        b'\xf0\x9f\xa4\xa6',
        False,
        (
            'facepalm',
            'person_facepalming'
        ),
        None,
    ),
    Unicode(
        'face_palm_tone1',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbb',
        False,
        (
            'facepalm_tone1',
            'person_facepalming_tone1'
        ),
        None,
    ),
    Unicode(
        'face_palm_tone2',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbc',
        False,
        (
            'facepalm_tone2',
            'person_facepalming_tone2'
        ),
        None,
    ),
    Unicode(
        'face_palm_tone3',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbd',
        False,
        (
            'facepalm_tone3',
            'person_facepalming_tone3'
        ),
        None,
    ),
    Unicode(
        'face_palm_tone4',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbe',
        False,
        (
            'facepalm_tone4',
            'person_facepalming_tone4'
        ),
        None,
    ),
    Unicode(
        'face_palm_tone5',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbf',
        False,
        (
            'facepalm_tone5',
            'person_facepalming_tone5'
        ),
        None,
    ),
    Unicode(
        'face_with_head_bandage',
        b'\xf0\x9f\xa4\x95',
        False,
        (
            'head_bandage',
        ),
        None,
    ),
    Unicode(
        'face_with_rolling_eyes',
        b'\xf0\x9f\x99\x84',
        False,
        (
            'rolling_eyes',
        ),
        None,
    ),
    Unicode(
        'face_with_thermometer',
        b'\xf0\x9f\xa4\x92',
        False,
        (
            'thermometer_face',
        ),
        None,
    ),
    Unicode(
        'factory',
        b'\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'fallen_leaf',
        b'\xf0\x9f\x8d\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'family',
        b'\xf0\x9f\x91\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'family_mmb',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'family_mmbb',
        (
            b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa8\xe2\x80'
            b'\x8d\xf0\x9f\x91\xa6\xe2\x80\x8d\xf0\x9f\x91\xa6'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'family_mmg',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'family_mmgb',
        (
            b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa8\xe2\x80'
            b'\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa6'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'family_mmgg',
        (
            b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa8\xe2\x80'
            b'\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa7'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'family_mwbb',
        (
            b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80'
            b'\x8d\xf0\x9f\x91\xa6\xe2\x80\x8d\xf0\x9f\x91\xa6'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'family_mwg',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'family_mwgb',
        (
            b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80'
            b'\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa6'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'family_mwgg',
        (
            b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80'
            b'\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa7'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'family_wwb',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'family_wwbb',
        (
            b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80'
            b'\x8d\xf0\x9f\x91\xa6\xe2\x80\x8d\xf0\x9f\x91\xa6'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'family_wwg',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'family_wwgb',
        (
            b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80'
            b'\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa6'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'family_wwgg',
        (
            b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80'
            b'\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa7'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'fast_forward',
        b'\xe2\x8f\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'fax',
        b'\xf0\x9f\x93\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'fearful',
        b'\xf0\x9f\x98\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'feet',
        b'\xf0\x9f\x90\xbe',
        False,
        (
            'paw_prints',
        ),
        None,
    ),
    Unicode(
        'fencer',
        b'\xf0\x9f\xa4\xba',
        False,
        (
            'fencing',
            'person_fencing'
        ),
        None,
    ),
    Unicode(
        'ferris_wheel',
        b'\xf0\x9f\x8e\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'ferry',
        b'\xe2\x9b\xb4',
        True,
        None,
        None,
    ),
    Unicode(
        'field_hockey',
        b'\xf0\x9f\x8f\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'file_cabinet',
        b'\xf0\x9f\x97\x84',
        True,
        None,
        None,
    ),
    Unicode(
        'file_folder',
        b'\xf0\x9f\x93\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'film_frames',
        b'\xf0\x9f\x8e\x9e',
        True,
        None,
        None,
    ),
    Unicode(
        'film_projector',
        b'\xf0\x9f\x93\xbd',
        True,
        None,
        None,
    ),
    Unicode(
        'fingers_crossed',
        b'\xf0\x9f\xa4\x9e',
        False,
        (
            'hand_with_index_and_middle_finger_crossed',
        ),
        None,
    ),
    Unicode(
        'fingers_crossed_tone1',
        b'\xf0\x9f\xa4\x9e\xf0\x9f\x8f\xbb',
        False,
        (
            'hand_with_index_and_middle_finger_crossed_tone1',
            'hand_with_index_and_middle_fingers_crossed_tone1'
        ),
        None,
    ),
    Unicode(
        'fingers_crossed_tone2',
        b'\xf0\x9f\xa4\x9e\xf0\x9f\x8f\xbc',
        False,
        (
            'hand_with_index_and_middle_finger_crossed_tone2',
            'hand_with_index_and_middle_fingers_crossed_tone2'
        ),
        None,
    ),
    Unicode(
        'fingers_crossed_tone3',
        b'\xf0\x9f\xa4\x9e\xf0\x9f\x8f\xbd',
        False,
        (
            'hand_with_index_and_middle_finger_crossed_tone3',
            'hand_with_index_and_middle_fingers_crossed_tone3'
        ),
        None,
    ),
    Unicode(
        'fingers_crossed_tone4',
        b'\xf0\x9f\xa4\x9e\xf0\x9f\x8f\xbe',
        False,
        (
            'hand_with_index_and_middle_finger_crossed_tone4',
            'hand_with_index_and_middle_fingers_crossed_tone4'
        ),
        None,
    ),
    Unicode(
        'fingers_crossed_tone5',
        b'\xf0\x9f\xa4\x9e\xf0\x9f\x8f\xbf',
        False,
        (
            'hand_with_index_and_middle_finger_crossed_tone5',
            'hand_with_index_and_middle_fingers_crossed_tone5'
        ),
        None,
    ),
    Unicode(
        'fire',
        b'\xf0\x9f\x94\xa5',
        False,
        (
            'flame',
        ),
        None,
    ),
    Unicode(
        'fire_engine',
        b'\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'fireworks',
        b'\xf0\x9f\x8e\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'first_place',
        b'\xf0\x9f\xa5\x87',
        False,
        (
            'first_place_medal',
        ),
        None,
    ),
    Unicode(
        'first_quarter_moon',
        b'\xf0\x9f\x8c\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'first_quarter_moon_with_face',
        b'\xf0\x9f\x8c\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'fish',
        b'\xf0\x9f\x90\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'fish_cake',
        b'\xf0\x9f\x8d\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'fishing_pole_and_fish',
        b'\xf0\x9f\x8e\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'fist',
        b'\xe2\x9c\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'fist_tone1',
        b'\xe2\x9c\x8a\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'fist_tone2',
        b'\xe2\x9c\x8a\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'fist_tone3',
        b'\xe2\x9c\x8a\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'fist_tone4',
        b'\xe2\x9c\x8a\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'fist_tone5',
        b'\xe2\x9c\x8a\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'five',
        b'5\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'flag_ac',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ad',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ae',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_af',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ag',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ai',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_al',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_am',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ao',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_aq',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ar',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_as',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_at',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_au',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_aw',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ax',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_az',
        b'\xf0\x9f\x87\xa6\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ba',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bb',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bd',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_be',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bf',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bg',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bh',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bi',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bj',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bl',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_black',
        b'\xf0\x9f\x8f\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bm',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bn',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bo',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bq',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_br',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bs',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bt',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bv',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bw',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_by',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_bz',
        b'\xf0\x9f\x87\xa7\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ca',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cc',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cd',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cf',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cg',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ch',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ci',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ck',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cl',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cm',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cn',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_co',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cp',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cr',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cu',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cv',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cw',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cx',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cy',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_cz',
        b'\xf0\x9f\x87\xa8\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_de',
        b'\xf0\x9f\x87\xa9\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_dg',
        b'\xf0\x9f\x87\xa9\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_dj',
        b'\xf0\x9f\x87\xa9\xf0\x9f\x87\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_dk',
        b'\xf0\x9f\x87\xa9\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_dm',
        b'\xf0\x9f\x87\xa9\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_do',
        b'\xf0\x9f\x87\xa9\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_dz',
        b'\xf0\x9f\x87\xa9\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ea',
        b'\xf0\x9f\x87\xaa\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ec',
        b'\xf0\x9f\x87\xaa\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ee',
        b'\xf0\x9f\x87\xaa\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_eg',
        b'\xf0\x9f\x87\xaa\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_eh',
        b'\xf0\x9f\x87\xaa\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_er',
        b'\xf0\x9f\x87\xaa\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_es',
        b'\xf0\x9f\x87\xaa\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_et',
        b'\xf0\x9f\x87\xaa\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_eu',
        b'\xf0\x9f\x87\xaa\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_fi',
        b'\xf0\x9f\x87\xab\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_fj',
        b'\xf0\x9f\x87\xab\xf0\x9f\x87\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_fk',
        b'\xf0\x9f\x87\xab\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_fm',
        b'\xf0\x9f\x87\xab\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_fo',
        b'\xf0\x9f\x87\xab\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_fr',
        b'\xf0\x9f\x87\xab\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ga',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gb',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gd',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ge',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gf',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gg',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gh',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gi',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gl',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gm',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gn',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gp',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gq',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gr',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gs',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gt',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gu',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gw',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_gy',
        b'\xf0\x9f\x87\xac\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_hk',
        b'\xf0\x9f\x87\xad\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_hm',
        b'\xf0\x9f\x87\xad\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_hn',
        b'\xf0\x9f\x87\xad\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_hr',
        b'\xf0\x9f\x87\xad\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ht',
        b'\xf0\x9f\x87\xad\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_hu',
        b'\xf0\x9f\x87\xad\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ic',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_id',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ie',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_il',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_im',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_in',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_io',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_iq',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ir',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_is',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_it',
        b'\xf0\x9f\x87\xae\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_je',
        b'\xf0\x9f\x87\xaf\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_jm',
        b'\xf0\x9f\x87\xaf\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_jo',
        b'\xf0\x9f\x87\xaf\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_jp',
        b'\xf0\x9f\x87\xaf\xf0\x9f\x87\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ke',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_kg',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_kh',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ki',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_km',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_kn',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_kp',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_kr',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_kw',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ky',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_kz',
        b'\xf0\x9f\x87\xb0\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_la',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_lb',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_lc',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_li',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_lk',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_lr',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ls',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_lt',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_lu',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_lv',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ly',
        b'\xf0\x9f\x87\xb1\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ma',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mc',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_md',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_me',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mf',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mg',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mh',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mk',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ml',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mm',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mn',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mo',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mp',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mq',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mr',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ms',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mt',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mu',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mv',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mw',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mx',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_my',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_mz',
        b'\xf0\x9f\x87\xb2\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_na',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_nc',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ne',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_nf',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ng',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ni',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_nl',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_no',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_np',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_nr',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_nu',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_nz',
        b'\xf0\x9f\x87\xb3\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_om',
        b'\xf0\x9f\x87\xb4\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pa',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pe',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pf',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pg',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ph',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pk',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pl',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pm',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pn',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pr',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ps',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pt',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_pw',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_py',
        b'\xf0\x9f\x87\xb5\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_qa',
        b'\xf0\x9f\x87\xb6\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_re',
        b'\xf0\x9f\x87\xb7\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ro',
        b'\xf0\x9f\x87\xb7\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_rs',
        b'\xf0\x9f\x87\xb7\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ru',
        b'\xf0\x9f\x87\xb7\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_rw',
        b'\xf0\x9f\x87\xb7\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sa',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sb',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sc',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sd',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_se',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sg',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sh',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_si',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sj',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sk',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sl',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sm',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sn',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_so',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sr',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ss',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_st',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sv',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sx',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sy',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_sz',
        b'\xf0\x9f\x87\xb8\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ta',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tc',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_td',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tf',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tg',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_th',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tj',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tk',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tl',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tm',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tn',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_to',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tr',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tt',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tv',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tw',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_tz',
        b'\xf0\x9f\x87\xb9\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ua',
        b'\xf0\x9f\x87\xba\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ug',
        b'\xf0\x9f\x87\xba\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_um',
        b'\xf0\x9f\x87\xba\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_us',
        b'\xf0\x9f\x87\xba\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_uy',
        b'\xf0\x9f\x87\xba\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_uz',
        b'\xf0\x9f\x87\xba\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_va',
        b'\xf0\x9f\x87\xbb\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_vc',
        b'\xf0\x9f\x87\xbb\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ve',
        b'\xf0\x9f\x87\xbb\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_vg',
        b'\xf0\x9f\x87\xbb\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_vi',
        b'\xf0\x9f\x87\xbb\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_vn',
        b'\xf0\x9f\x87\xbb\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_vu',
        b'\xf0\x9f\x87\xbb\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_wf',
        b'\xf0\x9f\x87\xbc\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_white',
        b'\xf0\x9f\x8f\xb3',
        True,
        None,
        None,
    ),
    Unicode(
        'flag_ws',
        b'\xf0\x9f\x87\xbc\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_xk',
        b'\xf0\x9f\x87\xbd\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_ye',
        b'\xf0\x9f\x87\xbe\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_yt',
        b'\xf0\x9f\x87\xbe\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_za',
        b'\xf0\x9f\x87\xbf\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_zm',
        b'\xf0\x9f\x87\xbf\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_zw',
        b'\xf0\x9f\x87\xbf\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flags',
        b'\xf0\x9f\x8e\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'flashlight',
        b'\xf0\x9f\x94\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'fleur_de_lis',
        b'\xe2\x9a\x9c',
        True,
        None,
        None,
    ),
    Unicode(
        'floppy_disk',
        b'\xf0\x9f\x92\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'flower_playing_cards',
        b'\xf0\x9f\x8e\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'flushed',
        b'\xf0\x9f\x98\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'fog',
        b'\xf0\x9f\x8c\xab',
        True,
        None,
        None,
    ),
    Unicode(
        'foggy',
        b'\xf0\x9f\x8c\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'football',
        b'\xf0\x9f\x8f\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'footprints',
        b'\xf0\x9f\x91\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'fork_and_knife',
        b'\xf0\x9f\x8d\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'fork_and_knife_with_plate',
        b'\xf0\x9f\x8d\xbd',
        True,
        None,
        None,
    ),
    Unicode(
        'fountain',
        b'\xe2\x9b\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'four',
        b'4\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'four_leaf_clover',
        b'\xf0\x9f\x8d\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'fox',
        b'\xf0\x9f\xa6\x8a',
        False,
        (
            'fox_face',
        ),
        None,
    ),
    Unicode(
        'frame_photo',
        b'\xf0\x9f\x96\xbc',
        True,
        None,
        None,
    ),
    Unicode(
        'free',
        b'\xf0\x9f\x86\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'fried_shrimp',
        b'\xf0\x9f\x8d\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'fries',
        b'\xf0\x9f\x8d\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'frog',
        b'\xf0\x9f\x90\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'frowning2',
        b'\xe2\x98\xb9',
        True,
        None,
        None,
    ),
    Unicode(
        'fuelpump',
        b'\xe2\x9b\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'full_moon',
        b'\xf0\x9f\x8c\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'full_moon_with_face',
        b'\xf0\x9f\x8c\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'funeral_urn',
        b'\xe2\x9a\xb1',
        True,
        None,
        None,
    ),
    Unicode(
        'game_die',
        b'\xf0\x9f\x8e\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'gay_pride_flag',
        b'\xf0\x9f\x8f\xb3\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x8c\x88',
        False,
        (
            'rainbow_flag',
        ),
        None,
    ),
    Unicode(
        'gear',
        b'\xe2\x9a\x99',
        True,
        None,
        None,
    ),
    Unicode(
        'gem',
        b'\xf0\x9f\x92\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'gemini',
        b'\xe2\x99\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'ghost',
        b'\xf0\x9f\x91\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'gift',
        b'\xf0\x9f\x8e\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'gift_heart',
        b'\xf0\x9f\x92\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'girl',
        b'\xf0\x9f\x91\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'girl_tone1',
        b'\xf0\x9f\x91\xa7\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'girl_tone2',
        b'\xf0\x9f\x91\xa7\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'girl_tone3',
        b'\xf0\x9f\x91\xa7\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'girl_tone4',
        b'\xf0\x9f\x91\xa7\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'girl_tone5',
        b'\xf0\x9f\x91\xa7\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'glass_of_milk',
        b'\xf0\x9f\xa5\x9b',
        False,
        (
            'milk',
        ),
        None,
    ),
    Unicode(
        'globe_with_meridians',
        b'\xf0\x9f\x8c\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'goal',
        b'\xf0\x9f\xa5\x85',
        False,
        (
            'goal_net',
        ),
        None,
    ),
    Unicode(
        'goat',
        b'\xf0\x9f\x90\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'golf',
        b'\xe2\x9b\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'golfer',
        b'\xf0\x9f\x8f\x8c',
        True,
        None,
        None,
    ),
    Unicode(
        'golfer_tone1',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbb',
        False,
        (
            'person_golfing_tone1',
        ),
        None,
    ),
    Unicode(
        'golfer_tone2',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbc',
        False,
        (
            'person_golfing_tone2',
        ),
        None,
    ),
    Unicode(
        'golfer_tone3',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbd',
        False,
        (
            'person_golfing_tone3',
        ),
        None,
    ),
    Unicode(
        'golfer_tone4',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbe',
        False,
        (
            'person_golfing_tone4',
        ),
        None,
    ),
    Unicode(
        'golfer_tone5',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbf',
        False,
        (
            'person_golfing_tone5',
        ),
        None,
    ),
    Unicode(
        'gorilla',
        b'\xf0\x9f\xa6\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'grandma',
        b'\xf0\x9f\x91\xb5',
        False,
        (
            'older_woman',
        ),
        None,
    ),
    Unicode(
        'grandma_tone1',
        b'\xf0\x9f\x91\xb5\xf0\x9f\x8f\xbb',
        False,
        (
            'older_woman_tone1',
        ),
        None,
    ),
    Unicode(
        'grandma_tone2',
        b'\xf0\x9f\x91\xb5\xf0\x9f\x8f\xbc',
        False,
        (
            'older_woman_tone2',
        ),
        None,
    ),
    Unicode(
        'grandma_tone3',
        b'\xf0\x9f\x91\xb5\xf0\x9f\x8f\xbd',
        False,
        (
            'older_woman_tone3',
        ),
        None,
    ),
    Unicode(
        'grandma_tone4',
        b'\xf0\x9f\x91\xb5\xf0\x9f\x8f\xbe',
        False,
        (
            'older_woman_tone4',
        ),
        None,
    ),
    Unicode(
        'grandma_tone5',
        b'\xf0\x9f\x91\xb5\xf0\x9f\x8f\xbf',
        False,
        (
            'older_woman_tone5',
        ),
        None,
    ),
    Unicode(
        'grapes',
        b'\xf0\x9f\x8d\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'green_apple',
        b'\xf0\x9f\x8d\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'green_book',
        b'\xf0\x9f\x93\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'green_heart',
        b'\xf0\x9f\x92\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'green_salad',
        b'\xf0\x9f\xa5\x97',
        False,
        (
            'salad',
        ),
        None,
    ),
    Unicode(
        'grey_exclamation',
        b'\xe2\x9d\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'grey_question',
        b'\xe2\x9d\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'grimacing',
        b'\xf0\x9f\x98\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'grin',
        b'\xf0\x9f\x98\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'grinning',
        b'\xf0\x9f\x98\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'guardsman',
        b'\xf0\x9f\x92\x82',
        False,
        (
            'guard',
        ),
        None,
    ),
    Unicode(
        'guardsman_tone1',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbb',
        False,
        (
            'guard_tone1',
        ),
        None,
    ),
    Unicode(
        'guardsman_tone2',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbc',
        False,
        (
            'guard_tone2',
        ),
        None,
    ),
    Unicode(
        'guardsman_tone3',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbd',
        False,
        (
            'guard_tone3',
        ),
        None,
    ),
    Unicode(
        'guardsman_tone4',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbe',
        False,
        (
            'guard_tone4',
        ),
        None,
    ),
    Unicode(
        'guardsman_tone5',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbf',
        False,
        (
            'guard_tone5',
        ),
        None,
    ),
    Unicode(
        'guitar',
        b'\xf0\x9f\x8e\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'gun',
        b'\xf0\x9f\x94\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'haircut',
        b'\xf0\x9f\x92\x87',
        False,
        (
            'person_getting_haircut',
        ),
        None,
    ),
    Unicode(
        'haircut_tone1',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbb',
        False,
        (
            'person_getting_haircut_tone1',
        ),
        None,
    ),
    Unicode(
        'haircut_tone2',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbc',
        False,
        (
            'person_getting_haircut_tone2',
        ),
        None,
    ),
    Unicode(
        'haircut_tone3',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbd',
        False,
        (
            'person_getting_haircut_tone3',
        ),
        None,
    ),
    Unicode(
        'haircut_tone4',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbe',
        False,
        (
            'person_getting_haircut_tone4',
        ),
        None,
    ),
    Unicode(
        'haircut_tone5',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbf',
        False,
        (
            'person_getting_haircut_tone5',
        ),
        None,
    ),
    Unicode(
        'hamburger',
        b'\xf0\x9f\x8d\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'hammer',
        b'\xf0\x9f\x94\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'hammer_and_pick',
        b'\xe2\x9a\x92',
        True,
        None,
        None,
    ),
    Unicode(
        'hammer_and_wrench',
        b'\xf0\x9f\x9b\xa0',
        True,
        None,
        None,
    ),
    Unicode(
        'hamster',
        b'\xf0\x9f\x90\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'hand_splayed',
        b'\xf0\x9f\x96\x90',
        True,
        None,
        None,
    ),
    Unicode(
        'hand_splayed_tone1',
        b'\xf0\x9f\x96\x90\xf0\x9f\x8f\xbb',
        False,
        (
            'raised_hand_with_fingers_splayed_tone1',
        ),
        None,
    ),
    Unicode(
        'hand_splayed_tone2',
        b'\xf0\x9f\x96\x90\xf0\x9f\x8f\xbc',
        False,
        (
            'raised_hand_with_fingers_splayed_tone2',
        ),
        None,
    ),
    Unicode(
        'hand_splayed_tone3',
        b'\xf0\x9f\x96\x90\xf0\x9f\x8f\xbd',
        False,
        (
            'raised_hand_with_fingers_splayed_tone3',
        ),
        None,
    ),
    Unicode(
        'hand_splayed_tone4',
        b'\xf0\x9f\x96\x90\xf0\x9f\x8f\xbe',
        False,
        (
            'raised_hand_with_fingers_splayed_tone4',
        ),
        None,
    ),
    Unicode(
        'hand_splayed_tone5',
        b'\xf0\x9f\x96\x90\xf0\x9f\x8f\xbf',
        False,
        (
            'raised_hand_with_fingers_splayed_tone5',
        ),
        None,
    ),
    Unicode(
        'handbag',
        b'\xf0\x9f\x91\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'handball',
        b'\xf0\x9f\xa4\xbe',
        False,
        (
            'person_playing_handball',
        ),
        None,
    ),
    Unicode(
        'handball_tone1',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbb',
        False,
        (
            'person_playing_handball_tone1',
        ),
        None,
    ),
    Unicode(
        'handball_tone2',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbc',
        False,
        (
            'person_playing_handball_tone2',
        ),
        None,
    ),
    Unicode(
        'handball_tone3',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbd',
        False,
        (
            'person_playing_handball_tone3',
        ),
        None,
    ),
    Unicode(
        'handball_tone4',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbe',
        False,
        (
            'person_playing_handball_tone4',
        ),
        None,
    ),
    Unicode(
        'handball_tone5',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbf',
        False,
        (
            'person_playing_handball_tone5',
        ),
        None,
    ),
    Unicode(
        'handshake',
        b'\xf0\x9f\xa4\x9d',
        False,
        (
            'shaking_hands',
        ),
        None,
    ),
    Unicode(
        'hankey',
        b'\xf0\x9f\x92\xa9',
        False,
        (
            'poo',
            'poop',
            'shit'
        ),
        None,
    ),
    Unicode(
        'hash',
        b'#\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'hatched_chick',
        b'\xf0\x9f\x90\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'hatching_chick',
        b'\xf0\x9f\x90\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'headphones',
        b'\xf0\x9f\x8e\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'hear_no_evil',
        b'\xf0\x9f\x99\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'heart',
        b'\xe2\x9d\xa4',
        True,
        None,
        None,
    ),
    Unicode(
        'heart_decoration',
        b'\xf0\x9f\x92\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'heart_exclamation',
        b'\xe2\x9d\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'heart_eyes',
        b'\xf0\x9f\x98\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'heart_eyes_cat',
        b'\xf0\x9f\x98\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'heartbeat',
        b'\xf0\x9f\x92\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'heartpulse',
        b'\xf0\x9f\x92\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'hearts',
        b'\xe2\x99\xa5',
        True,
        None,
        None,
    ),
    Unicode(
        'heavy_check_mark',
        b'\xe2\x9c\x94',
        True,
        None,
        None,
    ),
    Unicode(
        'heavy_division_sign',
        b'\xe2\x9e\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'heavy_dollar_sign',
        b'\xf0\x9f\x92\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'heavy_minus_sign',
        b'\xe2\x9e\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'heavy_multiplication_x',
        b'\xe2\x9c\x96',
        True,
        None,
        None,
    ),
    Unicode(
        'heavy_plus_sign',
        b'\xe2\x9e\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'helicopter',
        b'\xf0\x9f\x9a\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'helmet_with_cross',
        b'\xe2\x9b\x91',
        True,
        None,
        None,
    ),
    Unicode(
        'herb',
        b'\xf0\x9f\x8c\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'hibiscus',
        b'\xf0\x9f\x8c\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'high_brightness',
        b'\xf0\x9f\x94\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'high_heel',
        b'\xf0\x9f\x91\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'hockey',
        b'\xf0\x9f\x8f\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'hole',
        b'\xf0\x9f\x95\xb3',
        True,
        None,
        None,
    ),
    Unicode(
        'homes',
        b'\xf0\x9f\x8f\x98',
        True,
        None,
        None,
    ),
    Unicode(
        'honey_pot',
        b'\xf0\x9f\x8d\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'horse',
        b'\xf0\x9f\x90\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'horse_racing',
        b'\xf0\x9f\x8f\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'horse_racing_tone1',
        b'\xf0\x9f\x8f\x87\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'horse_racing_tone2',
        b'\xf0\x9f\x8f\x87\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'horse_racing_tone3',
        b'\xf0\x9f\x8f\x87\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'horse_racing_tone4',
        b'\xf0\x9f\x8f\x87\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'horse_racing_tone5',
        b'\xf0\x9f\x8f\x87\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'hospital',
        b'\xf0\x9f\x8f\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'hot_dog',
        b'\xf0\x9f\x8c\xad',
        False,
        (
            'hotdog',
        ),
        None,
    ),
    Unicode(
        'hot_pepper',
        b'\xf0\x9f\x8c\xb6',
        True,
        None,
        None,
    ),
    Unicode(
        'hotel',
        b'\xf0\x9f\x8f\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'hotsprings',
        b'\xe2\x99\xa8',
        True,
        None,
        None,
    ),
    Unicode(
        'hourglass',
        b'\xe2\x8c\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'hourglass_flowing_sand',
        b'\xe2\x8f\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'house',
        b'\xf0\x9f\x8f\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'house_with_garden',
        b'\xf0\x9f\x8f\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'hugging',
        b'\xf0\x9f\xa4\x97',
        False,
        (
            'hugging_face',
        ),
        None,
    ),
    Unicode(
        'hushed',
        b'\xf0\x9f\x98\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'ice_cream',
        b'\xf0\x9f\x8d\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'ice_skate',
        b'\xe2\x9b\xb8',
        True,
        None,
        None,
    ),
    Unicode(
        'icecream',
        b'\xf0\x9f\x8d\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'id',
        b'\xf0\x9f\x86\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'ideograph_advantage',
        b'\xf0\x9f\x89\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'imp',
        b'\xf0\x9f\x91\xbf',
        False,
        None,
        (']:(', ']:-(', ']=(', ']=-('),
    ),
    Unicode(
        'inbox_tray',
        b'\xf0\x9f\x93\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'incoming_envelope',
        b'\xf0\x9f\x93\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'information_desk_person',
        b'\xf0\x9f\x92\x81',
        False,
        (
            'person_tipping_hand',
        ),
        None,
    ),
    Unicode(
        'information_desk_person_tone1',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbb',
        False,
        (
            'person_tipping_hand_tone1',
        ),
        None,
    ),
    Unicode(
        'information_desk_person_tone2',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbc',
        False,
        (
            'person_tipping_hand_tone2',
        ),
        None,
    ),
    Unicode(
        'information_desk_person_tone3',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbd',
        False,
        (
            'person_tipping_hand_tone3',
        ),
        None,
    ),
    Unicode(
        'information_desk_person_tone4',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbe',
        False,
        (
            'person_tipping_hand_tone4',
        ),
        None,
    ),
    Unicode(
        'information_desk_person_tone5',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbf',
        False,
        (
            'person_tipping_hand_tone5',
        ),
        None,
    ),
    Unicode(
        'information_source',
        b'\xe2\x84\xb9',
        True,
        None,
        None,
    ),
    Unicode(
        'interrobang',
        b'\xe2\x81\x89',
        True,
        None,
        None,
    ),
    Unicode(
        'iphone',
        b'\xf0\x9f\x93\xb1',
        False,
        (
            'mobile_phone',
        ),
        None,
    ),
    Unicode(
        'izakaya_lantern',
        b'\xf0\x9f\x8f\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'jack_o_lantern',
        b'\xf0\x9f\x8e\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'japan',
        b'\xf0\x9f\x97\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'japanese_castle',
        b'\xf0\x9f\x8f\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'japanese_goblin',
        b'\xf0\x9f\x91\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'japanese_ogre',
        b'\xf0\x9f\x91\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'jeans',
        b'\xf0\x9f\x91\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'joy_cat',
        b'\xf0\x9f\x98\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'joystick',
        b'\xf0\x9f\x95\xb9',
        True,
        None,
        None,
    ),
    Unicode(
        'juggler',
        b'\xf0\x9f\xa4\xb9',
        False,
        (
            'juggling',
            'person_juggling'
        ),
        None,
    ),
    Unicode(
        'juggler_tone1',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbb',
        False,
        (
            'juggling_tone1',
            'person_juggling_tone1'
        ),
        None,
    ),
    Unicode(
        'juggler_tone2',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbc',
        False,
        (
            'juggling_tone2',
            'person_juggling_tone2'
        ),
        None,
    ),
    Unicode(
        'juggler_tone3',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbd',
        False,
        (
            'juggling_tone3',
            'person_juggling_tone3'
        ),
        None,
    ),
    Unicode(
        'juggler_tone4',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbe',
        False,
        (
            'juggling_tone4',
            'person_juggling_tone4'
        ),
        None,
    ),
    Unicode(
        'juggler_tone5',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbf',
        False,
        (
            'juggling_tone5',
            'person_juggling_tone5'
        ),
        None,
    ),
    Unicode(
        'kaaba',
        b'\xf0\x9f\x95\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'karate_uniform',
        b'\xf0\x9f\xa5\x8b',
        False,
        (
            'martial_arts_uniform',
        ),
        None,
    ),
    Unicode(
        'key2',
        b'\xf0\x9f\x97\x9d',
        True,
        None,
        None,
    ),
    Unicode(
        'key',
        b'\xf0\x9f\x94\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'keyboard',
        b'\xe2\x8c\xa8',
        True,
        None,
        None,
    ),
    Unicode(
        'keycap_ten',
        b'\xf0\x9f\x94\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'kimono',
        b'\xf0\x9f\x91\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'kiss',
        b'\xf0\x9f\x92\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'kissing_cat',
        b'\xf0\x9f\x98\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'kissing_closed_eyes',
        b'\xf0\x9f\x98\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'kissing_heart',
        b'\xf0\x9f\x98\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'kissing_smiling_eyes',
        b'\xf0\x9f\x98\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'kiwi',
        b'\xf0\x9f\xa5\x9d',
        False,
        (
            'kiwifruit',
        ),
        None,
    ),
    Unicode(
        'knife',
        b'\xf0\x9f\x94\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'koala',
        b'\xf0\x9f\x90\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'koko',
        b'\xf0\x9f\x88\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'label',
        b'\xf0\x9f\x8f\xb7',
        True,
        None,
        None,
    ),
    Unicode(
        'large_blue_circle',
        b'\xf0\x9f\x94\xb5',
        False,
        (
            'blue_circle',
        ),
        None,
    ),
    Unicode(
        'large_blue_diamond',
        b'\xf0\x9f\x94\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'large_orange_diamond',
        b'\xf0\x9f\x94\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'last_quarter_moon',
        b'\xf0\x9f\x8c\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'last_quarter_moon_with_face',
        b'\xf0\x9f\x8c\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'laughing',
        b'\xf0\x9f\x98\x86',
        False,
        (
            'satisfied',
        ),
        ('x-)', 'X-)'),
    ),
    Unicode(
        'leaves',
        b'\xf0\x9f\x8d\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'ledger',
        b'\xf0\x9f\x93\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'left_facing_fist',
        b'\xf0\x9f\xa4\x9b',
        False,
        (
            'left_fist',
        ),
        None,
    ),
    Unicode(
        'left_facing_fist_tone1',
        b'\xf0\x9f\xa4\x9b\xf0\x9f\x8f\xbb',
        False,
        (
            'left_fist_tone1',
        ),
        None,
    ),
    Unicode(
        'left_facing_fist_tone2',
        b'\xf0\x9f\xa4\x9b\xf0\x9f\x8f\xbc',
        False,
        (
            'left_fist_tone2',
        ),
        None,
    ),
    Unicode(
        'left_facing_fist_tone3',
        b'\xf0\x9f\xa4\x9b\xf0\x9f\x8f\xbd',
        False,
        (
            'left_fist_tone3',
        ),
        None,
    ),
    Unicode(
        'left_facing_fist_tone4',
        b'\xf0\x9f\xa4\x9b\xf0\x9f\x8f\xbe',
        False,
        (
            'left_fist_tone4',
        ),
        None,
    ),
    Unicode(
        'left_facing_fist_tone5',
        b'\xf0\x9f\xa4\x9b\xf0\x9f\x8f\xbf',
        False,
        (
            'left_fist_tone5',
        ),
        None,
    ),
    Unicode(
        'left_luggage',
        b'\xf0\x9f\x9b\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'left_right_arrow',
        b'\xe2\x86\x94',
        True,
        None,
        None,
    ),
    Unicode(
        'left_speech_bubble',
        b'\xf0\x9f\x97\xa8',
        True,
        None,
        None,
    ),
    Unicode(
        'leftwards_arrow_with_hook',
        b'\xe2\x86\xa9',
        True,
        None,
        None,
    ),
    Unicode(
        'lemon',
        b'\xf0\x9f\x8d\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'leo',
        b'\xe2\x99\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'leopard',
        b'\xf0\x9f\x90\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'level_slider',
        b'\xf0\x9f\x8e\x9a',
        True,
        None,
        None,
    ),
    Unicode(
        'levitate',
        b'\xf0\x9f\x95\xb4',
        True,
        None,
        None,
    ),
    Unicode(
        'levitate_tone1',
        b'\xf0\x9f\x95\xb4\xf0\x9f\x8f\xbb',
        False,
        (
            'man_in_business_suit_levitating_tone1',
        ),
        None,
    ),
    Unicode(
        'levitate_tone2',
        b'\xf0\x9f\x95\xb4\xf0\x9f\x8f\xbc',
        False,
        (
            'man_in_business_suit_levitating_tone2',
        ),
        None,
    ),
    Unicode(
        'levitate_tone3',
        b'\xf0\x9f\x95\xb4\xf0\x9f\x8f\xbd',
        False,
        (
            'man_in_business_suit_levitating_tone3',
        ),
        None,
    ),
    Unicode(
        'levitate_tone4',
        b'\xf0\x9f\x95\xb4\xf0\x9f\x8f\xbe',
        False,
        (
            'man_in_business_suit_levitating_tone4',
        ),
        None,
    ),
    Unicode(
        'levitate_tone5',
        b'\xf0\x9f\x95\xb4\xf0\x9f\x8f\xbf',
        False,
        (
            'man_in_business_suit_levitating_tone5',
        ),
        None,
    ),
    Unicode(
        'liar',
        b'\xf0\x9f\xa4\xa5',
        False,
        (
            'lying_face',
        ),
        None,
    ),
    Unicode(
        'libra',
        b'\xe2\x99\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'lifter',
        b'\xf0\x9f\x8f\x8b',
        True,
        None,
        None,
    ),
    Unicode(
        'lifter_tone1',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbb',
        False,
        (
            'weight_lifter_tone1',
            'person_lifting_weights_tone1'
        ),
        None,
    ),
    Unicode(
        'lifter_tone2',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbc',
        False,
        (
            'weight_lifter_tone2',
            'person_lifting_weights_tone2'
        ),
        None,
    ),
    Unicode(
        'lifter_tone3',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbd',
        False,
        (
            'weight_lifter_tone3',
            'person_lifting_weights_tone3'
        ),
        None,
    ),
    Unicode(
        'lifter_tone4',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbe',
        False,
        (
            'weight_lifter_tone4',
            'person_lifting_weights_tone4'
        ),
        None,
    ),
    Unicode(
        'lifter_tone5',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbf',
        False,
        (
            'weight_lifter_tone5',
            'person_lifting_weights_tone5'
        ),
        None,
    ),
    Unicode(
        'light_rail',
        b'\xf0\x9f\x9a\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'link',
        b'\xf0\x9f\x94\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'linked_paperclips',
        b'\xf0\x9f\x96\x87',
        True,
        None,
        None,
    ),
    Unicode(
        'lion',
        b'\xf0\x9f\xa6\x81',
        False,
        (
            'lion_face',
        ),
        None,
    ),
    Unicode(
        'lips',
        b'\xf0\x9f\x91\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'lipstick',
        b'\xf0\x9f\x92\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'lizard',
        b'\xf0\x9f\xa6\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'lock',
        b'\xf0\x9f\x94\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'lock_with_ink_pen',
        b'\xf0\x9f\x94\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'lollipop',
        b'\xf0\x9f\x8d\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'loop',
        b'\xe2\x9e\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'loud_sound',
        b'\xf0\x9f\x94\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'loudspeaker',
        b'\xf0\x9f\x93\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'love_hotel',
        b'\xf0\x9f\x8f\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'love_letter',
        b'\xf0\x9f\x92\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'low_brightness',
        b'\xf0\x9f\x94\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'lower_left_ballpoint_pen',
        b'\xf0\x9f\x96\x8a',
        True,
        None,
        None,
    ),
    Unicode(
        'lower_left_fountain_pen',
        b'\xf0\x9f\x96\x8b',
        True,
        None,
        None,
    ),
    Unicode(
        'lower_left_paintbrush',
        b'\xf0\x9f\x96\x8c',
        True,
        None,
        None,
    ),
    Unicode(
        'm',
        b'\xe2\x93\x82',
        True,
        None,
        None,
    ),
    Unicode(
        'mag',
        b'\xf0\x9f\x94\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'mag_right',
        b'\xf0\x9f\x94\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'mahjong',
        b'\xf0\x9f\x80\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'mailbox',
        b'\xf0\x9f\x93\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'mailbox_closed',
        b'\xf0\x9f\x93\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'mailbox_with_mail',
        b'\xf0\x9f\x93\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'mailbox_with_no_mail',
        b'\xf0\x9f\x93\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'male_dancer',
        b'\xf0\x9f\x95\xba',
        False,
        (
            'man_dancing',
        ),
        None,
    ),
    Unicode(
        'male_dancer_tone1',
        b'\xf0\x9f\x95\xba\xf0\x9f\x8f\xbb',
        False,
        (
            'man_dancing_tone1',
        ),
        None,
    ),
    Unicode(
        'male_dancer_tone2',
        b'\xf0\x9f\x95\xba\xf0\x9f\x8f\xbc',
        False,
        (
            'man_dancing_tone2',
        ),
        None,
    ),
    Unicode(
        'male_dancer_tone3',
        b'\xf0\x9f\x95\xba\xf0\x9f\x8f\xbd',
        False,
        (
            'man_dancing_tone3',
        ),
        None,
    ),
    Unicode(
        'male_dancer_tone4',
        b'\xf0\x9f\x95\xba\xf0\x9f\x8f\xbe',
        False,
        (
            'man_dancing_tone4',
        ),
        None,
    ),
    Unicode(
        'male_dancer_tone5',
        b'\xf0\x9f\x95\xba\xf0\x9f\x8f\xbf',
        False,
        (
            'man_dancing_tone5',
        ),
        None,
    ),
    Unicode(
        'man',
        b'\xf0\x9f\x91\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo',
        b'\xf0\x9f\xa4\xb5',
        True,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone1',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbb',
        True,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone2',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbc',
        True,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone3',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbd',
        True,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone4',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbe',
        True,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone5',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbf',
        True,
        None,
        None,
    ),
    Unicode(
        'man_with_gua_pi_mao',
        b'\xf0\x9f\x91\xb2',
        False,
        (
            'man_with_chinese_cap',
        ),
        None,
    ),
    Unicode(
        'man_with_gua_pi_mao_tone1',
        b'\xf0\x9f\x91\xb2\xf0\x9f\x8f\xbb',
        False,
        (
            'man_with_chinese_cap_tone1',
        ),
        None,
    ),
    Unicode(
        'man_with_gua_pi_mao_tone2',
        b'\xf0\x9f\x91\xb2\xf0\x9f\x8f\xbc',
        False,
        (
            'man_with_chinese_cap_tone2',
        ),
        None,
    ),
    Unicode(
        'man_with_gua_pi_mao_tone3',
        b'\xf0\x9f\x91\xb2\xf0\x9f\x8f\xbd',
        False,
        (
            'man_with_chinese_cap_tone3',
        ),
        None,
    ),
    Unicode(
        'man_with_gua_pi_mao_tone4',
        b'\xf0\x9f\x91\xb2\xf0\x9f\x8f\xbe',
        False,
        (
            'man_with_chinese_cap_tone4',
        ),
        None,
    ),
    Unicode(
        'man_with_gua_pi_mao_tone5',
        b'\xf0\x9f\x91\xb2\xf0\x9f\x8f\xbf',
        False,
        (
            'man_with_chinese_cap_tone5',
        ),
        None,
    ),
    Unicode(
        'man_with_turban',
        b'\xf0\x9f\x91\xb3',
        False,
        (
            'person_wearing_turban',
        ),
        None,
    ),
    Unicode(
        'man_with_turban_tone1',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbb',
        False,
        (
            'person_wearing_turban_tone1',
        ),
        None,
    ),
    Unicode(
        'man_with_turban_tone2',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbc',
        False,
        (
            'person_wearing_turban_tone2',
        ),
        None,
    ),
    Unicode(
        'man_with_turban_tone3',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbd',
        False,
        (
            'person_wearing_turban_tone3',
        ),
        None,
    ),
    Unicode(
        'man_with_turban_tone4',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbe',
        False,
        (
            'person_wearing_turban_tone4',
        ),
        None,
    ),
    Unicode(
        'man_with_turban_tone5',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbf',
        False,
        (
            'person_wearing_turban_tone5',
        ),
        None,
    ),
    Unicode(
        'mans_shoe',
        b'\xf0\x9f\x91\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'map',
        b'\xf0\x9f\x97\xba',
        True,
        None,
        None,
    ),
    Unicode(
        'maple_leaf',
        b'\xf0\x9f\x8d\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'mask',
        b'\xf0\x9f\x98\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'massage',
        b'\xf0\x9f\x92\x86',
        False,
        (
            'person_getting_massage',
        ),
        None,
    ),
    Unicode(
        'massage_tone1',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbb',
        False,
        (
            'person_getting_massage_tone1',
        ),
        None,
    ),
    Unicode(
        'massage_tone2',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbc',
        False,
        (
            'person_getting_massage_tone2',
        ),
        None,
    ),
    Unicode(
        'massage_tone3',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbd',
        False,
        (
            'person_getting_massage_tone3',
        ),
        None,
    ),
    Unicode(
        'massage_tone4',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbe',
        False,
        (
            'person_getting_massage_tone4',
        ),
        None,
    ),
    Unicode(
        'massage_tone5',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbf',
        False,
        (
            'person_getting_massage_tone5',
        ),
        None,
    ),
    Unicode(
        'meat_on_bone',
        b'\xf0\x9f\x8d\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'medal',
        b'\xf0\x9f\x8f\x85',
        False,
        (
            'sports_medal',
        ),
        None,
    ),
    Unicode(
        'mega',
        b'\xf0\x9f\x93\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'melon',
        b'\xf0\x9f\x8d\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'menorah',
        b'\xf0\x9f\x95\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'mens',
        b'\xf0\x9f\x9a\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'metal',
        b'\xf0\x9f\xa4\x98',
        False,
        (
            'sign_of_the_horns',
        ),
        None,
    ),
    Unicode(
        'metal_tone1',
        b'\xf0\x9f\xa4\x98\xf0\x9f\x8f\xbb',
        False,
        (
            'sign_of_the_horns_tone1',
        ),
        None,
    ),
    Unicode(
        'metal_tone2',
        b'\xf0\x9f\xa4\x98\xf0\x9f\x8f\xbc',
        False,
        (
            'sign_of_the_horns_tone2',
        ),
        None,
    ),
    Unicode(
        'metal_tone3',
        b'\xf0\x9f\xa4\x98\xf0\x9f\x8f\xbd',
        False,
        (
            'sign_of_the_horns_tone3',
        ),
        None,
    ),
    Unicode(
        'metal_tone4',
        b'\xf0\x9f\xa4\x98\xf0\x9f\x8f\xbe',
        False,
        (
            'sign_of_the_horns_tone4',
        ),
        None,
    ),
    Unicode(
        'metal_tone5',
        b'\xf0\x9f\xa4\x98\xf0\x9f\x8f\xbf',
        False,
        (
            'sign_of_the_horns_tone5',
        ),
        None,
    ),
    Unicode(
        'metro',
        b'\xf0\x9f\x9a\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'microphone2',
        b'\xf0\x9f\x8e\x99',
        True,
        None,
        None,
    ),
    Unicode(
        'microphone',
        b'\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'microscope',
        b'\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'middle_finger',
        b'\xf0\x9f\x96\x95',
        False,
        (
            'reversed_hand_with_middle_finger_extended',
        ),
        None,
    ),
    Unicode(
        'middle_finger_tone1',
        b'\xf0\x9f\x96\x95\xf0\x9f\x8f\xbb',
        False,
        (
            'reversed_hand_with_middle_finger_extended_tone1',
        ),
        None,
    ),
    Unicode(
        'middle_finger_tone2',
        b'\xf0\x9f\x96\x95\xf0\x9f\x8f\xbc',
        False,
        (
            'reversed_hand_with_middle_finger_extended_tone2',
        ),
        None,
    ),
    Unicode(
        'middle_finger_tone3',
        b'\xf0\x9f\x96\x95\xf0\x9f\x8f\xbd',
        False,
        (
            'reversed_hand_with_middle_finger_extended_tone3',
        ),
        None,
    ),
    Unicode(
        'middle_finger_tone4',
        b'\xf0\x9f\x96\x95\xf0\x9f\x8f\xbe',
        False,
        (
            'reversed_hand_with_middle_finger_extended_tone4',
        ),
        None,
    ),
    Unicode(
        'middle_finger_tone5',
        b'\xf0\x9f\x96\x95\xf0\x9f\x8f\xbf',
        False,
        (
            'reversed_hand_with_middle_finger_extended_tone5',
        ),
        None,
    ),
    Unicode(
        'military_medal',
        b'\xf0\x9f\x8e\x96',
        True,
        None,
        None,
    ),
    Unicode(
        'milky_way',
        b'\xf0\x9f\x8c\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'minibus',
        b'\xf0\x9f\x9a\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'minidisc',
        b'\xf0\x9f\x92\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'mobile_phone_off',
        b'\xf0\x9f\x93\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'money_mouth',
        b'\xf0\x9f\xa4\x91',
        False,
        (
            'money_mouth_face',
        ),
        None,
    ),
    Unicode(
        'money_with_wings',
        b'\xf0\x9f\x92\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'moneybag',
        b'\xf0\x9f\x92\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'monkey',
        b'\xf0\x9f\x90\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'monkey_face',
        b'\xf0\x9f\x90\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'monorail',
        b'\xf0\x9f\x9a\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'mortar_board',
        b'\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'mosque',
        b'\xf0\x9f\x95\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'mother_christmas',
        b'\xf0\x9f\xa4\xb6',
        False,
        (
            'mrs_claus',
        ),
        None,
    ),
    Unicode(
        'mother_christmas_tone1',
        b'\xf0\x9f\xa4\xb6\xf0\x9f\x8f\xbb',
        False,
        (
            'mrs_claus_tone1',
        ),
        None,
    ),
    Unicode(
        'mother_christmas_tone2',
        b'\xf0\x9f\xa4\xb6\xf0\x9f\x8f\xbc',
        False,
        (
            'mrs_claus_tone2',
        ),
        None,
    ),
    Unicode(
        'mother_christmas_tone3',
        b'\xf0\x9f\xa4\xb6\xf0\x9f\x8f\xbd',
        False,
        (
            'mrs_claus_tone3',
        ),
        None,
    ),
    Unicode(
        'mother_christmas_tone4',
        b'\xf0\x9f\xa4\xb6\xf0\x9f\x8f\xbe',
        False,
        (
            'mrs_claus_tone4',
        ),
        None,
    ),
    Unicode(
        'mother_christmas_tone5',
        b'\xf0\x9f\xa4\xb6\xf0\x9f\x8f\xbf',
        False,
        (
            'mrs_claus_tone5',
        ),
        None,
    ),
    Unicode(
        'motor_scooter',
        b'\xf0\x9f\x9b\xb5',
        False,
        (
            'motorbike',
        ),
        None,
    ),
    Unicode(
        'motorboat',
        b'\xf0\x9f\x9b\xa5',
        True,
        None,
        None,
    ),
    Unicode(
        'motorcycle',
        b'\xf0\x9f\x8f\x8d',
        True,
        None,
        None,
    ),
    Unicode(
        'motorway',
        b'\xf0\x9f\x9b\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'mount_fuji',
        b'\xf0\x9f\x97\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'mountain',
        b'\xe2\x9b\xb0',
        True,
        None,
        None,
    ),
    Unicode(
        'mountain_bicyclist',
        b'\xf0\x9f\x9a\xb5',
        False,
        (
            'person_mountain_biking',
        ),
        None,
    ),
    Unicode(
        'mountain_bicyclist_tone1',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbb',
        False,
        (
            'person_mountain_biking_tone1',
        ),
        None,
    ),
    Unicode(
        'mountain_bicyclist_tone2',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbc',
        False,
        (
            'person_mountain_biking_tone2',
        ),
        None,
    ),
    Unicode(
        'mountain_bicyclist_tone3',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbd',
        False,
        (
            'person_mountain_biking_tone3',
        ),
        None,
    ),
    Unicode(
        'mountain_bicyclist_tone4',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbe',
        False,
        (
            'person_mountain_biking_tone4',
        ),
        None,
    ),
    Unicode(
        'mountain_bicyclist_tone5',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbf',
        False,
        (
            'person_mountain_biking_tone5',
        ),
        None,
    ),
    Unicode(
        'mountain_cableway',
        b'\xf0\x9f\x9a\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'mountain_railway',
        b'\xf0\x9f\x9a\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'mountain_snow',
        b'\xf0\x9f\x8f\x94',
        True,
        None,
        None,
    ),
    Unicode(
        'mouse2',
        b'\xf0\x9f\x90\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'mouse',
        b'\xf0\x9f\x90\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'mouse_three_button',
        b'\xf0\x9f\x96\xb1',
        True,
        None,
        None,
    ),
    Unicode(
        'movie_camera',
        b'\xf0\x9f\x8e\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'moyai',
        b'\xf0\x9f\x97\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'muscle',
        b'\xf0\x9f\x92\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'muscle_tone1',
        b'\xf0\x9f\x92\xaa\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'muscle_tone2',
        b'\xf0\x9f\x92\xaa\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'muscle_tone3',
        b'\xf0\x9f\x92\xaa\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'muscle_tone4',
        b'\xf0\x9f\x92\xaa\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'muscle_tone5',
        b'\xf0\x9f\x92\xaa\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'mushroom',
        b'\xf0\x9f\x8d\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'musical_keyboard',
        b'\xf0\x9f\x8e\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'musical_note',
        b'\xf0\x9f\x8e\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'musical_score',
        b'\xf0\x9f\x8e\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'mute',
        b'\xf0\x9f\x94\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'nail_care',
        b'\xf0\x9f\x92\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'nail_care_tone1',
        b'\xf0\x9f\x92\x85\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'nail_care_tone2',
        b'\xf0\x9f\x92\x85\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'nail_care_tone3',
        b'\xf0\x9f\x92\x85\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'nail_care_tone4',
        b'\xf0\x9f\x92\x85\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'nail_care_tone5',
        b'\xf0\x9f\x92\x85\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'name_badge',
        b'\xf0\x9f\x93\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'national_park',
        b'\xf0\x9f\x8f\x9e',
        True,
        None,
        None,
    ),
    Unicode(
        'nauseated_face',
        b'\xf0\x9f\xa4\xa2',
        False,
        (
            'sick',
        ),
        None,
    ),
    Unicode(
        'necktie',
        b'\xf0\x9f\x91\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'negative_squared_cross_mark',
        b'\xe2\x9d\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'nerd',
        b'\xf0\x9f\xa4\x93',
        False,
        (
            'nerd_face',
        ),
        None,
    ),
    Unicode(
        'new',
        b'\xf0\x9f\x86\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'new_moon',
        b'\xf0\x9f\x8c\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'new_moon_with_face',
        b'\xf0\x9f\x8c\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'newspaper2',
        b'\xf0\x9f\x97\x9e',
        True,
        None,
        None,
    ),
    Unicode(
        'newspaper',
        b'\xf0\x9f\x93\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'next_track',
        b'\xe2\x8f\xad',
        True,
        None,
        None,
    ),
    Unicode(
        'ng',
        b'\xf0\x9f\x86\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'night_with_stars',
        b'\xf0\x9f\x8c\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'nine',
        b'9\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'no_bell',
        b'\xf0\x9f\x94\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'no_bicycles',
        b'\xf0\x9f\x9a\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'no_entry',
        b'\xe2\x9b\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'no_entry_sign',
        b'\xf0\x9f\x9a\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'no_good',
        b'\xf0\x9f\x99\x85',
        False,
        (
            'person_gesturing_no',
        ),
        None,
    ),
    Unicode(
        'no_good_tone1',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbb',
        False,
        (
            'person_gesturing_no_tone1',
        ),
        None,
    ),
    Unicode(
        'no_good_tone2',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbc',
        False,
        (
            'person_gesturing_no_tone2',
        ),
        None,
    ),
    Unicode(
        'no_good_tone3',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbd',
        False,
        (
            'person_gesturing_no_tone3',
        ),
        None,
    ),
    Unicode(
        'no_good_tone4',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbe',
        False,
        (
            'person_gesturing_no_tone4',
        ),
        None,
    ),
    Unicode(
        'no_good_tone5',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbf',
        False,
        (
            'person_gesturing_no_tone5',
        ),
        None,
    ),
    Unicode(
        'no_mobile_phones',
        b'\xf0\x9f\x93\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'no_mouth',
        b'\xf0\x9f\x98\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'no_pedestrians',
        b'\xf0\x9f\x9a\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'no_smoking',
        b'\xf0\x9f\x9a\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'non_potable_water',
        b'\xf0\x9f\x9a\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'nose',
        b'\xf0\x9f\x91\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'nose_tone1',
        b'\xf0\x9f\x91\x83\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'nose_tone2',
        b'\xf0\x9f\x91\x83\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'nose_tone3',
        b'\xf0\x9f\x91\x83\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'nose_tone4',
        b'\xf0\x9f\x91\x83\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'nose_tone5',
        b'\xf0\x9f\x91\x83\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'notebook',
        b'\xf0\x9f\x93\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'notebook_with_decorative_cover',
        b'\xf0\x9f\x93\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'notepad_spiral',
        b'\xf0\x9f\x97\x92',
        True,
        None,
        None,
    ),
    Unicode(
        'notes',
        b'\xf0\x9f\x8e\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'nut_and_bolt',
        b'\xf0\x9f\x94\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'o2',
        b'\xf0\x9f\x85\xbe',
        True,
        None,
        None,
    ),
    Unicode(
        'o',
        b'\xe2\xad\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'ocean',
        b'\xf0\x9f\x8c\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'octagonal_sign',
        b'\xf0\x9f\x9b\x91',
        False,
        (
            'stop_sign',
        ),
        None,
    ),
    Unicode(
        'octopus',
        b'\xf0\x9f\x90\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'oden',
        b'\xf0\x9f\x8d\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'office',
        b'\xf0\x9f\x8f\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'oil',
        b'\xf0\x9f\x9b\xa2',
        True,
        None,
        None,
    ),
    Unicode(
        'ok',
        b'\xf0\x9f\x86\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'ok_hand',
        b'\xf0\x9f\x91\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'ok_hand_tone1',
        b'\xf0\x9f\x91\x8c\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'ok_hand_tone2',
        b'\xf0\x9f\x91\x8c\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'ok_hand_tone3',
        b'\xf0\x9f\x91\x8c\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'ok_hand_tone4',
        b'\xf0\x9f\x91\x8c\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'ok_hand_tone5',
        b'\xf0\x9f\x91\x8c\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'ok_woman',
        b'\xf0\x9f\x99\x86',
        False,
        (
            'person_gesturing_ok',
        ),
        None,
    ),
    Unicode(
        'ok_woman_tone1',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbb',
        False,
        (
            'person_gesturing_ok_tone1',
        ),
        None,
    ),
    Unicode(
        'ok_woman_tone2',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbc',
        False,
        (
            'person_gesturing_ok_tone2',
        ),
        None,
    ),
    Unicode(
        'ok_woman_tone3',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbd',
        False,
        (
            'person_gesturing_ok_tone3',
        ),
        None,
    ),
    Unicode(
        'ok_woman_tone4',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbe',
        False,
        (
            'person_gesturing_ok_tone4',
        ),
        None,
    ),
    Unicode(
        'ok_woman_tone5',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbf',
        False,
        (
            'person_gesturing_ok_tone5',
        ),
        None,
    ),
    Unicode(
        'older_man',
        b'\xf0\x9f\x91\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'older_man_tone1',
        b'\xf0\x9f\x91\xb4\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'older_man_tone2',
        b'\xf0\x9f\x91\xb4\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'older_man_tone3',
        b'\xf0\x9f\x91\xb4\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'older_man_tone4',
        b'\xf0\x9f\x91\xb4\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'older_man_tone5',
        b'\xf0\x9f\x91\xb4\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'om_symbol',
        b'\xf0\x9f\x95\x89',
        True,
        None,
        None,
    ),
    Unicode(
        'on',
        b'\xf0\x9f\x94\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'oncoming_automobile',
        b'\xf0\x9f\x9a\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'oncoming_bus',
        b'\xf0\x9f\x9a\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'oncoming_police_car',
        b'\xf0\x9f\x9a\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'oncoming_taxi',
        b'\xf0\x9f\x9a\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'one',
        b'1\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'open_file_folder',
        b'\xf0\x9f\x93\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'open_hands',
        b'\xf0\x9f\x91\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'open_hands_tone1',
        b'\xf0\x9f\x91\x90\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'open_hands_tone2',
        b'\xf0\x9f\x91\x90\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'open_hands_tone3',
        b'\xf0\x9f\x91\x90\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'open_hands_tone4',
        b'\xf0\x9f\x91\x90\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'open_hands_tone5',
        b'\xf0\x9f\x91\x90\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'ophiuchus',
        b'\xe2\x9b\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'orange_book',
        b'\xf0\x9f\x93\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'orthodox_cross',
        b'\xe2\x98\xa6',
        True,
        None,
        None,
    ),
    Unicode(
        'outbox_tray',
        b'\xf0\x9f\x93\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'owl',
        b'\xf0\x9f\xa6\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'ox',
        b'\xf0\x9f\x90\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'package',
        b'\xf0\x9f\x93\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'paella',
        b'\xf0\x9f\xa5\x98',
        False,
        (
            'shallow_pan_of_food',
        ),
        None,
    ),
    Unicode(
        'page_facing_up',
        b'\xf0\x9f\x93\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'page_with_curl',
        b'\xf0\x9f\x93\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'pager',
        b'\xf0\x9f\x93\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_tree',
        b'\xf0\x9f\x8c\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'pancakes',
        b'\xf0\x9f\xa5\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'panda_face',
        b'\xf0\x9f\x90\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'paperclip',
        b'\xf0\x9f\x93\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'parking',
        b'\xf0\x9f\x85\xbf',
        True,
        None,
        None,
    ),
    Unicode(
        'part_alternation_mark',
        b'\xe3\x80\xbd',
        True,
        None,
        None,
    ),
    Unicode(
        'partly_sunny',
        b'\xe2\x9b\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'passport_control',
        b'\xf0\x9f\x9b\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'peace',
        b'\xe2\x98\xae',
        True,
        None,
        None,
    ),
    Unicode(
        'peach',
        b'\xf0\x9f\x8d\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'peanuts',
        b'\xf0\x9f\xa5\x9c',
        False,
        (
            'shelled_peanut',
        ),
        None,
    ),
    Unicode(
        'pear',
        b'\xf0\x9f\x8d\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'pencil2',
        b'\xe2\x9c\x8f',
        True,
        None,
        None,
    ),
    Unicode(
        'pencil',
        b'\xf0\x9f\x93\x9d',
        False,
        (
            'memo',
        ),
        None,
    ),
    Unicode(
        'penguin',
        b'\xf0\x9f\x90\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'pensive',
        b'\xf0\x9f\x98\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'performing_arts',
        b'\xf0\x9f\x8e\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'persevere',
        b'\xf0\x9f\x98\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'person_frowning',
        b'\xf0\x9f\x99\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'person_frowning_tone1',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'person_frowning_tone2',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_frowning_tone3',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'person_frowning_tone4',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'person_frowning_tone5',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_blond_hair',
        b'\xf0\x9f\x91\xb1',
        False,
        (
            'blond_haired_person',
        ),
        None,
    ),
    Unicode(
        'person_with_blond_hair_tone1',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbb',
        False,
        (
            'blond_haired_person_tone1',
        ),
        None,
    ),
    Unicode(
        'person_with_blond_hair_tone2',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbc',
        False,
        (
            'blond_haired_person_tone2',
        ),
        None,
    ),
    Unicode(
        'person_with_blond_hair_tone3',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbd',
        False,
        (
            'blond_haired_person_tone3',
        ),
        None,
    ),
    Unicode(
        'person_with_blond_hair_tone4',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbe',
        False,
        (
            'blond_haired_person_tone4',
        ),
        None,
    ),
    Unicode(
        'person_with_blond_hair_tone5',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbf',
        False,
        (
            'blond_haired_person_tone5',
        ),
        None,
    ),
    Unicode(
        'person_with_pouting_face',
        b'\xf0\x9f\x99\x8e',
        False,
        (
            'person_pouting',
        ),
        None,
    ),
    Unicode(
        'person_with_pouting_face_tone1',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbb',
        False,
        (
            'person_pouting_tone1',
        ),
        None,
    ),
    Unicode(
        'person_with_pouting_face_tone2',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbc',
        False,
        (
            'person_pouting_tone2',
        ),
        None,
    ),
    Unicode(
        'person_with_pouting_face_tone3',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbd',
        False,
        (
            'person_pouting_tone3',
        ),
        None,
    ),
    Unicode(
        'person_with_pouting_face_tone4',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbe',
        False,
        (
            'person_pouting_tone4',
        ),
        None,
    ),
    Unicode(
        'person_with_pouting_face_tone5',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbf',
        False,
        (
            'person_pouting_tone5',
        ),
        None,
    ),
    Unicode(
        'pick',
        b'\xe2\x9b\x8f',
        True,
        None,
        None,
    ),
    Unicode(
        'pig2',
        b'\xf0\x9f\x90\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'pig',
        b'\xf0\x9f\x90\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'pig_nose',
        b'\xf0\x9f\x90\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'pill',
        b'\xf0\x9f\x92\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'pineapple',
        b'\xf0\x9f\x8d\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'ping_pong',
        b'\xf0\x9f\x8f\x93',
        False,
        (
            'table_tennis',
        ),
        None,
    ),
    Unicode(
        'pisces',
        b'\xe2\x99\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'pizza',
        b'\xf0\x9f\x8d\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'place_of_worship',
        b'\xf0\x9f\x9b\x90',
        False,
        (
            'worship_symbol',
        ),
        None,
    ),
    Unicode(
        'play_pause',
        b'\xe2\x8f\xaf',
        True,
        None,
        None,
    ),
    Unicode(
        'point_down',
        b'\xf0\x9f\x91\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'point_down_tone1',
        b'\xf0\x9f\x91\x87\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'point_down_tone2',
        b'\xf0\x9f\x91\x87\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'point_down_tone3',
        b'\xf0\x9f\x91\x87\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'point_down_tone4',
        b'\xf0\x9f\x91\x87\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'point_down_tone5',
        b'\xf0\x9f\x91\x87\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'point_left',
        b'\xf0\x9f\x91\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'point_left_tone1',
        b'\xf0\x9f\x91\x88\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'point_left_tone2',
        b'\xf0\x9f\x91\x88\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'point_left_tone3',
        b'\xf0\x9f\x91\x88\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'point_left_tone4',
        b'\xf0\x9f\x91\x88\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'point_left_tone5',
        b'\xf0\x9f\x91\x88\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'point_right',
        b'\xf0\x9f\x91\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'point_right_tone1',
        b'\xf0\x9f\x91\x89\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'point_right_tone2',
        b'\xf0\x9f\x91\x89\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'point_right_tone3',
        b'\xf0\x9f\x91\x89\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'point_right_tone4',
        b'\xf0\x9f\x91\x89\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'point_right_tone5',
        b'\xf0\x9f\x91\x89\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up',
        b'\xe2\x98\x9d',
        True,
        None,
        None,
    ),
    Unicode(
        'point_up_tone1',
        b'\xe2\x98\x9d\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_tone2',
        b'\xe2\x98\x9d\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_tone3',
        b'\xe2\x98\x9d\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_tone4',
        b'\xe2\x98\x9d\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_tone5',
        b'\xe2\x98\x9d\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_2',
        b'\xf0\x9f\x91\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_2_tone1',
        b'\xf0\x9f\x91\x86\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_2_tone2',
        b'\xf0\x9f\x91\x86\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_2_tone3',
        b'\xf0\x9f\x91\x86\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_2_tone4',
        b'\xf0\x9f\x91\x86\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up_2_tone5',
        b'\xf0\x9f\x91\x86\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'police_car',
        b'\xf0\x9f\x9a\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'poodle',
        b'\xf0\x9f\x90\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'popcorn',
        b'\xf0\x9f\x8d\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'post_office',
        b'\xf0\x9f\x8f\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'postal_horn',
        b'\xf0\x9f\x93\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'postbox',
        b'\xf0\x9f\x93\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'potable_water',
        b'\xf0\x9f\x9a\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'potato',
        b'\xf0\x9f\xa5\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'pouch',
        b'\xf0\x9f\x91\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'poultry_leg',
        b'\xf0\x9f\x8d\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'pound',
        b'\xf0\x9f\x92\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'pouting_cat',
        b'\xf0\x9f\x98\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'pray',
        b'\xf0\x9f\x99\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'pray_tone1',
        b'\xf0\x9f\x99\x8f\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'pray_tone2',
        b'\xf0\x9f\x99\x8f\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'pray_tone3',
        b'\xf0\x9f\x99\x8f\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'pray_tone4',
        b'\xf0\x9f\x99\x8f\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'pray_tone5',
        b'\xf0\x9f\x99\x8f\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'prayer_beads',
        b'\xf0\x9f\x93\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'previous_track',
        b'\xe2\x8f\xae',
        True,
        None,
        None,
    ),
    Unicode(
        'prince',
        b'\xf0\x9f\xa4\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'prince_tone1',
        b'\xf0\x9f\xa4\xb4\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'prince_tone2',
        b'\xf0\x9f\xa4\xb4\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'prince_tone3',
        b'\xf0\x9f\xa4\xb4\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'prince_tone4',
        b'\xf0\x9f\xa4\xb4\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'prince_tone5',
        b'\xf0\x9f\xa4\xb4\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'princess',
        b'\xf0\x9f\x91\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'princess_tone1',
        b'\xf0\x9f\x91\xb8\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'princess_tone2',
        b'\xf0\x9f\x91\xb8\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'princess_tone3',
        b'\xf0\x9f\x91\xb8\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'princess_tone4',
        b'\xf0\x9f\x91\xb8\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'princess_tone5',
        b'\xf0\x9f\x91\xb8\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'printer',
        b'\xf0\x9f\x96\xa8',
        True,
        None,
        None,
    ),
    Unicode(
        'punch',
        b'\xf0\x9f\x91\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'punch_tone1',
        b'\xf0\x9f\x91\x8a\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'punch_tone2',
        b'\xf0\x9f\x91\x8a\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'punch_tone3',
        b'\xf0\x9f\x91\x8a\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'punch_tone4',
        b'\xf0\x9f\x91\x8a\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'punch_tone5',
        b'\xf0\x9f\x91\x8a\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'purple_heart',
        b'\xf0\x9f\x92\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'purse',
        b'\xf0\x9f\x91\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'pushpin',
        b'\xf0\x9f\x93\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'put_litter_in_its_place',
        b'\xf0\x9f\x9a\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'question',
        b'\xe2\x9d\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'rabbit2',
        b'\xf0\x9f\x90\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'rabbit',
        b'\xf0\x9f\x90\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'race_car',
        b'\xf0\x9f\x8f\x8e',
        True,
        None,
        None,
    ),
    Unicode(
        'racehorse',
        b'\xf0\x9f\x90\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'radio',
        b'\xf0\x9f\x93\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'radio_button',
        b'\xf0\x9f\x94\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'radioactive',
        b'\xe2\x98\xa2',
        True,
        None,
        None,
    ),
    Unicode(
        'railroad_track',
        b'\xf0\x9f\x9b\xa4',
        True,
        None,
        None,
    ),
    Unicode(
        'railway_car',
        b'\xf0\x9f\x9a\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'rainbow',
        b'\xf0\x9f\x8c\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hand',
        b'\xe2\x9c\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hand_tone1',
        b'\xe2\x9c\x8b\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hand_tone2',
        b'\xe2\x9c\x8b\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hand_tone3',
        b'\xe2\x9c\x8b\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hand_tone4',
        b'\xe2\x9c\x8b\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hand_tone5',
        b'\xe2\x9c\x8b\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hand_with_part_between_middle_and_ring_fingers',
        b'\xf0\x9f\x96\x96',
        False,
        (
            'vulcan',
        ),
        None,
    ),
    Unicode(
        'raised_hand_with_part_between_middle_and_ring_fingers_tone1',
        b'\xf0\x9f\x96\x96\xf0\x9f\x8f\xbb',
        False,
        (
            'vulcan_tone1',
        ),
        None,
    ),
    Unicode(
        'raised_hand_with_part_between_middle_and_ring_fingers_tone2',
        b'\xf0\x9f\x96\x96\xf0\x9f\x8f\xbc',
        False,
        (
            'vulcan_tone2',
        ),
        None,
    ),
    Unicode(
        'raised_hand_with_part_between_middle_and_ring_fingers_tone3',
        b'\xf0\x9f\x96\x96\xf0\x9f\x8f\xbd',
        False,
        (
            'vulcan_tone3',
        ),
        None,
    ),
    Unicode(
        'raised_hand_with_part_between_middle_and_ring_fingers_tone4',
        b'\xf0\x9f\x96\x96\xf0\x9f\x8f\xbe',
        False,
        (
            'vulcan_tone4',
        ),
        None,
    ),
    Unicode(
        'raised_hand_with_part_between_middle_and_ring_fingers_tone5',
        b'\xf0\x9f\x96\x96\xf0\x9f\x8f\xbf',
        False,
        (
            'vulcan_tone5',
        ),
        None,
    ),
    Unicode(
        'raised_hands',
        b'\xf0\x9f\x99\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hands_tone1',
        b'\xf0\x9f\x99\x8c\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hands_tone2',
        b'\xf0\x9f\x99\x8c\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hands_tone3',
        b'\xf0\x9f\x99\x8c\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hands_tone4',
        b'\xf0\x9f\x99\x8c\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'raised_hands_tone5',
        b'\xf0\x9f\x99\x8c\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'raising_hand',
        b'\xf0\x9f\x99\x8b',
        False,
        (
            'person_raising_hand',
        ),
        None,
    ),
    Unicode(
        'raising_hand_tone1',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbb',
        False,
        (
            'person_raising_hand_tone1',
        ),
        None,
    ),
    Unicode(
        'raising_hand_tone2',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbc',
        False,
        (
            'person_raising_hand_tone2',
        ),
        None,
    ),
    Unicode(
        'raising_hand_tone3',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbd',
        False,
        (
            'person_raising_hand_tone3',
        ),
        None,
    ),
    Unicode(
        'raising_hand_tone4',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbe',
        False,
        (
            'person_raising_hand_tone4',
        ),
        None,
    ),
    Unicode(
        'raising_hand_tone5',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbf',
        False,
        (
            'person_raising_hand_tone5',
        ),
        None,
    ),
    Unicode(
        'ram',
        b'\xf0\x9f\x90\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'ramen',
        b'\xf0\x9f\x8d\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'rat',
        b'\xf0\x9f\x90\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'record_button',
        b'\xe2\x8f\xba',
        True,
        None,
        None,
    ),
    Unicode(
        'recycle',
        b'\xe2\x99\xbb',
        True,
        None,
        None,
    ),
    Unicode(
        'red_car',
        b'\xf0\x9f\x9a\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'red_circle',
        b'\xf0\x9f\x94\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_a',
        b'\xf0\x9f\x87\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_b',
        b'\xf0\x9f\x87\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_c',
        b'\xf0\x9f\x87\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_d',
        b'\xf0\x9f\x87\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_e',
        b'\xf0\x9f\x87\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_f',
        b'\xf0\x9f\x87\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_g',
        b'\xf0\x9f\x87\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_h',
        b'\xf0\x9f\x87\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_i',
        b'\xf0\x9f\x87\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_j',
        b'\xf0\x9f\x87\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_k',
        b'\xf0\x9f\x87\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_l',
        b'\xf0\x9f\x87\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_m',
        b'\xf0\x9f\x87\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_n',
        b'\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_o',
        b'\xf0\x9f\x87\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_p',
        b'\xf0\x9f\x87\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_q',
        b'\xf0\x9f\x87\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_r',
        b'\xf0\x9f\x87\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_s',
        b'\xf0\x9f\x87\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_t',
        b'\xf0\x9f\x87\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_u',
        b'\xf0\x9f\x87\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_v',
        b'\xf0\x9f\x87\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_w',
        b'\xf0\x9f\x87\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_x',
        b'\xf0\x9f\x87\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_y',
        b'\xf0\x9f\x87\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'regional_indicator_z',
        b'\xf0\x9f\x87\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'registered',
        b'\xc2\xae',
        True,
        None,
        None,
    ),
    Unicode(
        'relaxed',
        b'\xe2\x98\xba',
        True,
        None,
        None,
    ),
    Unicode(
        'relieved',
        b'\xf0\x9f\x98\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'reminder_ribbon',
        b'\xf0\x9f\x8e\x97',
        True,
        None,
        None,
    ),
    Unicode(
        'repeat',
        b'\xf0\x9f\x94\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'repeat_one',
        b'\xf0\x9f\x94\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'restroom',
        b'\xf0\x9f\x9a\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'revolving_hearts',
        b'\xf0\x9f\x92\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'rewind',
        b'\xe2\x8f\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'rhino',
        b'\xf0\x9f\xa6\x8f',
        False,
        (
            'rhinoceros',
        ),
        None,
    ),
    Unicode(
        'ribbon',
        b'\xf0\x9f\x8e\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'rice',
        b'\xf0\x9f\x8d\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'rice_ball',
        b'\xf0\x9f\x8d\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'rice_cracker',
        b'\xf0\x9f\x8d\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'rice_scene',
        b'\xf0\x9f\x8e\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'right_facing_fist',
        b'\xf0\x9f\xa4\x9c',
        False,
        (
            'right_fist',
        ),
        None,
    ),
    Unicode(
        'right_facing_fist_tone1',
        b'\xf0\x9f\xa4\x9c\xf0\x9f\x8f\xbb',
        False,
        (
            'right_fist_tone1',
        ),
        None,
    ),
    Unicode(
        'right_facing_fist_tone2',
        b'\xf0\x9f\xa4\x9c\xf0\x9f\x8f\xbc',
        False,
        (
            'right_fist_tone2',
        ),
        None,
    ),
    Unicode(
        'right_facing_fist_tone3',
        b'\xf0\x9f\xa4\x9c\xf0\x9f\x8f\xbd',
        False,
        (
            'right_fist_tone3',
        ),
        None,
    ),
    Unicode(
        'right_facing_fist_tone4',
        b'\xf0\x9f\xa4\x9c\xf0\x9f\x8f\xbe',
        False,
        (
            'right_fist_tone4',
        ),
        None,
    ),
    Unicode(
        'right_facing_fist_tone5',
        b'\xf0\x9f\xa4\x9c\xf0\x9f\x8f\xbf',
        False,
        (
            'right_fist_tone5',
        ),
        None,
    ),
    Unicode(
        'ring',
        b'\xf0\x9f\x92\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'robot',
        b'\xf0\x9f\xa4\x96',
        False,
        (
            'robot_face',
        ),
        None,
    ),
    Unicode(
        'rocket',
        b'\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'rofl',
        b'\xf0\x9f\xa4\xa3',
        False,
        (
            'rolling_on_the_floor_laughing',
        ),
        None,
    ),
    Unicode(
        'roller_coaster',
        b'\xf0\x9f\x8e\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'rooster',
        b'\xf0\x9f\x90\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'rose',
        b'\xf0\x9f\x8c\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'rosette',
        b'\xf0\x9f\x8f\xb5',
        True,
        None,
        None,
    ),
    Unicode(
        'rotating_light',
        b'\xf0\x9f\x9a\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'round_pushpin',
        b'\xf0\x9f\x93\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'rowboat',
        b'\xf0\x9f\x9a\xa3',
        False,
        (
            'person_rowing_boat',
        ),
        None,
    ),
    Unicode(
        'rowboat_tone1',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbb',
        False,
        (
            'person_rowing_boat_tone1',
        ),
        None,
    ),
    Unicode(
        'rowboat_tone2',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbc',
        False,
        (
            'person_rowing_boat_tone2',
        ),
        None,
    ),
    Unicode(
        'rowboat_tone3',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbd',
        False,
        (
            'person_rowing_boat_tone3',
        ),
        None,
    ),
    Unicode(
        'rowboat_tone4',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbe',
        False,
        (
            'person_rowing_boat_tone4',
        ),
        None,
    ),
    Unicode(
        'rowboat_tone5',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbf',
        False,
        (
            'person_rowing_boat_tone5',
        ),
        None,
    ),
    Unicode(
        'rugby_football',
        b'\xf0\x9f\x8f\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'runner',
        b'\xf0\x9f\x8f\x83',
        False,
        (
            'person_running',
        ),
        None,
    ),
    Unicode(
        'runner_tone1',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbb',
        False,
        (
            'person_running_tone1',
        ),
        None,
    ),
    Unicode(
        'runner_tone2',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbc',
        False,
        (
            'person_running_tone2',
        ),
        None,
    ),
    Unicode(
        'runner_tone3',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbd',
        False,
        (
            'person_running_tone3',
        ),
        None,
    ),
    Unicode(
        'runner_tone4',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbe',
        False,
        (
            'person_running_tone4',
        ),
        None,
    ),
    Unicode(
        'runner_tone5',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbf',
        False,
        (
            'person_running_tone5',
        ),
        None,
    ),
    Unicode(
        'running_shirt_with_sash',
        b'\xf0\x9f\x8e\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'sa',
        b'\xf0\x9f\x88\x82',
        True,
        None,
        None,
    ),
    Unicode(
        'sagittarius',
        b'\xe2\x99\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'sailboat',
        b'\xe2\x9b\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'sake',
        b'\xf0\x9f\x8d\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'sandal',
        b'\xf0\x9f\x91\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'santa',
        b'\xf0\x9f\x8e\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'santa_tone1',
        b'\xf0\x9f\x8e\x85\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'santa_tone2',
        b'\xf0\x9f\x8e\x85\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'santa_tone3',
        b'\xf0\x9f\x8e\x85\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'santa_tone4',
        b'\xf0\x9f\x8e\x85\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'santa_tone5',
        b'\xf0\x9f\x8e\x85\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'satellite',
        b'\xf0\x9f\x93\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'satellite_orbital',
        b'\xf0\x9f\x9b\xb0',
        True,
        None,
        None,
    ),
    Unicode(
        'saxophone',
        b'\xf0\x9f\x8e\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'scales',
        b'\xe2\x9a\x96',
        True,
        None,
        None,
    ),
    Unicode(
        'school',
        b'\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'school_satchel',
        b'\xf0\x9f\x8e\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'scissors',
        b'\xe2\x9c\x82',
        True,
        None,
        None,
    ),
    Unicode(
        'scooter',
        b'\xf0\x9f\x9b\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'scorpion',
        b'\xf0\x9f\xa6\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'scorpius',
        b'\xe2\x99\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'scream',
        b'\xf0\x9f\x98\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'scream_cat',
        b'\xf0\x9f\x99\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'scroll',
        b'\xf0\x9f\x93\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'seat',
        b'\xf0\x9f\x92\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'second_place',
        b'\xf0\x9f\xa5\x88',
        False,
        (
            'second_place_medal',
        ),
        None,
    ),
    Unicode(
        'secret',
        b'\xe3\x8a\x99',
        True,
        None,
        None,
    ),
    Unicode(
        'see_no_evil',
        b'\xf0\x9f\x99\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'seedling',
        b'\xf0\x9f\x8c\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'selfie',
        b'\xf0\x9f\xa4\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'selfie_tone1',
        b'\xf0\x9f\xa4\xb3\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'selfie_tone2',
        b'\xf0\x9f\xa4\xb3\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'selfie_tone3',
        b'\xf0\x9f\xa4\xb3\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'selfie_tone4',
        b'\xf0\x9f\xa4\xb3\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'selfie_tone5',
        b'\xf0\x9f\xa4\xb3\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'seven',
        b'7\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'shamrock',
        b'\xe2\x98\x98',
        True,
        None,
        None,
    ),
    Unicode(
        'shark',
        b'\xf0\x9f\xa6\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'shaved_ice',
        b'\xf0\x9f\x8d\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'sheep',
        b'\xf0\x9f\x90\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'shell',
        b'\xf0\x9f\x90\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'shield',
        b'\xf0\x9f\x9b\xa1',
        True,
        None,
        None,
    ),
    Unicode(
        'shinto_shrine',
        b'\xe2\x9b\xa9',
        True,
        None,
        None,
    ),
    Unicode(
        'ship',
        b'\xf0\x9f\x9a\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'shirt',
        b'\xf0\x9f\x91\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'shopping_bags',
        b'\xf0\x9f\x9b\x8d',
        True,
        None,
        None,
    ),
    Unicode(
        'shopping_cart',
        b'\xf0\x9f\x9b\x92',
        False,
        (
            'shopping_trolley',
        ),
        None,
    ),
    Unicode(
        'shower',
        b'\xf0\x9f\x9a\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'shrimp',
        b'\xf0\x9f\xa6\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'shrug',
        b'\xf0\x9f\xa4\xb7',
        False,
        (
            'person_shrugging',
        ),
        None,
    ),
    Unicode(
        'shrug_tone1',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbb',
        False,
        (
            'person_shrugging_tone1',
        ),
        None,
    ),
    Unicode(
        'shrug_tone2',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbc',
        False,
        (
            'person_shrugging_tone2',
        ),
        None,
    ),
    Unicode(
        'shrug_tone3',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbd',
        False,
        (
            'person_shrugging_tone3',
        ),
        None,
    ),
    Unicode(
        'shrug_tone4',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbe',
        False,
        (
            'person_shrugging_tone4',
        ),
        None,
    ),
    Unicode(
        'shrug_tone5',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbf',
        False,
        (
            'person_shrugging_tone5',
        ),
        None,
    ),
    Unicode(
        'signal_strength',
        b'\xf0\x9f\x93\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'six',
        b'6\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'six_pointed_star',
        b'\xf0\x9f\x94\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'skeleton',
        b'\xf0\x9f\x92\x80',
        False,
        (
            'skull',
        ),
        None,
    ),
    Unicode(
        'ski',
        b'\xf0\x9f\x8e\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'skier',
        b'\xe2\x9b\xb7',
        True,
        None,
        None,
    ),
    Unicode(
        'skier_tone1',
        b'\xe2\x9b\xb7\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'skier_tone2',
        b'\xe2\x9b\xb7\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'skier_tone3',
        b'\xe2\x9b\xb7\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'skier_tone4',
        b'\xe2\x9b\xb7\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'skier_tone5',
        b'\xe2\x9b\xb7\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'skull_and_crossbones',
        b'\xe2\x98\xa0',
        True,
        None,
        None,
    ),
    Unicode(
        'sleeping',
        b'\xf0\x9f\x98\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'sleeping_accommodation',
        b'\xf0\x9f\x9b\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'sleeping_accommodation_tone1',
        b'\xf0\x9f\x9b\x8c\xf0\x9f\x8f\xbb',
        False,
        (
            'person_in_bed_tone1',
        ),
        None,
    ),
    Unicode(
        'sleeping_accommodation_tone2',
        b'\xf0\x9f\x9b\x8c\xf0\x9f\x8f\xbc',
        False,
        (
            'person_in_bed_tone2',
        ),
        None,
    ),
    Unicode(
        'sleeping_accommodation_tone3',
        b'\xf0\x9f\x9b\x8c\xf0\x9f\x8f\xbd',
        False,
        (
            'person_in_bed_tone3',
        ),
        None,
    ),
    Unicode(
        'sleeping_accommodation_tone4',
        b'\xf0\x9f\x9b\x8c\xf0\x9f\x8f\xbe',
        False,
        (
            'person_in_bed_tone4',
        ),
        None,
    ),
    Unicode(
        'sleeping_accommodation_tone5',
        b'\xf0\x9f\x9b\x8c\xf0\x9f\x8f\xbf',
        False,
        (
            'person_in_bed_tone5',
        ),
        None,
    ),
    Unicode(
        'sleepy',
        b'\xf0\x9f\x98\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'sleuth_or_spy',
        b'\xf0\x9f\x95\xb5',
        False,
        (
            'spy',
        ),
        None,
    ),
    Unicode(
        'sleuth_or_spy_tone1',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbb',
        False,
        (
            'spy_tone1',
            'detective_tone1'
        ),
        None,
    ),
    Unicode(
        'sleuth_or_spy_tone2',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbc',
        False,
        (
            'spy_tone2',
            'detective_tone2'
        ),
        None,
    ),
    Unicode(
        'sleuth_or_spy_tone3',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbd',
        False,
        (
            'spy_tone3',
            'detective_tone3'
        ),
        None,
    ),
    Unicode(
        'sleuth_or_spy_tone4',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbe',
        False,
        (
            'spy_tone4',
            'detective_tone4'
        ),
        None,
    ),
    Unicode(
        'sleuth_or_spy_tone5',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbf',
        False,
        (
            'spy_tone5',
            'detective_tone5'
        ),
        None,
    ),
    Unicode(
        'slight_frown',
        b'\xf0\x9f\x99\x81',
        False,
        (
            'slightly_frowning_face',
        ),
        None,
    ),
    Unicode(
        'slight_smile',
        b'\xf0\x9f\x99\x82',
        False,
        (
            'slightly_smiling_face',
        ),
        (':)', ':-)', '=)', '=-)'),
    ),
    Unicode(
        'slot_machine',
        b'\xf0\x9f\x8e\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'small_blue_diamond',
        b'\xf0\x9f\x94\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'small_orange_diamond',
        b'\xf0\x9f\x94\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'small_red_triangle',
        b'\xf0\x9f\x94\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'small_red_triangle_down',
        b'\xf0\x9f\x94\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'smile_cat',
        b'\xf0\x9f\x98\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'smiley_cat',
        b'\xf0\x9f\x98\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'smiling_imp',
        b'\xf0\x9f\x98\x88',
        False,
        None,
        (']:)', ']:-)', ']=)', ']=-)'),
    ),
    Unicode(
        'smirk',
        b'\xf0\x9f\x98\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'smirk_cat',
        b'\xf0\x9f\x98\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'smoking',
        b'\xf0\x9f\x9a\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'snail',
        b'\xf0\x9f\x90\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'snake',
        b'\xf0\x9f\x90\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'sneeze',
        b'\xf0\x9f\xa4\xa7',
        False,
        (
            'sneezing_face',
        ),
        None,
    ),
    Unicode(
        'snowboarder',
        b'\xf0\x9f\x8f\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'snowboarder_tone1',
        b'\xf0\x9f\x8f\x82\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'snowboarder_tone2',
        b'\xf0\x9f\x8f\x82\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'snowboarder_tone3',
        b'\xf0\x9f\x8f\x82\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'snowboarder_tone4',
        b'\xf0\x9f\x8f\x82\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'snowboarder_tone5',
        b'\xf0\x9f\x8f\x82\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'snowflake',
        b'\xe2\x9d\x84',
        True,
        None,
        None,
    ),
    Unicode(
        'snowman2',
        b'\xe2\x98\x83',
        True,
        None,
        None,
    ),
    Unicode(
        'snowman',
        b'\xe2\x9b\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'soccer',
        b'\xe2\x9a\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'soon',
        b'\xf0\x9f\x94\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'sos',
        b'\xf0\x9f\x86\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'sound',
        b'\xf0\x9f\x94\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'space_invader',
        b'\xf0\x9f\x91\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'spades',
        b'\xe2\x99\xa0',
        True,
        None,
        None,
    ),
    Unicode(
        'spaghetti',
        b'\xf0\x9f\x8d\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'sparkle',
        b'\xe2\x9d\x87',
        True,
        None,
        None,
    ),
    Unicode(
        'sparkler',
        b'\xf0\x9f\x8e\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'sparkles',
        b'\xe2\x9c\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'sparkling_heart',
        b'\xf0\x9f\x92\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'speak_no_evil',
        b'\xf0\x9f\x99\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'speaker',
        b'\xf0\x9f\x94\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'speaking_head',
        b'\xf0\x9f\x97\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'speech_balloon',
        b'\xf0\x9f\x92\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'speedboat',
        b'\xf0\x9f\x9a\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'spider',
        b'\xf0\x9f\x95\xb7',
        True,
        None,
        None,
    ),
    Unicode(
        'spider_web',
        b'\xf0\x9f\x95\xb8',
        True,
        None,
        None,
    ),
    Unicode(
        'spoon',
        b'\xf0\x9f\xa5\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'squid',
        b'\xf0\x9f\xa6\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'stadium',
        b'\xf0\x9f\x8f\x9f',
        True,
        None,
        None,
    ),
    Unicode(
        'star2',
        b'\xf0\x9f\x8c\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'star',
        b'\xe2\xad\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'star_and_crescent',
        b'\xe2\x98\xaa',
        True,
        None,
        None,
    ),
    Unicode(
        'star_of_david',
        b'\xe2\x9c\xa1',
        True,
        None,
        None,
    ),
    Unicode(
        'stars',
        b'\xf0\x9f\x8c\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'station',
        b'\xf0\x9f\x9a\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'statue_of_liberty',
        b'\xf0\x9f\x97\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'steam_locomotive',
        b'\xf0\x9f\x9a\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'stew',
        b'\xf0\x9f\x8d\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'stop_button',
        b'\xe2\x8f\xb9',
        True,
        None,
        None,
    ),
    Unicode(
        'stopwatch',
        b'\xe2\x8f\xb1',
        True,
        None,
        None,
    ),
    Unicode(
        'straight_ruler',
        b'\xf0\x9f\x93\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'strawberry',
        b'\xf0\x9f\x8d\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'stuck_out_tongue_closed_eyes',
        b'\xf0\x9f\x98\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'stuck_out_tongue_winking_eye',
        b'\xf0\x9f\x98\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'stuffed_flatbread',
        b'\xf0\x9f\xa5\x99',
        False,
        (
            'stuffed_pita',
        ),
        None,
    ),
    Unicode(
        'sun_with_face',
        b'\xf0\x9f\x8c\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'sunflower',
        b'\xf0\x9f\x8c\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'sunny',
        b'\xe2\x98\x80',
        True,
        None,
        None,
    ),
    Unicode(
        'sunrise',
        b'\xf0\x9f\x8c\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'sunrise_over_mountains',
        b'\xf0\x9f\x8c\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'surfer',
        b'\xf0\x9f\x8f\x84',
        False,
        (
            'person_surfing',
        ),
        None,
    ),
    Unicode(
        'surfer_tone1',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbb',
        False,
        (
            'person_surfing_tone1',
        ),
        None,
    ),
    Unicode(
        'surfer_tone2',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbc',
        False,
        (
            'person_surfing_tone2',
        ),
        None,
    ),
    Unicode(
        'surfer_tone3',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbd',
        False,
        (
            'person_surfing_tone3',
        ),
        None,
    ),
    Unicode(
        'surfer_tone4',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbe',
        False,
        (
            'person_surfing_tone4',
        ),
        None,
    ),
    Unicode(
        'surfer_tone5',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbf',
        False,
        (
            'person_surfing_tone5',
        ),
        None,
    ),
    Unicode(
        'sushi',
        b'\xf0\x9f\x8d\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'suspension_railway',
        b'\xf0\x9f\x9a\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'sweat_drops',
        b'\xf0\x9f\x92\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'sweet_potato',
        b'\xf0\x9f\x8d\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'swimmer',
        b'\xf0\x9f\x8f\x8a',
        False,
        (
            'person_swimming',
        ),
        None,
    ),
    Unicode(
        'swimmer_tone1',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbb',
        False,
        (
            'person_swimming_tone1',
        ),
        None,
    ),
    Unicode(
        'swimmer_tone2',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbc',
        False,
        (
            'person_swimming_tone2',
        ),
        None,
    ),
    Unicode(
        'swimmer_tone3',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbd',
        False,
        (
            'person_swimming_tone3',
        ),
        None,
    ),
    Unicode(
        'swimmer_tone4',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbe',
        False,
        (
            'person_swimming_tone4',
        ),
        None,
    ),
    Unicode(
        'swimmer_tone5',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbf',
        False,
        (
            'person_swimming_tone5',
        ),
        None,
    ),
    Unicode(
        'symbols',
        b'\xf0\x9f\x94\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'synagogue',
        b'\xf0\x9f\x95\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'syringe',
        b'\xf0\x9f\x92\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'taco',
        b'\xf0\x9f\x8c\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'tada',
        b'\xf0\x9f\x8e\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'tanabata_tree',
        b'\xf0\x9f\x8e\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'tangerine',
        b'\xf0\x9f\x8d\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'taurus',
        b'\xe2\x99\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'taxi',
        b'\xf0\x9f\x9a\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'tea',
        b'\xf0\x9f\x8d\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'telephone',
        b'\xe2\x98\x8e',
        True,
        None,
        None,
    ),
    Unicode(
        'telephone_receiver',
        b'\xf0\x9f\x93\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'telescope',
        b'\xf0\x9f\x94\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'tennis',
        b'\xf0\x9f\x8e\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'tent',
        b'\xe2\x9b\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'thermometer',
        b'\xf0\x9f\x8c\xa1',
        True,
        None,
        None,
    ),
    Unicode(
        'thinking',
        b'\xf0\x9f\xa4\x94',
        False,
        (
            'thinking_face',
        ),
        None,
    ),
    Unicode(
        'third_place',
        b'\xf0\x9f\xa5\x89',
        False,
        (
            'third_place_medal',
        ),
        None,
    ),
    Unicode(
        'thought_balloon',
        b'\xf0\x9f\x92\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'three',
        b'3\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'thunder_cloud_and_rain',
        b'\xe2\x9b\x88',
        True,
        None,
        None,
    ),
    Unicode(
        'ticket',
        b'\xf0\x9f\x8e\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'tiger2',
        b'\xf0\x9f\x90\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'tiger',
        b'\xf0\x9f\x90\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'timer',
        b'\xe2\x8f\xb2',
        True,
        None,
        None,
    ),
    Unicode(
        'tired_face',
        b'\xf0\x9f\x98\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'tm',
        b'\xe2\x84\xa2',
        True,
        None,
        None,
    ),
    Unicode(
        'toilet',
        b'\xf0\x9f\x9a\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'tokyo_tower',
        b'\xf0\x9f\x97\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'tomato',
        b'\xf0\x9f\x8d\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'tongue',
        b'\xf0\x9f\x91\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'top',
        b'\xf0\x9f\x94\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'tophat',
        b'\xf0\x9f\x8e\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'trackball',
        b'\xf0\x9f\x96\xb2',
        True,
        None,
        None,
    ),
    Unicode(
        'tractor',
        b'\xf0\x9f\x9a\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'traffic_light',
        b'\xf0\x9f\x9a\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'train2',
        b'\xf0\x9f\x9a\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'train',
        b'\xf0\x9f\x9a\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'tram',
        b'\xf0\x9f\x9a\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'triangular_flag_on_post',
        b'\xf0\x9f\x9a\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'triangular_ruler',
        b'\xf0\x9f\x93\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'trident',
        b'\xf0\x9f\x94\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'triumph',
        b'\xf0\x9f\x98\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'trolleybus',
        b'\xf0\x9f\x9a\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'trophy',
        b'\xf0\x9f\x8f\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'tropical_drink',
        b'\xf0\x9f\x8d\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'tropical_fish',
        b'\xf0\x9f\x90\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'truck',
        b'\xf0\x9f\x9a\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'trumpet',
        b'\xf0\x9f\x8e\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'tulip',
        b'\xf0\x9f\x8c\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'tumbler_glass',
        b'\xf0\x9f\xa5\x83',
        False,
        (
            'whisky',
        ),
        None,
    ),
    Unicode(
        'turkey',
        b'\xf0\x9f\xa6\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'turtle',
        b'\xf0\x9f\x90\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'tv',
        b'\xf0\x9f\x93\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'twisted_rightwards_arrows',
        b'\xf0\x9f\x94\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'two',
        b'2\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'two_hearts',
        b'\xf0\x9f\x92\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'two_men_holding_hands',
        b'\xf0\x9f\x91\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'two_women_holding_hands',
        b'\xf0\x9f\x91\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'u5272',
        b'\xf0\x9f\x88\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'u5408',
        b'\xf0\x9f\x88\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'u55b6',
        b'\xf0\x9f\x88\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'u6307',
        b'\xf0\x9f\x88\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'u6708',
        b'\xf0\x9f\x88\xb7',
        True,
        None,
        None,
    ),
    Unicode(
        'u6709',
        b'\xf0\x9f\x88\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'u6e80',
        b'\xf0\x9f\x88\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'u7121',
        b'\xf0\x9f\x88\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'u7533',
        b'\xf0\x9f\x88\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'u7981',
        b'\xf0\x9f\x88\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'u7a7a',
        b'\xf0\x9f\x88\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'umbrella2',
        b'\xe2\x98\x82',
        True,
        None,
        None,
    ),
    Unicode(
        'umbrella',
        b'\xe2\x98\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'underage',
        b'\xf0\x9f\x94\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'unicorn',
        b'\xf0\x9f\xa6\x84',
        False,
        (
            'unicorn_face',
        ),
        None,
    ),
    Unicode(
        'unlock',
        b'\xf0\x9f\x94\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'up',
        b'\xf0\x9f\x86\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'upside_down',
        b'\xf0\x9f\x99\x83',
        False,
        (
            'upside_down_face',
        ),
        None,
    ),
    Unicode(
        'v',
        b'\xe2\x9c\x8c',
        True,
        None,
        None,
    ),
    Unicode(
        'v_tone1',
        b'\xe2\x9c\x8c\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'v_tone2',
        b'\xe2\x9c\x8c\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'v_tone3',
        b'\xe2\x9c\x8c\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'v_tone4',
        b'\xe2\x9c\x8c\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'v_tone5',
        b'\xe2\x9c\x8c\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'vertical_traffic_light',
        b'\xf0\x9f\x9a\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'vhs',
        b'\xf0\x9f\x93\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'vibration_mode',
        b'\xf0\x9f\x93\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'video_camera',
        b'\xf0\x9f\x93\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'video_game',
        b'\xf0\x9f\x8e\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'violin',
        b'\xf0\x9f\x8e\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'virgo',
        b'\xe2\x99\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'volcano',
        b'\xf0\x9f\x8c\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'volleyball',
        b'\xf0\x9f\x8f\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'vs',
        b'\xf0\x9f\x86\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'walking',
        b'\xf0\x9f\x9a\xb6',
        False,
        (
            'person_walking',
        ),
        None,
    ),
    Unicode(
        'walking_tone1',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbb',
        False,
        (
            'person_walking_tone1',
        ),
        None,
    ),
    Unicode(
        'walking_tone2',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbc',
        False,
        (
            'person_walking_tone2',
        ),
        None,
    ),
    Unicode(
        'walking_tone3',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbd',
        False,
        (
            'person_walking_tone3',
        ),
        None,
    ),
    Unicode(
        'walking_tone4',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbe',
        False,
        (
            'person_walking_tone4',
        ),
        None,
    ),
    Unicode(
        'walking_tone5',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbf',
        False,
        (
            'person_walking_tone5',
        ),
        None,
    ),
    Unicode(
        'waning_crescent_moon',
        b'\xf0\x9f\x8c\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'waning_gibbous_moon',
        b'\xf0\x9f\x8c\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'warning',
        b'\xe2\x9a\xa0',
        True,
        None,
        None,
    ),
    Unicode(
        'wastebasket',
        b'\xf0\x9f\x97\x91',
        True,
        None,
        None,
    ),
    Unicode(
        'watch',
        b'\xe2\x8c\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'water_buffalo',
        b'\xf0\x9f\x90\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'water_polo',
        b'\xf0\x9f\xa4\xbd',
        False,
        (
            'person_playing_water_polo',
        ),
        None,
    ),
    Unicode(
        'water_polo_tone1',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbb',
        False,
        (
            'person_playing_water_polo_tone1',
        ),
        None,
    ),
    Unicode(
        'water_polo_tone2',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbc',
        False,
        (
            'person_playing_water_polo_tone2',
        ),
        None,
    ),
    Unicode(
        'water_polo_tone3',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbd',
        False,
        (
            'person_playing_water_polo_tone3',
        ),
        None,
    ),
    Unicode(
        'water_polo_tone4',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbe',
        False,
        (
            'person_playing_water_polo_tone4',
        ),
        None,
    ),
    Unicode(
        'water_polo_tone5',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbf',
        False,
        (
            'person_playing_water_polo_tone5',
        ),
        None,
    ),
    Unicode(
        'watermelon',
        b'\xf0\x9f\x8d\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'wave',
        b'\xf0\x9f\x91\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'wave_tone1',
        b'\xf0\x9f\x91\x8b\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'wave_tone2',
        b'\xf0\x9f\x91\x8b\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'wave_tone3',
        b'\xf0\x9f\x91\x8b\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'wave_tone4',
        b'\xf0\x9f\x91\x8b\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'wave_tone5',
        b'\xf0\x9f\x91\x8b\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'wavy_dash',
        b'\xe3\x80\xb0',
        True,
        None,
        None,
    ),
    Unicode(
        'waxing_crescent_moon',
        b'\xf0\x9f\x8c\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'waxing_gibbous_moon',
        b'\xf0\x9f\x8c\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'wc',
        b'\xf0\x9f\x9a\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'weary',
        b'\xf0\x9f\x98\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'wedding',
        b'\xf0\x9f\x92\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'whale2',
        b'\xf0\x9f\x90\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'whale',
        b'\xf0\x9f\x90\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'wheel_of_dharma',
        b'\xe2\x98\xb8',
        True,
        None,
        None,
    ),
    Unicode(
        'wheelchair',
        b'\xe2\x99\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'white_check_mark',
        b'\xe2\x9c\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'white_circle',
        b'\xe2\x9a\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'white_flower',
        b'\xf0\x9f\x92\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'white_large_square',
        b'\xe2\xac\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'white_medium_small_square',
        b'\xe2\x97\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'white_medium_square',
        b'\xe2\x97\xbb',
        True,
        None,
        None,
    ),
    Unicode(
        'white_small_square',
        b'\xe2\x96\xab',
        True,
        None,
        None,
    ),
    Unicode(
        'white_square_button',
        b'\xf0\x9f\x94\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'white_sun_behind_cloud',
        b'\xf0\x9f\x8c\xa5',
        True,
        None,
        None,
    ),
    Unicode(
        'white_sun_behind_cloud_with_rain',
        b'\xf0\x9f\x8c\xa6',
        True,
        None,
        None,
    ),
    Unicode(
        'white_sun_small_cloud',
        b'\xf0\x9f\x8c\xa4',
        True,
        None,
        None,
    ),
    Unicode(
        'wilted_flower',
        b'\xf0\x9f\xa5\x80',
        False,
        (
            'wilted_rose',
        ),
        None,
    ),
    Unicode(
        'wind_blowing_face',
        b'\xf0\x9f\x8c\xac',
        True,
        None,
        None,
    ),
    Unicode(
        'wind_chime',
        b'\xf0\x9f\x8e\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'wine_glass',
        b'\xf0\x9f\x8d\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'wink',
        b'\xf0\x9f\x98\x89',
        False,
        None,
        (';)', ';-)'),
    ),
    Unicode(
        'wolf',
        b'\xf0\x9f\x90\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'woman',
        b'\xf0\x9f\x91\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'womans_clothes',
        b'\xf0\x9f\x91\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'womans_hat',
        b'\xf0\x9f\x91\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'womens',
        b'\xf0\x9f\x9a\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'worried',
        b'\xf0\x9f\x98\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'wrench',
        b'\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'wrestlers',
        b'\xf0\x9f\xa4\xbc',
        False,
        (
            'wrestling',
            'people_wrestling'
        ),
        None,
    ),
    Unicode(
        'writing_hand',
        b'\xe2\x9c\x8d',
        True,
        None,
        None,
    ),
    Unicode(
        'writing_hand_tone1',
        b'\xe2\x9c\x8d\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'writing_hand_tone2',
        b'\xe2\x9c\x8d\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'writing_hand_tone3',
        b'\xe2\x9c\x8d\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'writing_hand_tone4',
        b'\xe2\x9c\x8d\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'writing_hand_tone5',
        b'\xe2\x9c\x8d\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'x',
        b'\xe2\x9d\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'yellow_heart',
        b'\xf0\x9f\x92\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'yen',
        b'\xf0\x9f\x92\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'yin_yang',
        b'\xe2\x98\xaf',
        True,
        None,
        None,
    ),
    Unicode(
        'yum',
        b'\xf0\x9f\x98\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'zap',
        b'\xe2\x9a\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'zero',
        b'0\xe2\x83\xa3',
        True,
        None,
        None,
    ),
    Unicode(
        'zipper_mouth',
        b'\xf0\x9f\xa4\x90',
        False,
        (
            'zipper_mouth_face',
        ),
        None,
    ),
    Unicode(
        'zzz',
        b'\xf0\x9f\x92\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'a',
        b'\xf0\x9f\x85\xb0\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'abacus',
        b'\xf0\x9f\xa7\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'adhesive_bandage',
        b'\xf0\x9f\xa9\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'admission_tickets',
        b'\xf0\x9f\x8e\x9f\xef\xb8\x8f',
        False,
        (
            'tickets',
        ),
        None,
    ),
    Unicode(
        'adult',
        b'\xf0\x9f\xa7\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'adult_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'adult_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'adult_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'adult_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'adult_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'airplane',
        b'\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'airplane_small',
        b'\xf0\x9f\x9b\xa9\xef\xb8\x8f',
        False,
        (
            'small_airplane',
        ),
        None,
    ),
    Unicode(
        'alembic',
        b'\xe2\x9a\x97\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'anger_right',
        b'\xf0\x9f\x97\xaf\xef\xb8\x8f',
        False,
        (
            'right_anger_bubble',
        ),
        None,
    ),
    Unicode(
        'arrow_backward',
        b'\xe2\x97\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_down',
        b'\xe2\xac\x87\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_forward',
        b'\xe2\x96\xb6\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_heading_down',
        b'\xe2\xa4\xb5\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_heading_up',
        b'\xe2\xa4\xb4\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_left',
        b'\xe2\xac\x85\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_lower_left',
        b'\xe2\x86\x99\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_lower_right',
        b'\xe2\x86\x98\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_right',
        b'\xe2\x9e\xa1\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_right_hook',
        b'\xe2\x86\xaa\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_up',
        b'\xe2\xac\x86\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_up_down',
        b'\xe2\x86\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_upper_left',
        b'\xe2\x86\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'arrow_upper_right',
        b'\xe2\x86\x97\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'asterisk',
        b'*\xef\xb8\x8f\xe2\x83\xa3',
        False,
        (
            'keycap_asterisk',
        ),
        None,
    ),
    Unicode(
        'atom',
        b'\xe2\x9a\x9b\xef\xb8\x8f',
        False,
        (
            'atom_symbol',
        ),
        None,
    ),
    Unicode(
        'auto_rickshaw',
        b'\xf0\x9f\x9b\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'axe',
        b'\xf0\x9f\xaa\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'b',
        b'\xf0\x9f\x85\xb1\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'badger',
        b'\xf0\x9f\xa6\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'bagel',
        b'\xf0\x9f\xa5\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'ballet_shoes',
        b'\xf0\x9f\xa9\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'ballot_box',
        b'\xf0\x9f\x97\xb3\xef\xb8\x8f',
        False,
        (
            'ballot_box_with_ballot',
        ),
        None,
    ),
    Unicode(
        'ballot_box_with_check',
        b'\xe2\x98\x91\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'bangbang',
        b'\xe2\x80\xbc\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'banjo',
        b'\xf0\x9f\xaa\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'basket',
        b'\xf0\x9f\xa7\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'basketball_player',
        b'\xe2\x9b\xb9\xef\xb8\x8f',
        False,
        (
            'person_bouncing_ball',
            'person_with_ball'
        ),
        None,
    ),
    Unicode(
        'beach',
        b'\xf0\x9f\x8f\x96\xef\xb8\x8f',
        False,
        (
            'beach_with_umbrella',
        ),
        None,
    ),
    Unicode(
        'beach_umbrella',
        b'\xe2\x9b\xb1\xef\xb8\x8f',
        False,
        (
            'umbrella_on_ground',
        ),
        None,
    ),
    Unicode(
        'bearded_person',
        b'\xf0\x9f\xa7\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'bearded_person_tone1',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'bearded_person_tone2',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'bearded_person_tone3',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'bearded_person_tone4',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'bearded_person_tone5',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'bed',
        b'\xf0\x9f\x9b\x8f\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'bellhop',
        b'\xf0\x9f\x9b\x8e\xef\xb8\x8f',
        False,
        (
            'bellhop_bell',
        ),
        None,
    ),
    Unicode(
        'beverage_box',
        b'\xf0\x9f\xa7\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'billed_cap',
        b'\xf0\x9f\xa7\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'biohazard',
        b'\xe2\x98\xa3\xef\xb8\x8f',
        False,
        (
            'biohazard_sign',
        ),
        None,
    ),
    Unicode(
        'black_medium_square',
        b'\xe2\x97\xbc\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'black_nib',
        b'\xe2\x9c\x92\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'black_small_square',
        b'\xe2\x96\xaa\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_man',
        b'\xf0\x9f\x91\xb1\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_man_tone1',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_man_tone2',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_man_tone3',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_man_tone4',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_man_tone5',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_woman',
        b'\xf0\x9f\x91\xb1\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_woman_tone1',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_woman_tone2',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_woman_tone3',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_woman_tone4',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blond_haired_woman_tone5',
        b'\xf0\x9f\x91\xb1\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'blue_square',
        b'\xf0\x9f\x9f\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'bone',
        b'\xf0\x9f\xa6\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'bowl_with_spoon',
        b'\xf0\x9f\xa5\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'brain',
        b'\xf0\x9f\xa7\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'breast_feeding',
        b'\xf0\x9f\xa4\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'breast_feeding_tone1',
        b'\xf0\x9f\xa4\xb1\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'breast_feeding_tone2',
        b'\xf0\x9f\xa4\xb1\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'breast_feeding_tone3',
        b'\xf0\x9f\xa4\xb1\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'breast_feeding_tone4',
        b'\xf0\x9f\xa4\xb1\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'breast_feeding_tone5',
        b'\xf0\x9f\xa4\xb1\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'bricks',
        b'\xf0\x9f\xa7\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'briefs',
        b'\xf0\x9f\xa9\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'broccoli',
        b'\xf0\x9f\xa5\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'broom',
        b'\xf0\x9f\xa7\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'brown_circle',
        b'\xf0\x9f\x9f\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'brown_heart',
        b'\xf0\x9f\xa4\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'brown_square',
        b'\xf0\x9f\x9f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'building_construction',
        b'\xf0\x9f\x8f\x97\xef\xb8\x8f',
        False,
        (
            'construction_site',
        ),
        None,
    ),
    Unicode(
        'butter',
        b'\xf0\x9f\xa7\x88',
        False,
        None,
        None,
    ),
    Unicode(
        'calendar_spiral',
        b'\xf0\x9f\x97\x93\xef\xb8\x8f',
        False,
        (
            'spiral_calendar_pad',
        ),
        None,
    ),
    Unicode(
        'camping',
        b'\xf0\x9f\x8f\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'candle',
        b'\xf0\x9f\x95\xaf\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'canned_food',
        b'\xf0\x9f\xa5\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'card_box',
        b'\xf0\x9f\x97\x83\xef\xb8\x8f',
        False,
        (
            'card_file_box',
        ),
        None,
    ),
    Unicode(
        'card_index_dividers',
        b'\xf0\x9f\x97\x82\xef\xb8\x8f',
        False,
        (
            'dividers',
        ),
        None,
    ),
    Unicode(
        'chains',
        b'\xe2\x9b\x93\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'chair',
        b'\xf0\x9f\xaa\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'chess_pawn',
        b'\xe2\x99\x9f\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'child',
        b'\xf0\x9f\xa7\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'child_tone1',
        b'\xf0\x9f\xa7\x92\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'child_tone2',
        b'\xf0\x9f\xa7\x92\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'child_tone3',
        b'\xf0\x9f\xa7\x92\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'child_tone4',
        b'\xf0\x9f\xa7\x92\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'child_tone5',
        b'\xf0\x9f\xa7\x92\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'chipmunk',
        b'\xf0\x9f\x90\xbf\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'chopsticks',
        b'\xf0\x9f\xa5\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'cityscape',
        b'\xf0\x9f\x8f\x99\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'classical_building',
        b'\xf0\x9f\x8f\x9b\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'clock',
        b'\xf0\x9f\x95\xb0\xef\xb8\x8f',
        False,
        (
            'mantlepiece_clock',
        ),
        None,
    ),
    Unicode(
        'cloud',
        b'\xe2\x98\x81\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'cloud_lightning',
        b'\xf0\x9f\x8c\xa9\xef\xb8\x8f',
        False,
        (
            'cloud_with_lightning',
        ),
        None,
    ),
    Unicode(
        'cloud_rain',
        b'\xf0\x9f\x8c\xa7\xef\xb8\x8f',
        False,
        (
            'cloud_with_rain',
        ),
        None,
    ),
    Unicode(
        'cloud_snow',
        b'\xf0\x9f\x8c\xa8\xef\xb8\x8f',
        False,
        (
            'cloud_with_snow',
        ),
        None,
    ),
    Unicode(
        'cloud_tornado',
        b'\xf0\x9f\x8c\xaa\xef\xb8\x8f',
        False,
        (
            'cloud_with_tornado',
        ),
        None,
    ),
    Unicode(
        'clubs',
        b'\xe2\x99\xa3\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'coat',
        b'\xf0\x9f\xa7\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'coconut',
        b'\xf0\x9f\xa5\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'coffin',
        b'\xe2\x9a\xb0\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'cold_face',
        b'\xf0\x9f\xa5\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'comet',
        b'\xe2\x98\x84\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'compass',
        b'\xf0\x9f\xa7\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'compression',
        b'\xf0\x9f\x97\x9c\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'congratulations',
        b'\xe3\x8a\x97\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'control_knobs',
        b'\xf0\x9f\x8e\x9b\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'copyright',
        b'\xc2\xa9\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'couch',
        b'\xf0\x9f\x9b\x8b\xef\xb8\x8f',
        False,
        (
            'couch_and_lamp',
        ),
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'crayon',
        b'\xf0\x9f\x96\x8d\xef\xb8\x8f',
        False,
        (
            'lower_left_crayon',
        ),
        None,
    ),
    Unicode(
        'cricket',
        b'\xf0\x9f\xa6\x97',
        False,
        (
            'cricket_game',
            'cricket_bat_ball'
        ),
        None,
    ),
    Unicode(
        'cross',
        b'\xe2\x9c\x9d\xef\xb8\x8f',
        False,
        (
            'latin_cross',
        ),
        None,
    ),
    Unicode(
        'crossed_swords',
        b'\xe2\x9a\x94\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'cruise_ship',
        b'\xf0\x9f\x9b\xb3\xef\xb8\x8f',
        False,
        (
            'passenger_ship',
        ),
        None,
    ),
    Unicode(
        'cup_with_straw',
        b'\xf0\x9f\xa5\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'cupcake',
        b'\xf0\x9f\xa7\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'curling_stone',
        b'\xf0\x9f\xa5\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'cut_of_meat',
        b'\xf0\x9f\xa5\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'dagger',
        b'\xf0\x9f\x97\xa1\xef\xb8\x8f',
        False,
        (
            'dagger_knife',
        ),
        None,
    ),
    Unicode(
        'dark_sunglasses',
        b'\xf0\x9f\x95\xb6\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_man',
        b'\xf0\x9f\xa7\x8f\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_man_tone1',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_man_tone2',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_man_tone3',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_man_tone4',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_man_tone5',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_person',
        b'\xf0\x9f\xa7\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_person_tone1',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_person_tone2',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_person_tone3',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_person_tone4',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_person_tone5',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_woman',
        b'\xf0\x9f\xa7\x8f\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_woman_tone1',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_woman_tone2',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_woman_tone3',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_woman_tone4',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'deaf_woman_tone5',
        b'\xf0\x9f\xa7\x8f\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'derelict_house_building',
        b'\xf0\x9f\x8f\x9a\xef\xb8\x8f',
        False,
        (
            'house_abandoned',
        ),
        None,
    ),
    Unicode(
        'desert',
        b'\xf0\x9f\x8f\x9c\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'desert_island',
        b'\xf0\x9f\x8f\x9d\xef\xb8\x8f',
        False,
        (
            'island',
        ),
        None,
    ),
    Unicode(
        'desktop',
        b'\xf0\x9f\x96\xa5\xef\xb8\x8f',
        False,
        (
            'desktop_computer',
        ),
        None,
    ),
    Unicode(
        'detective',
        b'\xf0\x9f\x95\xb5\xef\xb8\x8f',
        False,
        (
            'sleuth_or_spy',
            'spy'
        ),
        None,
    ),
    Unicode(
        'diamonds',
        b'\xe2\x99\xa6\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'diving_mask',
        b'\xf0\x9f\xa4\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'diya_lamp',
        b'\xf0\x9f\xaa\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'dna',
        b'\xf0\x9f\xa7\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'double_vertical_bar',
        b'\xe2\x8f\xb8\xef\xb8\x8f',
        False,
        (
            'pause_button',
        ),
        None,
    ),
    Unicode(
        'dove',
        b'\xf0\x9f\x95\x8a\xef\xb8\x8f',
        False,
        (
            'dove_of_peace',
        ),
        None,
    ),
    Unicode(
        'drop_of_blood',
        b'\xf0\x9f\xa9\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'dumpling',
        b'\xf0\x9f\xa5\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_with_hearing_aid',
        b'\xf0\x9f\xa6\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_with_hearing_aid_tone1',
        b'\xf0\x9f\xa6\xbb\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_with_hearing_aid_tone2',
        b'\xf0\x9f\xa6\xbb\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_with_hearing_aid_tone3',
        b'\xf0\x9f\xa6\xbb\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_with_hearing_aid_tone4',
        b'\xf0\x9f\xa6\xbb\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'ear_with_hearing_aid_tone5',
        b'\xf0\x9f\xa6\xbb\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'eight',
        b'8\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'eight_pointed_black_star',
        b'\xe2\x9c\xb4\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'eight_spoked_asterisk',
        b'\xe2\x9c\xb3\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'eject',
        b'\xe2\x8f\x8f\xef\xb8\x8f',
        False,
        (
            'eject_symbol',
        ),
        None,
    ),
    Unicode(
        'elf',
        b'\xf0\x9f\xa7\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'elf_tone1',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'elf_tone2',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'elf_tone3',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'elf_tone4',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'elf_tone5',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'england',
        (
            b'\xf0\x9f\x8f\xb4\xf3\xa0\x81\xa7\xf3\xa0\x81\xa2\xf3\xa0'
            b'\x81\xa5\xf3\xa0\x81\xae\xf3\xa0\x81\xa7\xf3\xa0\x81\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'envelope',
        b'\xe2\x9c\x89\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'exploding_head',
        b'\xf0\x9f\xa4\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'eye',
        b'\xf0\x9f\x91\x81\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'face_vomiting',
        b'\xf0\x9f\xa4\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'face_with_hand_over_mouth',
        b'\xf0\x9f\xa4\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'face_with_monocle',
        b'\xf0\x9f\xa7\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'face_with_raised_eyebrow',
        b'\xf0\x9f\xa4\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'face_with_symbols_over_mouth',
        b'\xf0\x9f\xa4\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'fairy',
        b'\xf0\x9f\xa7\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'fairy_tone1',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'fairy_tone2',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'fairy_tone3',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'fairy_tone4',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'fairy_tone5',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'falafel',
        b'\xf0\x9f\xa7\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'family_man_boy',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'family_man_boy_boy',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa6\xe2\x80\x8d\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'family_man_girl',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'family_man_girl_boy',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'family_man_girl_girl',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'family_man_woman_boy',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'family_woman_boy',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'family_woman_boy_boy',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa6\xe2\x80\x8d\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'family_woman_girl',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'family_woman_girl_boy',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'family_woman_girl_girl',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x91\xa7\xe2\x80\x8d\xf0\x9f\x91\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'female_sign',
        b'\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'ferry',
        b'\xe2\x9b\xb4\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'file_cabinet',
        b'\xf0\x9f\x97\x84\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'film_frames',
        b'\xf0\x9f\x8e\x9e\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'film_projector',
        b'\xf0\x9f\x93\xbd\xef\xb8\x8f',
        False,
        (
            'projector',
        ),
        None,
    ),
    Unicode(
        'fire_extinguisher',
        b'\xf0\x9f\xa7\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'firecracker',
        b'\xf0\x9f\xa7\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'five',
        b'5\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'flag_white',
        b'\xf0\x9f\x8f\xb3\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'flamingo',
        b'\xf0\x9f\xa6\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'fleur_de_lis',
        b'\xe2\x9a\x9c\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'flying_disc',
        b'\xf0\x9f\xa5\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'flying_saucer',
        b'\xf0\x9f\x9b\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'fog',
        b'\xf0\x9f\x8c\xab\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'foot',
        b'\xf0\x9f\xa6\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'foot_tone1',
        b'\xf0\x9f\xa6\xb6\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'foot_tone2',
        b'\xf0\x9f\xa6\xb6\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'foot_tone3',
        b'\xf0\x9f\xa6\xb6\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'foot_tone4',
        b'\xf0\x9f\xa6\xb6\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'foot_tone5',
        b'\xf0\x9f\xa6\xb6\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'fork_and_knife_with_plate',
        b'\xf0\x9f\x8d\xbd\xef\xb8\x8f',
        False,
        (
            'fork_knife_plate',
        ),
        None,
    ),
    Unicode(
        'fortune_cookie',
        b'\xf0\x9f\xa5\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'four',
        b'4\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'frame_photo',
        b'\xf0\x9f\x96\xbc\xef\xb8\x8f',
        False,
        (
            'frame_with_picture',
        ),
        None,
    ),
    Unicode(
        'frowning2',
        b'\xe2\x98\xb9\xef\xb8\x8f',
        False,
        (
            'white_frowning_face',
        ),
        None,
    ),
    Unicode(
        'funeral_urn',
        b'\xe2\x9a\xb1\xef\xb8\x8f',
        False,
        (
            'urn',
        ),
        None,
    ),
    Unicode(
        'garlic',
        b'\xf0\x9f\xa7\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'gear',
        b'\xe2\x9a\x99\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'genie',
        b'\xf0\x9f\xa7\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'giraffe',
        b'\xf0\x9f\xa6\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'gloves',
        b'\xf0\x9f\xa7\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'goggles',
        b'\xf0\x9f\xa5\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'golfer',
        b'\xf0\x9f\x8f\x8c\xef\xb8\x8f',
        False,
        (
            'person_golfing',
        ),
        None,
    ),
    Unicode(
        'green_circle',
        b'\xf0\x9f\x9f\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'green_square',
        b'\xf0\x9f\x9f\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'guide_dog',
        b'\xf0\x9f\xa6\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'hammer_and_pick',
        b'\xe2\x9a\x92\xef\xb8\x8f',
        False,
        (
            'hammer_pick',
        ),
        None,
    ),
    Unicode(
        'hammer_and_wrench',
        b'\xf0\x9f\x9b\xa0\xef\xb8\x8f',
        False,
        (
            'tools',
        ),
        None,
    ),
    Unicode(
        'hand_splayed',
        b'\xf0\x9f\x96\x90\xef\xb8\x8f',
        False,
        (
            'raised_hand_with_fingers_splayed',
        ),
        None,
    ),
    Unicode(
        'hash',
        b'#\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'heart',
        b'\xe2\x9d\xa4\xef\xb8\x8f',
        False,
        None,
        ('<3', '♡'),
    ),
    Unicode(
        'heart_exclamation',
        b'\xe2\x9d\xa3\xef\xb8\x8f',
        False,
        (
            'heavy_heart_exclamation_mark_ornament',
        ),
        None,
    ),
    Unicode(
        'hearts',
        b'\xe2\x99\xa5\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'heavy_check_mark',
        b'\xe2\x9c\x94\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'heavy_multiplication_x',
        b'\xe2\x9c\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'hedgehog',
        b'\xf0\x9f\xa6\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'helmet_with_cross',
        b'\xe2\x9b\x91\xef\xb8\x8f',
        False,
        (
            'helmet_with_white_cross',
        ),
        None,
    ),
    Unicode(
        'hiking_boot',
        b'\xf0\x9f\xa5\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'hindu_temple',
        b'\xf0\x9f\x9b\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'hippopotamus',
        b'\xf0\x9f\xa6\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'hole',
        b'\xf0\x9f\x95\xb3\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'homes',
        b'\xf0\x9f\x8f\x98\xef\xb8\x8f',
        False,
        (
            'house_buildings',
        ),
        None,
    ),
    Unicode(
        'hot_face',
        b'\xf0\x9f\xa5\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'hot_pepper',
        b'\xf0\x9f\x8c\xb6\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'hotsprings',
        b'\xe2\x99\xa8\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'ice_cube',
        b'\xf0\x9f\xa7\x8a',
        False,
        None,
        None,
    ),
    Unicode(
        'ice_skate',
        b'\xe2\x9b\xb8\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'infinity',
        b'\xe2\x99\xbe\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'information_source',
        b'\xe2\x84\xb9\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'interrobang',
        b'\xe2\x81\x89\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'jigsaw',
        b'\xf0\x9f\xa7\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'joystick',
        b'\xf0\x9f\x95\xb9\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'kangaroo',
        b'\xf0\x9f\xa6\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'key2',
        b'\xf0\x9f\x97\x9d\xef\xb8\x8f',
        False,
        (
            'old_key',
        ),
        None,
    ),
    Unicode(
        'keyboard',
        b'\xe2\x8c\xa8\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man',
        (
            b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kite',
        b'\xf0\x9f\xaa\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'lab_coat',
        b'\xf0\x9f\xa5\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'label',
        b'\xf0\x9f\x8f\xb7\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'lacrosse',
        b'\xf0\x9f\xa5\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'leafy_green',
        b'\xf0\x9f\xa5\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'left_right_arrow',
        b'\xe2\x86\x94\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'left_speech_bubble',
        b'\xf0\x9f\x97\xa8\xef\xb8\x8f',
        False,
        (
            'speech_left',
        ),
        None,
    ),
    Unicode(
        'leftwards_arrow_with_hook',
        b'\xe2\x86\xa9\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'leg',
        b'\xf0\x9f\xa6\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'leg_tone1',
        b'\xf0\x9f\xa6\xb5\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'leg_tone2',
        b'\xf0\x9f\xa6\xb5\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'leg_tone3',
        b'\xf0\x9f\xa6\xb5\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'leg_tone4',
        b'\xf0\x9f\xa6\xb5\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'leg_tone5',
        b'\xf0\x9f\xa6\xb5\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'level_slider',
        b'\xf0\x9f\x8e\x9a\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'levitate',
        b'\xf0\x9f\x95\xb4\xef\xb8\x8f',
        False,
        (
            'man_in_business_suit_levitating',
        ),
        None,
    ),
    Unicode(
        'lifter',
        b'\xf0\x9f\x8f\x8b\xef\xb8\x8f',
        False,
        (
            'person_lifting_weights',
            'weight_lifter'
        ),
        None,
    ),
    Unicode(
        'linked_paperclips',
        b'\xf0\x9f\x96\x87\xef\xb8\x8f',
        False,
        (
            'paperclips',
        ),
        None,
    ),
    Unicode(
        'llama',
        b'\xf0\x9f\xa6\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'lobster',
        b'\xf0\x9f\xa6\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'love_you_gesture',
        b'\xf0\x9f\xa4\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'love_you_gesture_tone1',
        b'\xf0\x9f\xa4\x9f\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'love_you_gesture_tone2',
        b'\xf0\x9f\xa4\x9f\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'love_you_gesture_tone3',
        b'\xf0\x9f\xa4\x9f\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'love_you_gesture_tone4',
        b'\xf0\x9f\xa4\x9f\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'love_you_gesture_tone5',
        b'\xf0\x9f\xa4\x9f\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'lower_left_ballpoint_pen',
        b'\xf0\x9f\x96\x8a\xef\xb8\x8f',
        False,
        (
            'pen_ballpoint',
        ),
        None,
    ),
    Unicode(
        'lower_left_fountain_pen',
        b'\xf0\x9f\x96\x8b\xef\xb8\x8f',
        False,
        (
            'pen_fountain',
        ),
        None,
    ),
    Unicode(
        'lower_left_paintbrush',
        b'\xf0\x9f\x96\x8c\xef\xb8\x8f',
        False,
        (
            'paintbrush',
        ),
        None,
    ),
    Unicode(
        'luggage',
        b'\xf0\x9f\xa7\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'm',
        b'\xe2\x93\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mage',
        b'\xf0\x9f\xa7\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'mage_tone1',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'mage_tone2',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'mage_tone3',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'mage_tone4',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'mage_tone5',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'magnet',
        b'\xf0\x9f\xa7\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'male_sign',
        b'\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_artist',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'man_artist_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'man_artist_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'man_artist_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'man_artist_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'man_artist_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'man_astronaut',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'man_astronaut_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'man_astronaut_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'man_astronaut_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'man_astronaut_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'man_astronaut_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bald',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bald_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bald_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bald_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bald_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bald_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'man_biking',
        b'\xf0\x9f\x9a\xb4\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_biking_tone1',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_biking_tone2',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_biking_tone3',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_biking_tone4',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_biking_tone5',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bouncing_ball',
        b'\xe2\x9b\xb9\xef\xb8\x8f\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bouncing_ball_tone1',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bouncing_ball_tone2',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bouncing_ball_tone3',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bouncing_ball_tone4',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bouncing_ball_tone5',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bowing',
        b'\xf0\x9f\x99\x87\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bowing_tone1',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bowing_tone2',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bowing_tone3',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bowing_tone4',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_bowing_tone5',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cartwheeling',
        b'\xf0\x9f\xa4\xb8\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cartwheeling_tone1',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cartwheeling_tone2',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cartwheeling_tone3',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cartwheeling_tone4',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cartwheeling_tone5',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_climbing',
        b'\xf0\x9f\xa7\x97\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_climbing_tone1',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_climbing_tone2',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_climbing_tone3',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_climbing_tone4',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_climbing_tone5',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_construction_worker',
        b'\xf0\x9f\x91\xb7\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_construction_worker_tone1',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_construction_worker_tone2',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_construction_worker_tone3',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_construction_worker_tone4',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_construction_worker_tone5',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cook',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cook_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cook_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cook_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cook_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_cook_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_curly_haired',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'man_curly_haired_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'man_curly_haired_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'man_curly_haired_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'man_curly_haired_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'man_curly_haired_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'man_detective',
        b'\xf0\x9f\x95\xb5\xef\xb8\x8f\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_detective_tone1',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_detective_tone2',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_detective_tone3',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_detective_tone4',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_detective_tone5',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_elf',
        b'\xf0\x9f\xa7\x9d\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_elf_tone1',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_elf_tone2',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_elf_tone3',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_elf_tone4',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_elf_tone5',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_facepalming',
        b'\xf0\x9f\xa4\xa6\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_facepalming_tone1',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_facepalming_tone2',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_facepalming_tone3',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_facepalming_tone4',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_facepalming_tone5',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_factory_worker',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'man_factory_worker_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'man_factory_worker_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'man_factory_worker_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'man_factory_worker_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'man_factory_worker_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'man_fairy',
        b'\xf0\x9f\xa7\x9a\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_fairy_tone1',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_fairy_tone2',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_fairy_tone3',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_fairy_tone4',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_fairy_tone5',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_farmer',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'man_farmer_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'man_farmer_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'man_farmer_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'man_farmer_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'man_farmer_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'man_firefighter',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'man_firefighter_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'man_firefighter_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'man_firefighter_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'man_firefighter_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'man_firefighter_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'man_frowning',
        b'\xf0\x9f\x99\x8d\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_frowning_tone1',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_frowning_tone2',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_frowning_tone3',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_frowning_tone4',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_frowning_tone5',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_genie',
        b'\xf0\x9f\xa7\x9e\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_no',
        b'\xf0\x9f\x99\x85\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_no_tone1',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_no_tone2',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_no_tone3',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_no_tone4',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_no_tone5',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_ok',
        b'\xf0\x9f\x99\x86\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_ok_tone1',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_ok_tone2',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_ok_tone3',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_ok_tone4',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_gesturing_ok_tone5',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_face_massage',
        b'\xf0\x9f\x92\x86\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_face_massage_tone1',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_face_massage_tone2',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_face_massage_tone3',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_face_massage_tone4',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_face_massage_tone5',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_haircut',
        b'\xf0\x9f\x92\x87\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_haircut_tone1',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_haircut_tone2',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_haircut_tone3',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_haircut_tone4',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_getting_haircut_tone5',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_golfing',
        b'\xf0\x9f\x8f\x8c\xef\xb8\x8f\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_golfing_tone1',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_golfing_tone2',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_golfing_tone3',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_golfing_tone4',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_golfing_tone5',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_guard',
        b'\xf0\x9f\x92\x82\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_guard_tone1',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_guard_tone2',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_guard_tone3',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_guard_tone4',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_guard_tone5',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_health_worker',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_health_worker_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_health_worker_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_health_worker_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_health_worker_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_health_worker_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_lotus_position',
        b'\xf0\x9f\xa7\x98\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_lotus_position_tone1',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_lotus_position_tone2',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_lotus_position_tone3',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_lotus_position_tone4',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_lotus_position_tone5',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_manual_wheelchair',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_manual_wheelchair_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_manual_wheelchair_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_manual_wheelchair_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_manual_wheelchair_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_manual_wheelchair_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_motorized_wheelchair',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_motorized_wheelchair_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_motorized_wheelchair_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_motorized_wheelchair_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_motorized_wheelchair_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_motorized_wheelchair_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_steamy_room',
        b'\xf0\x9f\xa7\x96\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_steamy_room_tone1',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_steamy_room_tone2',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_steamy_room_tone3',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_steamy_room_tone4',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_steamy_room_tone5',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_judge',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_judge_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_judge_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_judge_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_judge_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_judge_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_juggling',
        b'\xf0\x9f\xa4\xb9\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_juggling_tone1',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_juggling_tone2',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_juggling_tone3',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_juggling_tone4',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_juggling_tone5',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_kneeling',
        b'\xf0\x9f\xa7\x8e\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_kneeling_tone1',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_kneeling_tone2',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_kneeling_tone3',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_kneeling_tone4',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_kneeling_tone5',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_lifting_weights',
        b'\xf0\x9f\x8f\x8b\xef\xb8\x8f\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_lifting_weights_tone1',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_lifting_weights_tone2',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_lifting_weights_tone3',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_lifting_weights_tone4',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_lifting_weights_tone5',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mage',
        b'\xf0\x9f\xa7\x99\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mage_tone1',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mage_tone2',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mage_tone3',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mage_tone4',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mage_tone5',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mechanic',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mechanic_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mechanic_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mechanic_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mechanic_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mechanic_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mountain_biking',
        b'\xf0\x9f\x9a\xb5\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mountain_biking_tone1',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mountain_biking_tone2',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mountain_biking_tone3',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mountain_biking_tone4',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_mountain_biking_tone5',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_office_worker',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_office_worker_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_office_worker_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_office_worker_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_office_worker_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_office_worker_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pilot',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pilot_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pilot_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pilot_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pilot_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pilot_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_handball',
        b'\xf0\x9f\xa4\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_handball_tone1',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_handball_tone2',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_handball_tone3',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_handball_tone4',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_handball_tone5',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_water_polo',
        b'\xf0\x9f\xa4\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_water_polo_tone1',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_water_polo_tone2',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_water_polo_tone3',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_water_polo_tone4',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_playing_water_polo_tone5',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_police_officer',
        b'\xf0\x9f\x91\xae\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_police_officer_tone1',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_police_officer_tone2',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_police_officer_tone3',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_police_officer_tone4',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_police_officer_tone5',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pouting',
        b'\xf0\x9f\x99\x8e\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pouting_tone1',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pouting_tone2',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pouting_tone3',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pouting_tone4',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_pouting_tone5',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_raising_hand',
        b'\xf0\x9f\x99\x8b\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_raising_hand_tone1',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_raising_hand_tone2',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_raising_hand_tone3',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_raising_hand_tone4',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_raising_hand_tone5',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_red_haired',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'man_red_haired_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'man_red_haired_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'man_red_haired_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'man_red_haired_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'man_red_haired_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'man_rowing_boat',
        b'\xf0\x9f\x9a\xa3\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_rowing_boat_tone1',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_rowing_boat_tone2',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_rowing_boat_tone3',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_rowing_boat_tone4',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_rowing_boat_tone5',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_running',
        b'\xf0\x9f\x8f\x83\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_running_tone1',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_running_tone2',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_running_tone3',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_running_tone4',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_running_tone5',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_scientist',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'man_scientist_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'man_scientist_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'man_scientist_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'man_scientist_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'man_scientist_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'man_shrugging',
        b'\xf0\x9f\xa4\xb7\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_shrugging_tone1',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_shrugging_tone2',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_shrugging_tone3',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_shrugging_tone4',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_shrugging_tone5',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_singer',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'man_singer_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'man_singer_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'man_singer_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'man_singer_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'man_singer_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'man_standing',
        b'\xf0\x9f\xa7\x8d\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_standing_tone1',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_standing_tone2',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_standing_tone3',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_standing_tone4',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_standing_tone5',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_student',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'man_student_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'man_student_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'man_student_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'man_student_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'man_student_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'man_superhero',
        b'\xf0\x9f\xa6\xb8\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_superhero_tone1',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_superhero_tone2',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_superhero_tone3',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_superhero_tone4',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_superhero_tone5',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_supervillain',
        b'\xf0\x9f\xa6\xb9\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_supervillain_tone1',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_supervillain_tone2',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_supervillain_tone3',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_supervillain_tone4',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_supervillain_tone5',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_surfing',
        b'\xf0\x9f\x8f\x84\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_surfing_tone1',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_surfing_tone2',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_surfing_tone3',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_surfing_tone4',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_surfing_tone5',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_swimming',
        b'\xf0\x9f\x8f\x8a\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_swimming_tone1',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_swimming_tone2',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_swimming_tone3',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_swimming_tone4',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_swimming_tone5',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_teacher',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'man_teacher_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'man_teacher_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'man_teacher_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'man_teacher_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'man_teacher_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'man_technologist',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'man_technologist_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'man_technologist_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'man_technologist_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'man_technologist_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'man_technologist_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tipping_hand',
        b'\xf0\x9f\x92\x81\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tipping_hand_tone1',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tipping_hand_tone2',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tipping_hand_tone3',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tipping_hand_tone4',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_tipping_hand_tone5',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_vampire',
        b'\xf0\x9f\xa7\x9b\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_vampire_tone1',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_vampire_tone2',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_vampire_tone3',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_vampire_tone4',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_vampire_tone5',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_walking',
        b'\xf0\x9f\x9a\xb6\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_walking_tone1',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_walking_tone2',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_walking_tone3',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_walking_tone4',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_walking_tone5',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_wearing_turban',
        b'\xf0\x9f\x91\xb3\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_wearing_turban_tone1',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_wearing_turban_tone2',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_wearing_turban_tone3',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_wearing_turban_tone4',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_wearing_turban_tone5',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_white_haired',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_white_haired_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_white_haired_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_white_haired_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_white_haired_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_white_haired_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_probing_cane',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_probing_cane_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_probing_cane_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_probing_cane_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_probing_cane_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_probing_cane_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'man_zombie',
        b'\xf0\x9f\xa7\x9f\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mango',
        b'\xf0\x9f\xa5\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'manual_wheelchair',
        b'\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'map',
        b'\xf0\x9f\x97\xba\xef\xb8\x8f',
        False,
        (
            'world_map',
        ),
        None,
    ),
    Unicode(
        'mate',
        b'\xf0\x9f\xa7\x89',
        False,
        None,
        None,
    ),
    Unicode(
        'mechanical_arm',
        b'\xf0\x9f\xa6\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'mechanical_leg',
        b'\xf0\x9f\xa6\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'medical_symbol',
        b'\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'men_with_bunny_ears_partying',
        b'\xf0\x9f\x91\xaf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'men_wrestling',
        b'\xf0\x9f\xa4\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mermaid',
        b'\xf0\x9f\xa7\x9c\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mermaid_tone1',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mermaid_tone2',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mermaid_tone3',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mermaid_tone4',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mermaid_tone5',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'merman',
        b'\xf0\x9f\xa7\x9c\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'merman_tone1',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'merman_tone2',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'merman_tone3',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'merman_tone4',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'merman_tone5',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'merperson',
        b'\xf0\x9f\xa7\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'merperson_tone1',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'merperson_tone2',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'merperson_tone3',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'merperson_tone4',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'merperson_tone5',
        b'\xf0\x9f\xa7\x9c\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'microbe',
        b'\xf0\x9f\xa6\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'microphone2',
        b'\xf0\x9f\x8e\x99\xef\xb8\x8f',
        False,
        (
            'studio_microphone',
        ),
        None,
    ),
    Unicode(
        'military_medal',
        b'\xf0\x9f\x8e\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'moon_cake',
        b'\xf0\x9f\xa5\xae',
        False,
        None,
        None,
    ),
    Unicode(
        'mosquito',
        b'\xf0\x9f\xa6\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'motorboat',
        b'\xf0\x9f\x9b\xa5\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'motorcycle',
        b'\xf0\x9f\x8f\x8d\xef\xb8\x8f',
        False,
        (
            'racing_motorcycle',
        ),
        None,
    ),
    Unicode(
        'motorized_wheelchair',
        b'\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'motorway',
        b'\xf0\x9f\x9b\xa3\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mountain',
        b'\xe2\x9b\xb0\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'mountain_snow',
        b'\xf0\x9f\x8f\x94\xef\xb8\x8f',
        False,
        (
            'snow_capped_mountain',
        ),
        None,
    ),
    Unicode(
        'mouse_three_button',
        b'\xf0\x9f\x96\xb1\xef\xb8\x8f',
        False,
        (
            'three_button_mouse',
        ),
        None,
    ),
    Unicode(
        'national_park',
        b'\xf0\x9f\x8f\x9e\xef\xb8\x8f',
        False,
        (
            'park',
        ),
        None,
    ),
    Unicode(
        'nazar_amulet',
        b'\xf0\x9f\xa7\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'newspaper2',
        b'\xf0\x9f\x97\x9e\xef\xb8\x8f',
        False,
        (
            'rolled_up_newspaper',
        ),
        None,
    ),
    Unicode(
        'next_track',
        b'\xe2\x8f\xad\xef\xb8\x8f',
        False,
        (
            'track_next',
        ),
        None,
    ),
    Unicode(
        'nine',
        b'9\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'notepad_spiral',
        b'\xf0\x9f\x97\x92\xef\xb8\x8f',
        False,
        (
            'spiral_note_pad',
        ),
        None,
    ),
    Unicode(
        'o2',
        b'\xf0\x9f\x85\xbe\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'oil',
        b'\xf0\x9f\x9b\xa2\xef\xb8\x8f',
        False,
        (
            'oil_drum',
        ),
        None,
    ),
    Unicode(
        'older_adult',
        b'\xf0\x9f\xa7\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'older_adult_tone1',
        b'\xf0\x9f\xa7\x93\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'older_adult_tone2',
        b'\xf0\x9f\xa7\x93\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'older_adult_tone3',
        b'\xf0\x9f\xa7\x93\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'older_adult_tone4',
        b'\xf0\x9f\xa7\x93\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'older_adult_tone5',
        b'\xf0\x9f\xa7\x93\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'om_symbol',
        b'\xf0\x9f\x95\x89\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'one',
        b'1\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'one_piece_swimsuit',
        b'\xf0\x9f\xa9\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'onion',
        b'\xf0\x9f\xa7\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'orange_circle',
        b'\xf0\x9f\x9f\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'orange_heart',
        b'\xf0\x9f\xa7\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'orange_square',
        b'\xf0\x9f\x9f\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'orangutan',
        b'\xf0\x9f\xa6\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'orthodox_cross',
        b'\xe2\x98\xa6\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'otter',
        b'\xf0\x9f\xa6\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'oyster',
        b'\xf0\x9f\xa6\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'palms_up_together',
        b'\xf0\x9f\xa4\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'palms_up_together_tone1',
        b'\xf0\x9f\xa4\xb2\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'palms_up_together_tone2',
        b'\xf0\x9f\xa4\xb2\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'palms_up_together_tone3',
        b'\xf0\x9f\xa4\xb2\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'palms_up_together_tone4',
        b'\xf0\x9f\xa4\xb2\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'palms_up_together_tone5',
        b'\xf0\x9f\xa4\xb2\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'parachute',
        b'\xf0\x9f\xaa\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'parking',
        b'\xf0\x9f\x85\xbf\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'parrot',
        b'\xf0\x9f\xa6\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'part_alternation_mark',
        b'\xe3\x80\xbd\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'partying_face',
        b'\xf0\x9f\xa5\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'peace',
        b'\xe2\x98\xae\xef\xb8\x8f',
        False,
        (
            'peace_symbol',
        ),
        None,
    ),
    Unicode(
        'peacock',
        b'\xf0\x9f\xa6\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'pencil2',
        b'\xe2\x9c\x8f\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'person_climbing',
        b'\xf0\x9f\xa7\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'person_climbing_tone1',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'person_climbing_tone2',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_climbing_tone3',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'person_climbing_tone4',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'person_climbing_tone5',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_lotus_position',
        b'\xf0\x9f\xa7\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_lotus_position_tone1',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_lotus_position_tone2',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_lotus_position_tone3',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_lotus_position_tone4',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_lotus_position_tone5',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_steamy_room',
        b'\xf0\x9f\xa7\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_steamy_room_tone1',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_steamy_room_tone2',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_steamy_room_tone3',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_steamy_room_tone4',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_steamy_room_tone5',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_kneeling',
        b'\xf0\x9f\xa7\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'person_kneeling_tone1',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'person_kneeling_tone2',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_kneeling_tone3',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'person_kneeling_tone4',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'person_kneeling_tone5',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_standing',
        b'\xf0\x9f\xa7\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'person_standing_tone1',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'person_standing_tone2',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_standing_tone3',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'person_standing_tone4',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'person_standing_tone5',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'petri_dish',
        b'\xf0\x9f\xa7\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'pick',
        b'\xe2\x9b\x8f\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'pie',
        b'\xf0\x9f\xa5\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'pinching_hand',
        b'\xf0\x9f\xa4\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'pinching_hand_tone1',
        b'\xf0\x9f\xa4\x8f\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'pinching_hand_tone2',
        b'\xf0\x9f\xa4\x8f\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'pinching_hand_tone3',
        b'\xf0\x9f\xa4\x8f\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'pinching_hand_tone4',
        b'\xf0\x9f\xa4\x8f\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'pinching_hand_tone5',
        b'\xf0\x9f\xa4\x8f\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'pirate_flag',
        b'\xf0\x9f\x8f\xb4\xe2\x80\x8d\xe2\x98\xa0\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'play_pause',
        b'\xe2\x8f\xaf\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'pleading_face',
        b'\xf0\x9f\xa5\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'point_up',
        b'\xe2\x98\x9d\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'pretzel',
        b'\xf0\x9f\xa5\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'previous_track',
        b'\xe2\x8f\xae\xef\xb8\x8f',
        False,
        (
            'track_previous',
        ),
        None,
    ),
    Unicode(
        'printer',
        b'\xf0\x9f\x96\xa8\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'probing_cane',
        b'\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'purple_circle',
        b'\xf0\x9f\x9f\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'purple_square',
        b'\xf0\x9f\x9f\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'raccoon',
        b'\xf0\x9f\xa6\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'race_car',
        b'\xf0\x9f\x8f\x8e\xef\xb8\x8f',
        False,
        (
            'racing_car',
        ),
        None,
    ),
    Unicode(
        'radioactive',
        b'\xe2\x98\xa2\xef\xb8\x8f',
        False,
        (
            'radioactive_sign',
        ),
        None,
    ),
    Unicode(
        'railroad_track',
        b'\xf0\x9f\x9b\xa4\xef\xb8\x8f',
        False,
        (
            'railway_track',
        ),
        None,
    ),
    Unicode(
        'razor',
        b'\xf0\x9f\xaa\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'receipt',
        b'\xf0\x9f\xa7\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'record_button',
        b'\xe2\x8f\xba\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'recycle',
        b'\xe2\x99\xbb\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'red_envelope',
        b'\xf0\x9f\xa7\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'red_square',
        b'\xf0\x9f\x9f\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'registered',
        b'\xc2\xae\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'relaxed',
        b'\xe2\x98\xba\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'reminder_ribbon',
        b'\xf0\x9f\x8e\x97\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'ringed_planet',
        b'\xf0\x9f\xaa\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'roll_of_paper',
        b'\xf0\x9f\xa7\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'rosette',
        b'\xf0\x9f\x8f\xb5\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'sa',
        b'\xf0\x9f\x88\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'safety_pin',
        b'\xf0\x9f\xa7\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'safety_vest',
        b'\xf0\x9f\xa6\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'salt',
        b'\xf0\x9f\xa7\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'sandwich',
        b'\xf0\x9f\xa5\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'sari',
        b'\xf0\x9f\xa5\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'satellite_orbital',
        b'\xf0\x9f\x9b\xb0\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'sauropod',
        b'\xf0\x9f\xa6\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'scales',
        b'\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'scarf',
        b'\xf0\x9f\xa7\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'scissors',
        b'\xe2\x9c\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'scotland',
        (
            b'\xf0\x9f\x8f\xb4\xf3\xa0\x81\xa7\xf3\xa0\x81\xa2\xf3\xa0'
            b'\x81\xb3\xf3\xa0\x81\xa3\xf3\xa0\x81\xb4\xf3\xa0\x81\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'secret',
        b'\xe3\x8a\x99\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'service_dog',
        b'\xf0\x9f\x90\x95\xe2\x80\x8d\xf0\x9f\xa6\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'seven',
        b'7\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'shamrock',
        b'\xe2\x98\x98\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'shield',
        b'\xf0\x9f\x9b\xa1\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'shinto_shrine',
        b'\xe2\x9b\xa9\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'shopping_bags',
        b'\xf0\x9f\x9b\x8d\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'shorts',
        b'\xf0\x9f\xa9\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'shushing_face',
        b'\xf0\x9f\xa4\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'six',
        b'6\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'skateboard',
        b'\xf0\x9f\x9b\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'skier',
        b'\xe2\x9b\xb7\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'skull_and_crossbones',
        b'\xe2\x98\xa0\xef\xb8\x8f',
        False,
        (
            'skull_crossbones',
        ),
        None,
    ),
    Unicode(
        'skunk',
        b'\xf0\x9f\xa6\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'sled',
        b'\xf0\x9f\x9b\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'sloth',
        b'\xf0\x9f\xa6\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'smiling_face_with_3_hearts',
        b'\xf0\x9f\xa5\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'snowflake',
        b'\xe2\x9d\x84\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'snowman2',
        b'\xe2\x98\x83\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'soap',
        b'\xf0\x9f\xa7\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'socks',
        b'\xf0\x9f\xa7\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'softball',
        b'\xf0\x9f\xa5\x8e',
        False,
        None,
        None,
    ),
    Unicode(
        'spades',
        b'\xe2\x99\xa0\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'sparkle',
        b'\xe2\x9d\x87\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'speaking_head',
        b'\xf0\x9f\x97\xa3\xef\xb8\x8f',
        False,
        (
            'speaking_head_in_silhouette',
        ),
        None,
    ),
    Unicode(
        'spider',
        b'\xf0\x9f\x95\xb7\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'spider_web',
        b'\xf0\x9f\x95\xb8\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'sponge',
        b'\xf0\x9f\xa7\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'squeeze_bottle',
        b'\xf0\x9f\xa7\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'stadium',
        b'\xf0\x9f\x8f\x9f\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'star_and_crescent',
        b'\xe2\x98\xaa\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'star_of_david',
        b'\xe2\x9c\xa1\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'star_struck',
        b'\xf0\x9f\xa4\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'stethoscope',
        b'\xf0\x9f\xa9\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'stop_button',
        b'\xe2\x8f\xb9\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'stopwatch',
        b'\xe2\x8f\xb1\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'sunny',
        b'\xe2\x98\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'superhero',
        b'\xf0\x9f\xa6\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'superhero_tone1',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'superhero_tone2',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'superhero_tone3',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'superhero_tone4',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'superhero_tone5',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'supervillain',
        b'\xf0\x9f\xa6\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'supervillain_tone1',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'supervillain_tone2',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'supervillain_tone3',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'supervillain_tone4',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'supervillain_tone5',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'swan',
        b'\xf0\x9f\xa6\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        't_rex',
        b'\xf0\x9f\xa6\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'takeout_box',
        b'\xf0\x9f\xa5\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'teddy_bear',
        b'\xf0\x9f\xa7\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'telephone',
        b'\xe2\x98\x8e\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'test_tube',
        b'\xf0\x9f\xa7\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'thermometer',
        b'\xf0\x9f\x8c\xa1\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'thread',
        b'\xf0\x9f\xa7\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'three',
        b'3\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'thunder_cloud_and_rain',
        b'\xe2\x9b\x88\xef\xb8\x8f',
        False,
        (
            'thunder_cloud_rain',
        ),
        None,
    ),
    Unicode(
        'timer',
        b'\xe2\x8f\xb2\xef\xb8\x8f',
        False,
        (
            'timer_clock',
        ),
        None,
    ),
    Unicode(
        'tm',
        b'\xe2\x84\xa2\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'toolbox',
        b'\xf0\x9f\xa7\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'tooth',
        b'\xf0\x9f\xa6\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'trackball',
        b'\xf0\x9f\x96\xb2\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'two',
        b'2\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'u6708',
        b'\xf0\x9f\x88\xb7\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'umbrella2',
        b'\xe2\x98\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'united_nations',
        b'\xf0\x9f\x87\xba\xf0\x9f\x87\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'v',
        b'\xe2\x9c\x8c\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'vampire',
        b'\xf0\x9f\xa7\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'vampire_tone1',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'vampire_tone2',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'vampire_tone3',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'vampire_tone4',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'vampire_tone5',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'waffle',
        b'\xf0\x9f\xa7\x87',
        False,
        None,
        None,
    ),
    Unicode(
        'wales',
        (
            b'\xf0\x9f\x8f\xb4\xf3\xa0\x81\xa7\xf3\xa0\x81\xa2\xf3\xa0'
            b'\x81\xb7\xf3\xa0\x81\xac\xf3\xa0\x81\xb3\xf3\xa0\x81\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'warning',
        b'\xe2\x9a\xa0\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'wastebasket',
        b'\xf0\x9f\x97\x91\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'wavy_dash',
        b'\xe3\x80\xb0\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'wheel_of_dharma',
        b'\xe2\x98\xb8\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'white_heart',
        b'\xf0\x9f\xa4\x8d',
        False,
        None,
        None,
    ),
    Unicode(
        'white_medium_square',
        b'\xe2\x97\xbb\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'white_small_square',
        b'\xe2\x96\xab\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'white_sun_behind_cloud',
        b'\xf0\x9f\x8c\xa5\xef\xb8\x8f',
        False,
        (
            'white_sun_cloud',
        ),
        None,
    ),
    Unicode(
        'white_sun_behind_cloud_with_rain',
        b'\xf0\x9f\x8c\xa6\xef\xb8\x8f',
        False,
        (
            'white_sun_rain_cloud',
        ),
        None,
    ),
    Unicode(
        'white_sun_small_cloud',
        b'\xf0\x9f\x8c\xa4\xef\xb8\x8f',
        False,
        (
            'white_sun_with_small_cloud',
        ),
        None,
    ),
    Unicode(
        'wind_blowing_face',
        b'\xf0\x9f\x8c\xac\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_artist',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_artist_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_artist_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_artist_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_artist_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_artist_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_astronaut',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_astronaut_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_astronaut_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_astronaut_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_astronaut_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_astronaut_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bald',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bald_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bald_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bald_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bald_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bald_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_biking',
        b'\xf0\x9f\x9a\xb4\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_biking_tone1',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_biking_tone2',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_biking_tone3',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_biking_tone4',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_biking_tone5',
        b'\xf0\x9f\x9a\xb4\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bouncing_ball',
        b'\xe2\x9b\xb9\xef\xb8\x8f\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bouncing_ball_tone1',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bouncing_ball_tone2',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bouncing_ball_tone3',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bouncing_ball_tone4',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bouncing_ball_tone5',
        b'\xe2\x9b\xb9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bowing',
        b'\xf0\x9f\x99\x87\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bowing_tone1',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bowing_tone2',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bowing_tone3',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bowing_tone4',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_bowing_tone5',
        b'\xf0\x9f\x99\x87\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cartwheeling',
        b'\xf0\x9f\xa4\xb8\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cartwheeling_tone1',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cartwheeling_tone2',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cartwheeling_tone3',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cartwheeling_tone4',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cartwheeling_tone5',
        b'\xf0\x9f\xa4\xb8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_climbing',
        b'\xf0\x9f\xa7\x97\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_climbing_tone1',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_climbing_tone2',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_climbing_tone3',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_climbing_tone4',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_climbing_tone5',
        b'\xf0\x9f\xa7\x97\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_construction_worker',
        b'\xf0\x9f\x91\xb7\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_construction_worker_tone1',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_construction_worker_tone2',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_construction_worker_tone3',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_construction_worker_tone4',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_construction_worker_tone5',
        b'\xf0\x9f\x91\xb7\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cook',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cook_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cook_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cook_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cook_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_cook_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_curly_haired',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_curly_haired_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_curly_haired_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_curly_haired_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_curly_haired_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_curly_haired_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_detective',
        b'\xf0\x9f\x95\xb5\xef\xb8\x8f\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_detective_tone1',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_detective_tone2',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_detective_tone3',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_detective_tone4',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_detective_tone5',
        b'\xf0\x9f\x95\xb5\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_elf',
        b'\xf0\x9f\xa7\x9d\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_elf_tone1',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_elf_tone2',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_elf_tone3',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_elf_tone4',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_elf_tone5',
        b'\xf0\x9f\xa7\x9d\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_facepalming',
        b'\xf0\x9f\xa4\xa6\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_facepalming_tone1',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_facepalming_tone2',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_facepalming_tone3',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_facepalming_tone4',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_facepalming_tone5',
        b'\xf0\x9f\xa4\xa6\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_factory_worker',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_factory_worker_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_factory_worker_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_factory_worker_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_factory_worker_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_factory_worker_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_fairy',
        b'\xf0\x9f\xa7\x9a\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_fairy_tone1',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_fairy_tone2',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_fairy_tone3',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_fairy_tone4',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_fairy_tone5',
        b'\xf0\x9f\xa7\x9a\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_farmer',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_farmer_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_farmer_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_farmer_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_farmer_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_farmer_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_firefighter',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_firefighter_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_firefighter_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_firefighter_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_firefighter_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_firefighter_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_frowning',
        b'\xf0\x9f\x99\x8d\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_frowning_tone1',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_frowning_tone2',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_frowning_tone3',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_frowning_tone4',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_frowning_tone5',
        b'\xf0\x9f\x99\x8d\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_genie',
        b'\xf0\x9f\xa7\x9e\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_no',
        b'\xf0\x9f\x99\x85\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_no_tone1',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_no_tone2',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_no_tone3',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_no_tone4',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_no_tone5',
        b'\xf0\x9f\x99\x85\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_ok',
        b'\xf0\x9f\x99\x86\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_ok_tone1',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_ok_tone2',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_ok_tone3',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_ok_tone4',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_gesturing_ok_tone5',
        b'\xf0\x9f\x99\x86\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_face_massage',
        b'\xf0\x9f\x92\x86\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_face_massage_tone1',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_face_massage_tone2',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_face_massage_tone3',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_face_massage_tone4',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_face_massage_tone5',
        b'\xf0\x9f\x92\x86\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_haircut',
        b'\xf0\x9f\x92\x87\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_haircut_tone1',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_haircut_tone2',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_haircut_tone3',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_haircut_tone4',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_getting_haircut_tone5',
        b'\xf0\x9f\x92\x87\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_golfing',
        b'\xf0\x9f\x8f\x8c\xef\xb8\x8f\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_golfing_tone1',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_golfing_tone2',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_golfing_tone3',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_golfing_tone4',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_golfing_tone5',
        b'\xf0\x9f\x8f\x8c\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_guard',
        b'\xf0\x9f\x92\x82\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_guard_tone1',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_guard_tone2',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_guard_tone3',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_guard_tone4',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_guard_tone5',
        b'\xf0\x9f\x92\x82\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_health_worker',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_health_worker_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_health_worker_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_health_worker_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_health_worker_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_health_worker_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_lotus_position',
        b'\xf0\x9f\xa7\x98\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_lotus_position_tone1',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_lotus_position_tone2',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_lotus_position_tone3',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_lotus_position_tone4',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_lotus_position_tone5',
        b'\xf0\x9f\xa7\x98\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_manual_wheelchair',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_manual_wheelchair_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_manual_wheelchair_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_manual_wheelchair_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_manual_wheelchair_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_manual_wheelchair_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_motorized_wheelchair',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_motorized_wheelchair_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_motorized_wheelchair_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_motorized_wheelchair_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_motorized_wheelchair_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_motorized_wheelchair_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_steamy_room',
        b'\xf0\x9f\xa7\x96\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_steamy_room_tone1',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_steamy_room_tone2',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_steamy_room_tone3',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_steamy_room_tone4',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_steamy_room_tone5',
        b'\xf0\x9f\xa7\x96\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_judge',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_judge_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_judge_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_judge_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_judge_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_judge_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_juggling',
        b'\xf0\x9f\xa4\xb9\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_juggling_tone1',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_juggling_tone2',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_juggling_tone3',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_juggling_tone4',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_juggling_tone5',
        b'\xf0\x9f\xa4\xb9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_kneeling',
        b'\xf0\x9f\xa7\x8e\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_kneeling_tone1',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_kneeling_tone2',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_kneeling_tone3',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_kneeling_tone4',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_kneeling_tone5',
        b'\xf0\x9f\xa7\x8e\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_lifting_weights',
        b'\xf0\x9f\x8f\x8b\xef\xb8\x8f\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_lifting_weights_tone1',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_lifting_weights_tone2',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_lifting_weights_tone3',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_lifting_weights_tone4',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_lifting_weights_tone5',
        b'\xf0\x9f\x8f\x8b\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mage',
        b'\xf0\x9f\xa7\x99\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mage_tone1',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mage_tone2',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mage_tone3',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mage_tone4',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mage_tone5',
        b'\xf0\x9f\xa7\x99\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mechanic',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mechanic_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mechanic_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mechanic_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mechanic_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mechanic_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mountain_biking',
        b'\xf0\x9f\x9a\xb5\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mountain_biking_tone1',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mountain_biking_tone2',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mountain_biking_tone3',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mountain_biking_tone4',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_mountain_biking_tone5',
        b'\xf0\x9f\x9a\xb5\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_office_worker',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_office_worker_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_office_worker_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_office_worker_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_office_worker_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_office_worker_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pilot',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pilot_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pilot_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pilot_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pilot_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pilot_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_handball',
        b'\xf0\x9f\xa4\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_handball_tone1',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_handball_tone2',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_handball_tone3',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_handball_tone4',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_handball_tone5',
        b'\xf0\x9f\xa4\xbe\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_water_polo',
        b'\xf0\x9f\xa4\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_water_polo_tone1',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_water_polo_tone2',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_water_polo_tone3',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_water_polo_tone4',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_playing_water_polo_tone5',
        b'\xf0\x9f\xa4\xbd\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_police_officer',
        b'\xf0\x9f\x91\xae\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_police_officer_tone1',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_police_officer_tone2',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_police_officer_tone3',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_police_officer_tone4',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_police_officer_tone5',
        b'\xf0\x9f\x91\xae\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pouting',
        b'\xf0\x9f\x99\x8e\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pouting_tone1',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pouting_tone2',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pouting_tone3',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pouting_tone4',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_pouting_tone5',
        b'\xf0\x9f\x99\x8e\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_raising_hand',
        b'\xf0\x9f\x99\x8b\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_raising_hand_tone1',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_raising_hand_tone2',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_raising_hand_tone3',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_raising_hand_tone4',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_raising_hand_tone5',
        b'\xf0\x9f\x99\x8b\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_red_haired',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_red_haired_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_red_haired_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_red_haired_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_red_haired_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_red_haired_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_rowing_boat',
        b'\xf0\x9f\x9a\xa3\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_rowing_boat_tone1',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_rowing_boat_tone2',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_rowing_boat_tone3',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_rowing_boat_tone4',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_rowing_boat_tone5',
        b'\xf0\x9f\x9a\xa3\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_running',
        b'\xf0\x9f\x8f\x83\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_running_tone1',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_running_tone2',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_running_tone3',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_running_tone4',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_running_tone5',
        b'\xf0\x9f\x8f\x83\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_scientist',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_scientist_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_scientist_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_scientist_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_scientist_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_scientist_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_shrugging',
        b'\xf0\x9f\xa4\xb7\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_shrugging_tone1',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_shrugging_tone2',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_shrugging_tone3',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_shrugging_tone4',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_shrugging_tone5',
        b'\xf0\x9f\xa4\xb7\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_singer',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_singer_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_singer_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_singer_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_singer_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_singer_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_standing',
        b'\xf0\x9f\xa7\x8d\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_standing_tone1',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_standing_tone2',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_standing_tone3',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_standing_tone4',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_standing_tone5',
        b'\xf0\x9f\xa7\x8d\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_student',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_student_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_student_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_student_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_student_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_student_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_superhero',
        b'\xf0\x9f\xa6\xb8\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_superhero_tone1',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_superhero_tone2',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_superhero_tone3',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_superhero_tone4',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_superhero_tone5',
        b'\xf0\x9f\xa6\xb8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_supervillain',
        b'\xf0\x9f\xa6\xb9\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_supervillain_tone1',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_supervillain_tone2',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_supervillain_tone3',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_supervillain_tone4',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_supervillain_tone5',
        b'\xf0\x9f\xa6\xb9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_surfing',
        b'\xf0\x9f\x8f\x84\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_surfing_tone1',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_surfing_tone2',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_surfing_tone3',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_surfing_tone4',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_surfing_tone5',
        b'\xf0\x9f\x8f\x84\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_swimming',
        b'\xf0\x9f\x8f\x8a\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_swimming_tone1',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_swimming_tone2',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_swimming_tone3',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_swimming_tone4',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_swimming_tone5',
        b'\xf0\x9f\x8f\x8a\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_teacher',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_teacher_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_teacher_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_teacher_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_teacher_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_teacher_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_technologist',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_technologist_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_technologist_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_technologist_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_technologist_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_technologist_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tipping_hand',
        b'\xf0\x9f\x92\x81\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tipping_hand_tone1',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tipping_hand_tone2',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tipping_hand_tone3',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tipping_hand_tone4',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tipping_hand_tone5',
        b'\xf0\x9f\x92\x81\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_vampire',
        b'\xf0\x9f\xa7\x9b\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_vampire_tone1',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_vampire_tone2',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_vampire_tone3',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_vampire_tone4',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_vampire_tone5',
        b'\xf0\x9f\xa7\x9b\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_walking',
        b'\xf0\x9f\x9a\xb6\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_walking_tone1',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_walking_tone2',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_walking_tone3',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_walking_tone4',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_walking_tone5',
        b'\xf0\x9f\x9a\xb6\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_wearing_turban',
        b'\xf0\x9f\x91\xb3\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_wearing_turban_tone1',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_wearing_turban_tone2',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_wearing_turban_tone3',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_wearing_turban_tone4',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_wearing_turban_tone5',
        b'\xf0\x9f\x91\xb3\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_white_haired',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_white_haired_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_white_haired_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_white_haired_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_white_haired_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_white_haired_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_headscarf',
        b'\xf0\x9f\xa7\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_headscarf_tone1',
        b'\xf0\x9f\xa7\x95\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_headscarf_tone2',
        b'\xf0\x9f\xa7\x95\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_headscarf_tone3',
        b'\xf0\x9f\xa7\x95\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_headscarf_tone4',
        b'\xf0\x9f\xa7\x95\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_headscarf_tone5',
        b'\xf0\x9f\xa7\x95\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_probing_cane',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_probing_cane_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_probing_cane_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_probing_cane_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_probing_cane_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_probing_cane_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_zombie',
        b'\xf0\x9f\xa7\x9f\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'womans_flat_shoe',
        b'\xf0\x9f\xa5\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'women_with_bunny_ears_partying',
        b'\xf0\x9f\x91\xaf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'women_wrestling',
        b'\xf0\x9f\xa4\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woozy_face',
        b'\xf0\x9f\xa5\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'writing_hand',
        b'\xe2\x9c\x8d\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'yarn',
        b'\xf0\x9f\xa7\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'yawning_face',
        b'\xf0\x9f\xa5\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'yellow_circle',
        b'\xf0\x9f\x9f\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'yellow_square',
        b'\xf0\x9f\x9f\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'yin_yang',
        b'\xe2\x98\xaf\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'yo_yo',
        b'\xf0\x9f\xaa\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'zany_face',
        b'\xf0\x9f\xa4\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'zebra',
        b'\xf0\x9f\xa6\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'zero',
        b'0\xef\xb8\x8f\xe2\x83\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'zombie',
        b'\xf0\x9f\xa7\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'technologist_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'firefighter_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'man_feeding_baby_tone4',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'man_feeding_baby_tone5',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'technologist_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'firefighter_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone1',
        b'\xf0\x9f\x91\xad\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone1_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'sewing_needle',
        b'\xf0\x9f\xaa\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'technologist',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone1_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone1_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'technologist_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'technologist_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone1_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'plunger',
        b'\xf0\x9f\xaa\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'technologist_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x92\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone2_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone2',
        b'\xf0\x9f\x91\xad\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'bucket',
        b'\xf0\x9f\xaa\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone2_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone3_curly_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        (
            'person_medium_skin_tone_curly_hair',
        ),
        None,
    ),
    Unicode(
        'firefighter',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone2_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone5_curly_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        (
            'person_dark_skin_tone_curly_hair',
        ),
        None,
    ),
    Unicode(
        'women_holding_hands_tone2_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'firefighter_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'man_feeding_baby',
        b'\xf0\x9f\x91\xa8\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone3_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'firefighter_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'firefighter_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x9a\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'man_feeding_baby_tone1',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone3_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'man_feeding_baby_tone2',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone3',
        b'\xf0\x9f\x91\xad\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'man_feeding_baby_tone3',
        b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone3_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone3_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone4_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone4_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'teacher_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone4_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'teacher_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone4',
        b'\xf0\x9f\x91\xad\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone4_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone5_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone5_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone5_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone5_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'women_holding_hands_tone5',
        b'\xf0\x9f\x91\xad\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_curly_hair',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone1_curly_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        (
            'person_light_skin_tone_curly_hair',
        ),
        None,
    ),
    Unicode(
        'mouse_trap',
        b'\xf0\x9f\xaa\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone2_curly_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        (
            'person_medium_light_skin_tone_curly_hair',
        ),
        None,
    ),
    Unicode(
        'person_tone4_curly_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb1',
        False,
        (
            'person_medium_dark_skin_tone_curly_hair',
        ),
        None,
    ),
    Unicode(
        'piñata',
        b'\xf0\x9f\xaa\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'teacher',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'nesting_dolls',
        b'\xf0\x9f\xaa\x86',
        False,
        None,
        None,
    ),
    Unicode(
        'teacher_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'seal',
        b'\xf0\x9f\xa6\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'teacher_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'teacher_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8f\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'mammoth',
        b'\xf0\x9f\xa6\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'bison',
        b'\xf0\x9f\xa6\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'window',
        b'\xf0\x9f\xaa\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_veil_tone4',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_veil_tone5',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'health_worker_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'health_worker_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_veil',
        b'\xf0\x9f\x91\xb0\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_veil_tone1',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_veil_tone2',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'man_with_veil_tone3',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'health_worker',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'health_worker_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'health_worker_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'health_worker_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9a\x95\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'scientist_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'scientist_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_tuxedo_tone4',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'scientist',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_tuxedo_tone5',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'scientist_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'scientist_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'scientist_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x94\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'judge_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone4_white_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        (
            'person_medium_dark_skin_tone_white_hair',
        ),
        None,
    ),
    Unicode(
        'judge_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone5_white_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        (
            'person_dark_skin_tone_white_hair',
        ),
        None,
    ),
    Unicode(
        'woman_in_tuxedo_tone3',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'judge_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_manual_wheelchair',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_manual_wheelchair_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_manual_wheelchair_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'feather',
        b'\xf0\x9f\xaa\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'elevator',
        b'\xf0\x9f\x9b\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'rock',
        b'\xf0\x9f\xaa\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'wood',
        b'\xf0\x9f\xaa\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'potted_plant',
        b'\xf0\x9f\xaa\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'beaver',
        b'\xf0\x9f\xa6\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone3_red_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        (
            'person_medium_skin_tone_red_hair',
        ),
        None,
    ),
    Unicode(
        'person_tone5_red_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        (
            'person_dark_skin_tone_red_hair',
        ),
        None,
    ),
    Unicode(
        'factory_worker_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'factory_worker_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'tamale',
        b'\xf0\x9f\xab\x94',
        False,
        None,
        None,
    ),
    Unicode(
        'person_red_hair',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone1_red_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        (
            'person_light_skin_tone_red_hair',
        ),
        None,
    ),
    Unicode(
        'fondue',
        b'\xf0\x9f\xab\x95',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone2_red_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        (
            'person_medium_light_skin_tone_red_hair',
        ),
        None,
    ),
    Unicode(
        'person_tone4_red_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb0',
        False,
        (
            'person_medium_dark_skin_tone_red_hair',
        ),
        None,
    ),
    Unicode(
        'black_cat',
        b'\xf0\x9f\x90\x88\xe2\x80\x8d\xe2\xac\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'factory_worker',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'factory_worker_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'factory_worker_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'factory_worker_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8f\xad',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_motorized_wheelchair_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_white_hair',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'student_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'office_worker_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'office_worker_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'judge',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_tuxedo',
        b'\xf0\x9f\xa4\xb5\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_tuxedo_tone1',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_motorized_wheelchair_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone1_white_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        (
            'person_light_skin_tone_white_hair',
        ),
        None,
    ),
    Unicode(
        'judge_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone2_white_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        (
            'person_medium_light_skin_tone_white_hair',
        ),
        None,
    ),
    Unicode(
        'student_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'judge_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9a\x96\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_in_tuxedo_tone2',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone3_white_hair',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb3',
        False,
        (
            'person_medium_skin_tone_white_hair',
        ),
        None,
    ),
    Unicode(
        'student',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone3',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'person_in_tuxedo_tone3',
            'tuxedo_tone3'
        ),
        None,
    ),
    Unicode(
        'student_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone5',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'person_in_tuxedo_tone5',
            'tuxedo_tone5'
        ),
        None,
    ),
    Unicode(
        'person_in_motorized_wheelchair',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'student_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'student_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_motorized_wheelchair_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_motorized_wheelchair_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_motorized_wheelchair_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'farmer_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'farmer_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'office_worker',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'carpentry_saw',
        b'\xf0\x9f\xaa\x9a',
        False,
        None,
        None,
    ),
    Unicode(
        'office_worker_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'office_worker_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'pickup_truck',
        b'\xf0\x9f\x9b\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'office_worker_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x92\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_probing_cane_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_probing_cane_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo',
        b'\xf0\x9f\xa4\xb5\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'person_in_tuxedo',
        ),
        None,
    ),
    Unicode(
        'headstone',
        b'\xf0\x9f\xaa\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone1',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'person_in_tuxedo_tone1',
            'tuxedo_tone1'
        ),
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone2',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'person_in_tuxedo_tone2',
            'tuxedo_tone2'
        ),
        None,
    ),
    Unicode(
        'man_in_tuxedo_tone4',
        b'\xf0\x9f\xa4\xb5\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'person_in_tuxedo_tone4',
            'tuxedo_tone4'
        ),
        None,
    ),
    Unicode(
        'magic_wand',
        b'\xf0\x9f\xaa\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'farmer',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'farmer_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'farmer_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'farmer_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8c\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'roller_skate',
        b'\xf0\x9f\x9b\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'long_drum',
        b'\xf0\x9f\xaa\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_probing_cane',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_probing_cane_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_probing_cane_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_probing_cane_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xaf',
        False,
        None,
        None,
    ),
    Unicode(
        'artist_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'artist_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'accordion',
        b'\xf0\x9f\xaa\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'boomerang',
        b'\xf0\x9f\xaa\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'hook',
        b'\xf0\x9f\xaa\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'knot',
        b'\xf0\x9f\xaa\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'artist',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'artist_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'artist_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'artist_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone1',
        b'\xf0\x9f\x91\xac\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone1_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone1_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone1_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'transgender_flag',
        b'\xf0\x9f\x8f\xb3\xef\xb8\x8f\xe2\x80\x8d\xe2\x9a\xa7\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'pilot_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone1_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'thong_sandal',
        b'\xf0\x9f\xa9\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'pilot_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone2_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone2',
        b'\xf0\x9f\x91\xac\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone2_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone2_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'olive',
        b'\xf0\x9f\xab\x92',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone2_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone3_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone3_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'bell_pepper',
        b'\xf0\x9f\xab\x91',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone3',
        b'\xf0\x9f\x91\xac\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone3_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone3_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone4_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone4_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'blueberries',
        b'\xf0\x9f\xab\x90',
        False,
        None,
        None,
    ),
    Unicode(
        'pilot',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone4_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone4',
        b'\xf0\x9f\x91\xac\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'pilot_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone4_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'pilot_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'pilot_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9c\x88\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone5_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone5_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone5_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone5_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'men_holding_hands_tone5',
        b'\xf0\x9f\x91\xac\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'military_helmet',
        b'\xf0\x9f\xaa\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'teapot',
        b'\xf0\x9f\xab\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'bubble_tea',
        b'\xf0\x9f\xa7\x8b',
        False,
        None,
        None,
    ),
    Unicode(
        'ninja_tone4',
        b'\xf0\x9f\xa5\xb7\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'ninja_tone5',
        b'\xf0\x9f\xa5\xb7\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'smiling_face_with_tear',
        b'\xf0\x9f\xa5\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'singer_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'singer_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'ninja',
        b'\xf0\x9f\xa5\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'ninja_tone1',
        b'\xf0\x9f\xa5\xb7\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'person_feeding_baby_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'ladder',
        b'\xf0\x9f\xaa\x9c',
        False,
        None,
        None,
    ),
    Unicode(
        'ninja_tone2',
        b'\xf0\x9f\xa5\xb7\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_feeding_baby_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'ninja_tone3',
        b'\xf0\x9f\xa5\xb7\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'screwdriver',
        b'\xf0\x9f\xaa\x9b',
        False,
        None,
        None,
    ),
    Unicode(
        'mx_claus_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8e\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'mx_claus_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8e\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_manual_wheelchair_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'singer',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'singer_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'singer_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'singer_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'lungs',
        b'\xf0\x9f\xab\x81',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_veil_tone3',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_veil_tone5',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'person_feeding_baby',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'flatbread',
        b'\xf0\x9f\xab\x93',
        False,
        None,
        None,
    ),
    Unicode(
        'person_feeding_baby_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_feeding_baby_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_feeding_baby_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'people_hugging',
        b'\xf0\x9f\xab\x82',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_feeding_baby_tone3',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_feeding_baby_tone5',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'mx_claus',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x8e\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'mx_claus_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8e\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'mx_claus_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8e\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'mx_claus_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8e\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_veil',
        b'\xf0\x9f\x91\xb0\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        (
            'bride_with_veil',
        ),
        None,
    ),
    Unicode(
        'coin',
        b'\xf0\x9f\xaa\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_veil_tone1',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone4_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_veil_tone2',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone4_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_with_veil_tone4',
        b'\xf0\x9f\x91\xb0\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone4_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone5_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'anatomical_heart',
        b'\xf0\x9f\xab\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_feeding_baby',
        b'\xf0\x9f\x91\xa9\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone5_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_feeding_baby_tone1',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone5_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_feeding_baby_tone2',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone5_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_feeding_baby_tone4',
        b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8d\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'pinched_fingers_tone3',
        b'\xf0\x9f\xa4\x8c\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'pinched_fingers_tone5',
        b'\xf0\x9f\xa4\x8c\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'cockroach',
        b'\xf0\x9f\xaa\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'beetle',
        b'\xf0\x9f\xaa\xb2',
        False,
        (
            'lady_beetle',
        ),
        None,
    ),
    Unicode(
        'mechanic_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'mechanic_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'disguised_face',
        b'\xf0\x9f\xa5\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'pinched_fingers',
        b'\xf0\x9f\xa4\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'pinched_fingers_tone2',
        b'\xf0\x9f\xa4\x8c\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'fly',
        b'\xf0\x9f\xaa\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'pinched_fingers_tone1',
        b'\xf0\x9f\xa4\x8c\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'pinched_fingers_tone4',
        b'\xf0\x9f\xa4\x8c\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'placard',
        b'\xf0\x9f\xaa\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'polar_bear',
        b'\xf0\x9f\x90\xbb\xe2\x80\x8d\xe2\x9d\x84\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'astronaut_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'mechanic',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'astronaut_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'mechanic_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'mechanic_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'mechanic_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x94\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'cook_tone3',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'cook_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'mirror',
        b'\xf0\x9f\xaa\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'astronaut',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'toothbrush',
        b'\xf0\x9f\xaa\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'cook',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'astronaut_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'worm',
        b'\xf0\x9f\xaa\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'astronaut_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'dodo',
        b'\xf0\x9f\xa6\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'cook_tone1',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'astronaut_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x9a\x80',
        False,
        None,
        None,
    ),
    Unicode(
        'cook_tone2',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'cook_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\x8d\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'hut',
        b'\xf0\x9f\x9b\x96',
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone4_bald',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        (
            'person_medium_dark_skin_tone_bald',
        ),
        None,
    ),
    Unicode(
        'person_tone5_bald',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        (
            'person_dark_skin_tone_bald',
        ),
        None,
    ),
    Unicode(
        'people_holding_hands_tone1_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone1_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'person_bald',
        b'\xf0\x9f\xa7\x91\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone2_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone2_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'person_tone1_bald',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        (
            'person_light_skin_tone_bald',
        ),
        None,
    ),
    Unicode(
        'person_tone2_bald',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        (
            'person_medium_light_skin_tone_bald',
        ),
        None,
    ),
    Unicode(
        'person_tone3_bald',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xa6\xb2',
        False,
        (
            'person_medium_skin_tone_bald',
        ),
        None,
    ),
    Unicode(
        'people_holding_hands_tone2_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'transgender_symbol',
        b'\xe2\x9a\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone2_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone3_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone3_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone3_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone3_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone4_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone1_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'people_holding_hands_tone1_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone1',
        b'\xf0\x9f\x91\xab\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone1_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone1_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone1_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone1_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone2_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone2',
        b'\xf0\x9f\x91\xab\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_manual_wheelchair_tone4',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone2_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'person_in_manual_wheelchair_tone5',
        b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xa6\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone2_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone2_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone3_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone3_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone3',
        b'\xf0\x9f\x91\xab\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone3_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone3_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone4_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone4_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone4_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone4',
        b'\xf0\x9f\x91\xab\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone5',
        b'\xf0\x9f\x91\xab\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone4_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone5_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone5_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone5_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_and_man_holding_hands_tone5_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f'
            b'\xa4\x9d\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'man_tone4_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'man_medium_dark_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'man_tone5_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'man_dark_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone5_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_tone1',
        b'\xf0\x9f\x92\x91\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone1_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_tone1',
        b'\xf0\x9f\x92\x8f\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone1_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone1_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone1_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone1_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone1_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone1_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone1_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone1_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone1_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone1_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone1_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone1_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone1_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone1_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone1_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone1_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone1_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone2_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone1_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone2_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone2_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone1_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone2_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone2_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'man_beard',
        b'\xf0\x9f\xa7\x94\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_tone2',
        b'\xf0\x9f\x92\x91\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone2_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone2_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone2_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone2_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_tone2',
        b'\xf0\x9f\x92\x8f\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone2_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'man_tone1_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'man_light_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone2_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone2_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone2_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'man_tone2_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'man_medium_light_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone2_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone2_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone2_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'man_tone3_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x82\xef\xb8\x8f',
        False,
        (
            'man_medium_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone3_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone3_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone2_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone3_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone3_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone3_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone3_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone3_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone3_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone3_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_tone3',
        b'\xf0\x9f\x92\x91\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_tone3',
        b'\xf0\x9f\x92\x8f\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone3_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tone3_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        (
            'woman_medium_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone3_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone3_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tone5_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        (
            'woman_dark_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone3_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone3_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone3_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone3_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone4_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone3_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone3_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone4_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone4_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone4_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone4_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone4_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone4_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone4_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone4_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone4_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone4_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone4_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone4_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone4_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone4_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_tone4',
        b'\xf0\x9f\x92\x91\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone4_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_tone4',
        b'\xf0\x9f\x92\x8f\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone4_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone4_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone5_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone4_tone5',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone5_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone5_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone4_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone5_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone5_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone5_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone5_tone1',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone5_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone5_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone5_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone5_tone2',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone5_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone5_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_person_person_tone5_tone3',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_person_person_tone5_tone4',
        (
            b'\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa7\x91\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone5_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_tone5',
        b'\xf0\x9f\x92\x8f\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone5_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_tone5',
        b'\xf0\x9f\x92\x91\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_man_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_woman_woman_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone5_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone2_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_beard',
        b'\xf0\x9f\xa7\x94\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone2_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'woman_tone1_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        (
            'woman_light_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'woman_tone2_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        (
            'woman_medium_light_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'woman_tone4_beard',
        b'\xf0\x9f\xa7\x94\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x99\x80\xef\xb8\x8f',
        False,
        (
            'woman_medium_dark_skin_tone_beard',
        ),
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone2_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone1_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone1_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone1_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone1_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone1_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone1_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone1_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone1_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone1_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone1_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone3_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone1_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone1_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbb\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone2_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone2_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone2_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone2_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone2_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone2_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone2_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone2_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone2_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone2_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone2_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'mending_heart',
        b'\xe2\x9d\xa4\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\xa9\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone3_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone2_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone3_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone3_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone3_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone3_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone3_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone3_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone3_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone3_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'heart_on_fire',
        b'\xe2\x9d\xa4\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x94\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone3_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone3_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone3_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone3_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbd\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone4_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone4_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone4_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone4_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'face_exhaling',
        b'\xf0\x9f\x98\xae\xe2\x80\x8d\xf0\x9f\x92\xa8',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone4_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone4_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone4_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone4_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone4_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone4_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone4_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone4_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbe\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone5_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone5_tone1',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone5_tone1',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbb'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'face_in_clouds',
        b'\xf0\x9f\x98\xb6\xe2\x80\x8d\xf0\x9f\x8c\xab\xef\xb8\x8f',
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone5_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone5_tone2',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone5_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone5_tone3',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone5_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone5_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone5_tone4',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone5_tone3',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbd'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_man_man_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'couple_with_heart_man_man_tone5',
        (
            b'\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4'
            b'\xef\xb8\x8f\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone5_tone4',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbe'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_man_tone5',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa8\xf0\x9f\x8f\xbf'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'face_with_spiral_eyes',
        b'\xf0\x9f\x98\xb5\xe2\x80\x8d\xf0\x9f\x92\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'kiss_woman_woman_tone5_tone2',
        (
            b'\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbf\xe2\x80\x8d\xe2\x9d\xa4\xef\xb8\x8f\xe2'
            b'\x80\x8d\xf0\x9f\x92\x8b\xe2\x80\x8d\xf0\x9f\x91\xa9\xf0\x9f\x8f\xbc'
        ),
        False,
        None,
        None,
    ),
    Unicode(
        'infinity',
        b'\xe2\x99\xbe',
        True,
        None,
        None,
    ),
    Unicode(
        'dotted_line_face',
        b'\xf0\x9f\xab\xa5',
        False,
        None,
        None,
    ),
    Unicode(
        'face_holding_back_tears',
        b'\xf0\x9f\xa5\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'face_with_peeking_eye',
        b'\xf0\x9f\xab\xa3',
        False,
        None,
        None,
    ),
    Unicode(
        'face_with_open_eyes_and_hand_over_mouth',
        b'\xf0\x9f\xab\xa2',
        False,
        None,
        None,
    ),
    Unicode(
        'saluting_face',
        b'\xf0\x9f\xab\xa1',
        False,
        None,
        None,
    ),
    Unicode(
        'melting_face',
        b'\xf0\x9f\xab\xa0',
        False,
        None,
        None,
    ),
    Unicode(
        'face_with_diagonal_mouth',
        b'\xf0\x9f\xab\xa4',
        False,
        None,
        None,
    ),
    Unicode(
        'heart_hands',
        b'\xf0\x9f\xab\xb6',
        False,
        None,
        None,
    ),
    Unicode(
        'heart_hands_tone1',
        b'\xf0\x9f\xab\xb6\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'heart_hands_tone2',
        b'\xf0\x9f\xab\xb6\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'heart_hands_tone3',
        b'\xf0\x9f\xab\xb6\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'heart_hands_tone4',
        b'\xf0\x9f\xab\xb6\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'heart_hands_tone5',
        b'\xf0\x9f\xab\xb6\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone1',
        b'\xf0\x9f\xa4\x9d\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone1_tone2',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone1_tone3',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone1_tone4',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone1_tone5',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbb\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone2_tone1',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone2',
        b'\xf0\x9f\xa4\x9d\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone2_tone3',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone2_tone4',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone2_tone5',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbc\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone3_tone1',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone3_tone2',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone3',
        b'\xf0\x9f\xa4\x9d\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone3_tone4',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone3_tone5',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbd\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone4_tone1',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone4_tone2',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone4_tone3',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone4',
        b'\xf0\x9f\xa4\x9d\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone4_tone5',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbe\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone5_tone1',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone5_tone2',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone5_tone3',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone5_tone4',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbf\xe2\x80\x8d\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'handshake_tone5',
        b'\xf0\x9f\xa4\x9d\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'hand_with_index_finger_and_thumb_crossed',
        b'\xf0\x9f\xab\xb0',
        False,
        None,
        None,
    ),
    Unicode(
        'hand_with_index_finger_and_thumb_crossed_tone1',
        b'\xf0\x9f\xab\xb0\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'hand_with_index_finger_and_thumb_crossed_tone2',
        b'\xf0\x9f\xab\xb0\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'hand_with_index_finger_and_thumb_crossed_tone3',
        b'\xf0\x9f\xab\xb0\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'hand_with_index_finger_and_thumb_crossed_tone4',
        b'\xf0\x9f\xab\xb0\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'hand_with_index_finger_and_thumb_crossed_tone5',
        b'\xf0\x9f\xab\xb0\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_down_hand',
        b'\xf0\x9f\xab\xb3',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_down_hand_tone1',
        b'\xf0\x9f\xab\xb3\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_down_hand_tone2',
        b'\xf0\x9f\xab\xb3\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_down_hand_tone3',
        b'\xf0\x9f\xab\xb3\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_down_hand_tone4',
        b'\xf0\x9f\xab\xb3\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_down_hand_tone5',
        b'\xf0\x9f\xab\xb3\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_up_hand',
        b'\xf0\x9f\xab\xb4',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_up_hand_tone1',
        b'\xf0\x9f\xab\xb4\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_up_hand_tone2',
        b'\xf0\x9f\xab\xb4\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_up_hand_tone3',
        b'\xf0\x9f\xab\xb4\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_up_hand_tone4',
        b'\xf0\x9f\xab\xb4\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'palm_up_hand_tone5',
        b'\xf0\x9f\xab\xb4\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'leftwards_hand',
        b'\xf0\x9f\xab\xb2',
        False,
        None,
        None,
    ),
    Unicode(
        'leftwards_hand_tone1',
        b'\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'leftwards_hand_tone2',
        b'\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'leftwards_hand_tone3',
        b'\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'leftwards_hand_tone4',
        b'\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'leftwards_hand_tone5',
        b'\xf0\x9f\xab\xb2\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'rightwards_hand',
        b'\xf0\x9f\xab\xb1',
        False,
        None,
        None,
    ),
    Unicode(
        'rightwards_hand_tone1',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'rightwards_hand_tone2',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'rightwards_hand_tone3',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'rightwards_hand_tone4',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'rightwards_hand_tone5',
        b'\xf0\x9f\xab\xb1\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'index_pointing_at_the_viewer',
        b'\xf0\x9f\xab\xb5',
        False,
        None,
        None,
    ),
    Unicode(
        'index_pointing_at_the_viewer_tone1',
        b'\xf0\x9f\xab\xb5\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'index_pointing_at_the_viewer_tone2',
        b'\xf0\x9f\xab\xb5\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'index_pointing_at_the_viewer_tone3',
        b'\xf0\x9f\xab\xb5\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'index_pointing_at_the_viewer_tone4',
        b'\xf0\x9f\xab\xb5\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'index_pointing_at_the_viewer_tone5',
        b'\xf0\x9f\xab\xb5\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'biting_lip',
        b'\xf0\x9f\xab\xa6',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_crown',
        b'\xf0\x9f\xab\x85',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_crown_tone1',
        b'\xf0\x9f\xab\x85\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_crown_tone2',
        b'\xf0\x9f\xab\x85\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_crown_tone3',
        b'\xf0\x9f\xab\x85\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_crown_tone4',
        b'\xf0\x9f\xab\x85\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'person_with_crown_tone5',
        b'\xf0\x9f\xab\x85\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'troll',
        b'\xf0\x9f\xa7\x8c',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_person',
        b'\xf0\x9f\xab\x84',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_person_tone1',
        b'\xf0\x9f\xab\x84\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_person_tone2',
        b'\xf0\x9f\xab\x84\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_person_tone3',
        b'\xf0\x9f\xab\x84\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_person_tone4',
        b'\xf0\x9f\xab\x84\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_person_tone5',
        b'\xf0\x9f\xab\x84\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_man',
        b'\xf0\x9f\xab\x83',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_man_tone1',
        b'\xf0\x9f\xab\x83\xf0\x9f\x8f\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_man_tone2',
        b'\xf0\x9f\xab\x83\xf0\x9f\x8f\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_man_tone3',
        b'\xf0\x9f\xab\x83\xf0\x9f\x8f\xbd',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_man_tone4',
        b'\xf0\x9f\xab\x83\xf0\x9f\x8f\xbe',
        False,
        None,
        None,
    ),
    Unicode(
        'pregnant_man_tone5',
        b'\xf0\x9f\xab\x83\xf0\x9f\x8f\xbf',
        False,
        None,
        None,
    ),
    Unicode(
        'empty_nest',
        b'\xf0\x9f\xaa\xb9',
        False,
        None,
        None,
    ),
    Unicode(
        'nest_with_eggs',
        b'\xf0\x9f\xaa\xba',
        False,
        None,
        None,
    ),
    Unicode(
        'coral',
        b'\xf0\x9f\xaa\xb8',
        False,
        None,
        None,
    ),
    Unicode(
        'lotus',
        b'\xf0\x9f\xaa\xb7',
        False,
        None,
        None,
    ),
    Unicode(
        'bubbles',
        b'\xf0\x9f\xab\xa7',
        False,
        None,
        None,
    ),
    Unicode(
        'jar',
        b'\xf0\x9f\xab\x99',
        False,
        None,
        None,
    ),
    Unicode(
        'beans',
        b'\xf0\x9f\xab\x98',
        False,
        None,
        None,
    ),
    Unicode(
        'pouring_liquid',
        b'\xf0\x9f\xab\x97',
        False,
        None,
        None,
    ),
    Unicode(
        'playground_slide',
        b'\xf0\x9f\x9b\x9d',
        False,
        None,
        None,
    ),
    Unicode(
        'crutch',
        b'\xf0\x9f\xa9\xbc',
        False,
        None,
        None,
    ),
    Unicode(
        'wheel',
        b'\xf0\x9f\x9b\x9e',
        False,
        None,
        None,
    ),
    Unicode(
        'ring_buoy',
        b'\xf0\x9f\x9b\x9f',
        False,
        None,
        None,
    ),
    Unicode(
        'low_battery',
        b'\xf0\x9f\xaa\xab',
        False,
        None,
        None,
    ),
    Unicode(
        'identification_card',
        b'\xf0\x9f\xaa\xaa',
        False,
        None,
        None,
    ),
    Unicode(
        'hamsa',
        b'\xf0\x9f\xaa\xac',
        False,
        None,
        None,
    ),
    Unicode(
        'x_ray',
        b'\xf0\x9f\xa9\xbb',
        False,
        None,
        None,
    ),
    Unicode(
        'mirror_ball',
        b'\xf0\x9f\xaa\xa9',
        False,
        None,
        None,
    ),
    Unicode(
        'heavy_equals_sign',
        b'\xf0\x9f\x9f\xb0',
        False,
        None,
        None,
    ),
)


for unicode in UNICODES:
    Emoji._create_unicode(unicode, True)

unicode = None
del unicode
